/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core;

import org.apache.log4j.Logger;

import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.TimeZone;

/**
 * Class is used to format logging messages. Messages
 * are routed to stdout and through a log4j appender.
 */
public class Log {
    
    /**
     * Toggle indicates whether to print out logging message.
     */
    private static boolean m_quiet = true;
    
    /**
     * Program name (which is printed out when a logging call is made.
     */
    public static String PROG_NAME = "apel";

    /**
     * Return string containing date/time
     */
    private static String getDateTime() {
        SimpleDateFormat dateFormatter = new SimpleDateFormat(
                "EEE MMM dd HH:mm:ss zzz yyyy");
        dateFormatter.setTimeZone(TimeZone.getTimeZone("UTC"));
        String date = dateFormatter.format(new Date());
        if (date.charAt(8) == '0') {
            date = date.replaceFirst(" 0", "  ");
        }
        return date;
    }

    /**
     * Sets the program name that will be printed when each
     * user log message is displayed.
     */
    public static void setProgramName(String progName) {
    	PROG_NAME = progName;
    }
    
    /**
     * Switch verbose mode on/off. When off, normal messages are not printed,
     * only warning messages.
     */
    public static void setQuiet(boolean onOrOff) {
        m_quiet = onOrOff;
    }

    /**
     * Terminates the program following an error or exception
     *
     * @param log The log4j logger
     * @param message Describes what the program was trying to do
     * @param e Exception that was thrown, or null if none
     */
    public static void die(Logger log, String message, Exception e) {
        System.err.println(getDateTime() + ": " + PROG_NAME +
            " - program aborted");
        if (!message.equals("")) {
            System.err.println(getDateTime() + ": " + PROG_NAME +
                " - " + message);
        }
        log.error("Fatal error");
        log.error(message);
        if (e != null) {
            log.error(e);
            e.printStackTrace(System.err);
        }
        System.exit(1);
    }

    /**
     * Terminates the program following an error
     *
     * @param log is the log4j logger
     * @param message - Describes what the program was trying to do
     */
    public static void die(Logger log, String message) {
        die(log, message, null);
    }

    /**
     * Terminates the program following an error
     *
     * @param log is the log4j logger
     * @param e Exception that was thrown, or null if none
     */
    public static void die(Logger log, Exception e) {
        die(log, "", e);
    }
    
    /**
     * Display general message about recoverable condition inside program.
     *
     * @param log is the log4j logger
     * @param message is the message to display
     */
    public static void info(Logger log, String message) {
        if (!m_quiet) {
            System.out.println(getDateTime() + ": " + PROG_NAME + " - " +
                message);
            System.out.flush();
            log.info(message);
        }
    }

    /**
     * Display warning message about recoverable condition inside program.
     *
     * @param log is the log4j logger
     * @param message is the message to display
     * @param e is the exception, or null
     */
    public static void warn(Logger log, String message, Exception e) {
        System.err.println(getDateTime() + ": " + PROG_NAME +
            " - WARNING - " + message);
        log.warn(message);
        if (e != null) {
            log.warn(e);
            e.printStackTrace(System.err);
        }
    }

    /**
     * Display warning message about recoverable condition inside program.
     *
     * @param log is the log4j logger
     * @param message is the message to display
     */
    public static void warn(Logger log, String message) {
        warn(log, message, null);
    }
    
}
