/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core;

import java.io.FileInputStream;
import javax.crypto.CipherOutputStream;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.File;
import java.io.Writer;
import java.io.InputStream;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.interfaces.RSAPublicKey;
import java.security.KeyFactory;
import java.security.Security;
import java.security.spec.EncodedKeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.security.SecureRandom;
import javax.crypto.Cipher;
import org.bouncycastle.util.encoders.Base64;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import sun.misc.BASE64Decoder;
import sun.misc.BASE64Encoder;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import org.glite.apel.core.Log;
import org.apache.log4j.Logger;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateFactory;

/**
 * Contains a list of misc methods used throughout the apel application.
 */
public class SecurityUtils {
   /**
    * Version of encryption schema used by APEL
    */
    private static final String VERSION="*APEL V.0.2*";

    /**
     * Reference to logging tool.
     */
    private static final Logger LOG = Logger.getLogger(SecurityUtils.class.getName());
   
   /**
    * Encryption Algorithm Used
    */
    private static String ALGORITHM="RSA";

   /** 
    * Public key location 
    */
    private static String m_publicKeyLocation = "";

   /**
    * Cipher
    */
    private static Cipher cipher = null;

   /**
    * KeyFactory
    */
    private static KeyFactory keyFactory = null;

   /**
    * CryptoGraphically strong Pseudo-Random-Number-Generator
    */
    private static SecureRandom secureRandom = null;

   /**
    * BASE64 encoded string of the GOC RSA public key
    * 1024 RSA key is faster, but allows you to encrypt 117 bytes of data
    * (that is strings comprised of 117 ASCII characters)
    * Version 0.1
    */
    private static String publicKeyBase64EncodedString_1024 = 
    "MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCaFdmy11oUYa+Sc9LWIc98S+Ny" +
     "mhytrBtKIGoRcGiD0ERdQAHU9EaNd/I5P/L9kMjl/BRXeuHdxH6JCKhtki0kY1qs" +
     "gLoj+5aHclHpAwpJArE5ie4p+wW3Zwzwijx9OmXWYbMyjLnug8JnVR5VDQK1Ja0A" +
     "rbsX1nkr2pBSC2NH6QIDAQAB";


   /**
    * BASE64 encoded string of the GOC RSA public key (2048)
    * 2048 RSA key is slower, but allows you to encrypt 245 bytes of data
    * (that is strings comprised of 245 ASCII characters)
    * Version 0.2
    */
    private static String publicKeyBase64EncodedString_2048 =
      "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnoWnH3Hf8Rmdq8uunRaQ" + 
      "jSeAEiKubLL5Loe6Bp65gHHrtXtUSxV7S5SQnZnBqPuIZkhrIWix+PzcNyJWiykI" + 
      "IrlYC5WN2scteGPNPSI46RtIwo4RxDSVONy4Lp0feR82ROvcjZV5dOWP18yq73FK" + 
      "WE9U4WVixKzce/Y4FnjggErtIzcL+ZL5psmjbFimTF8y2VpJ3VVWs58kEBHf8hf9" + 
      "8CDwSagJuiV0Ro/o+UwxwBqgjYhCwXHVMF+uFzBvGhKCtVbFILHXV6AuBVyGFuJy" +    
      "OAxROkWdH16xH3X4cWvclDJTagLXAJWU0QK0u5eQJAN008LUIETHHKnMtZi7wrHa" +
      "cQIDAQAB";

   /**
    * PublicKey used for encrypting data
    */
    private static PublicKey publicKey = null;

    private static SecurityUtils SECURITYUTILS = new SecurityUtils();
    
    protected SecurityUtils() {
      Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
   
      try {

           // Set up random number generator
           secureRandom = SecureRandom.getInstance("SHA1PRNG","SUN");
           Log.info(LOG,"SecureRandom provider is " + secureRandom.getProvider() );
           //
           // Seed random number generator.
           // This should work on most UNIX boxes quite well.
           // Windows users will have to wait for SecureRandom to sel seed the first time it is called.
           //
           if (new File("/dev/urandom").exists())
           {
                Log.info(LOG,"Salting SecureRandom (SHA1PRNG) from /dev/urandom");
                byte[] salt = new byte[8192];
                new FileInputStream("/dev/urandom").read(salt);
                secureRandom.setSeed(salt);
                for (int t=0; t<salt.length; t++) salt[t] = 0;
                salt = null;
            }

	    System.out.println("KEY LOCATION = " + m_publicKeyLocation);

            // Derive PublicKey 
    	    if (m_publicKeyLocation.equals("")){
            	publicKey = getPublicKeyFromString(publicKeyBase64EncodedString_1024);
	    }
	    else {
		publicKey = getPublicKeyFromFile(m_publicKeyLocation);
	    }

            // Obtain a Cipher Object
            cipher = Cipher.getInstance("RSA/ECB/PKCS1Padding","BC");

            // Obtain a KeyFactory Object
            keyFactory = KeyFactory.getInstance(ALGORITHM);

        }catch( IOException e ){
            Log.info(LOG, "SecurityUtils IOException:" + e );
        }catch( NoSuchAlgorithmException e ){
            Log.info(LOG, "SecurityUtils NoSuchAlgorithmException:" + e );
        } catch (Exception e) {
            Log.info(LOG, "SecurityUtils Exception:" + e );
	    e.printStackTrace();
        }

    } // end constriuctor
 
   /**
    * Returns a reference to the global utility object.
    */
    public static SecurityUtils getInstance() {
    	return SECURITYUTILS;
    }
    

   /**
    * Generates Public Key from BASE64 encoded string
    * @param key BASE64 encoded string which represents the key
    * @return The PublicKey
    * @throws java.lang.Exception
    */

    private static PublicKey getPublicKeyFromString(String key) throws Exception
    {
        BASE64Decoder b64 = new BASE64Decoder();
        KeyFactory keyFactory = KeyFactory.getInstance(ALGORITHM);
        EncodedKeySpec publicKeySpec = new X509EncodedKeySpec(b64.decodeBuffer(key));
        PublicKey publicKey = keyFactory.generatePublic(publicKeySpec);
        return publicKey;
    }

    /**
     * Generates Public Key from BASE64 encoded string
     * @param key BASE64 encoded string which represents the key
     * @return The PublicKey
     * @throws java.lang.Exception
     */
     private static PublicKey getPublicKeyFromFile(String file) throws Exception
     {

	File keyFile = new File(file);        
	byte[] encodedKey = new byte[(int)keyFile.length()];

	new FileInputStream(keyFile).read(encodedKey);

	X509EncodedKeySpec publicKeySpec = new X509EncodedKeySpec(encodedKey);

	KeyFactory kf = KeyFactory.getInstance("RSA");
	PublicKey pk = kf.generatePublic(publicKeySpec);

	Cipher rsa = Cipher.getInstance("RSA");

	rsa.init(Cipher.ENCRYPT_MODE, pk);
	OutputStream os = new CipherOutputStream(
        new FileOutputStream("encrypted.rsa"), rsa);

	Writer out = new OutputStreamWriter(os);
	out.write("Hello World!!");
	out.close();
	os.close();

	return pk;

     }


   /**
    * Encode bytes array to BASE64 string
    * @param bytes
    * @return Encoded string
    */

    private static String encodeBase64(byte[] bytes)
    {
         Base64 b64 = new Base64();
         return new String(b64.encode(bytes));
    }
   /**
    * Append Random data to the input string
    * @param String
    * @return byte[]
    */
    private static byte[] randomiseData(String strin, SecureRandom secureRandom) throws Exception
    {
       // Check input string for null
       // Calling String methods causes a NullPointerException
       if (strin == null ) {
         strin = "NULL";
       }
       // Cristina del Cano Novales 2008-04-23
       // Bug fix #27375 The case for the UserDN should be preserved
       //String str  = strin.toLowerCase();
       String str = strin;
       String n1   = Double.toString(secureRandom.nextDouble());
       String n2   = Double.toString(secureRandom.nextDouble());
       String data = n1.substring(2) + "*" + str + "*" + n2.substring(2);
       int data_length = data.getBytes("UTF8").length;
       ///////////////////////////////////////////////////////
       // Check if the string to be encrypted is too long?
       // If yes, try to make the input string shorter.
       //  1) remove n2 and test
       //  2) remove n1 and test
       //  3) replace "ou=personal certificate" with "ou=pc" 
       //  4) Take only the email part of the dn
       ///////////////////////////////////////////////////////

       if ( data_length > 110 ) {

           String tmp = n1.substring(2) + "*" + str;
           //Log.info(LOG,"Input string length has exceeded modulus of encryption key:");
           int dn_length = str.getBytes("UTF8").length;
           int n1_length = n1.getBytes("UTF8").length;
           int n2_length = n2.getBytes("UTF8").length;
           //Log.info(LOG,"Lengths: " + "n1=" + n1_length + " n2=" + n2_length + " dn=" + dn_length);
           // reduce length of input string by removing n2 first
           tmp = n1.substring(2) + "*" + str;

           if (tmp.getBytes("UTF8").length < 110 ) {
                   data=tmp;
           }else if (str.getBytes("UTF8").length < 110 ) {
                   data=str;
           }else {
                   // apply methods to shorten the DN
                   tmp=str.replaceAll("ou=personal certificate","ou=pc");
                   if (tmp.getBytes("UTF8").length < 110 ) {
                           // printStats(tmp,"... \"ou=personal certificate\" replaced with \"ou=pc\" ");
                           data=tmp;
                   }else{
                           // try to remove everything but the email component
                           String [] email = str.split("email"); 
                           if (email[1].getBytes("UTF8").length < 110 ) {
                             data="email" + email[1];
                           }else{ 
                             printStats(tmp,"Failed to reduced size of input data!");
                           }
                   }
           }
       }
       return data.getBytes("UTF8");
    }
    /**
     * Print details about the input string.
     */
    public static void printStats(String data, String text) throws Exception {
     Log.info(LOG, "  " + text);
     Log.info(LOG, "Size of input string to encrypt = " + data.getBytes("UTF8").length );
     Log.info(LOG, "User DN : " + data);
    }
    /**
     * Encrypt data, and encode to base 64 representation.
     */
    public String EncryptDataToBase64 (String userDN)
    {
            
	try{
            // Derive PublicKey
            if (m_publicKeyLocation.equals("")){
                publicKey = getPublicKeyFromString(publicKeyBase64EncodedString_1024);
            }
            else {
                publicKey = getPublicKeyFromFile(m_publicKeyLocation);
            }
	}
        catch(Exception ex){
	   Log.info(LOG, "SecurityUtils Exception - Can't load the public key from file : " + ex);
	}


        byte[] plainText = null;
        byte[] ciphertextBytes = null; 
        String output = null;
         try {  
            plainText = randomiseData(userDN,secureRandom);
            // Initialize the cipher for encryption
            cipher.init(Cipher.ENCRYPT_MODE, publicKey, secureRandom );
            // Encrypt
            ciphertextBytes = cipher.doFinal(plainText);
            output = VERSION + encodeBase64(ciphertextBytes);
          } catch (Exception e) {
            Log.info(LOG,"SecurityUtils Exception - Unable to process UserDN : " + e );
            Log.info(LOG,"UserDN=\"" + userDN + "\"");
            output = "UserDN Encryption Failed";
          }
            // Encode to BASE64
            return output;
    }
    /**
     * Decrypt data, encode into UTF-8.
     * Implemented at GOC using private key.
     */
    public String DecryptBase64DataToString (String userDN)
    {
     return "Not Enabled on APEL Client";
    }

    public void setPublicKeyLocation(String publicKeyLocation){
	m_publicKeyLocation = publicKeyLocation;
    }      

}
