/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core;

import org.apache.log4j.Logger;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.GregorianCalendar;
import java.util.SimpleTimeZone;
import java.util.TimeZone;


/**
 * Conversions to ISO 8601 format times.
 */
public class Time {
    
    private static final Logger m_log = Logger.getLogger(Time.class.getName());
    private static final SimpleDateFormat LOCAL_DATE_TIME_STAMP_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss Z");
//    private static String timeZoneSpecification = null;
    static final long ONE_DAY_IN_MILLIS = 24 * 60 * 60 * 1000;
    static final TimeZone UTC_TIMEZONE = TimeZone.getTimeZone("GMT");
    GregorianCalendar m_calendar;
    SimpleDateFormat m_dateFormat;
    private String m_epochTimeSeconds;
    private TimeZone m_timeZone;

    /**
     * Constructor
     *
     * @param epochTimeSeconds as an epoch time in seconds - i.e. UTC since
     *        midnight at start of 1 January
     * @param timeZone the timezone to use
     */
    public Time(String epochTimeSeconds, TimeZone timeZone) {
        m_epochTimeSeconds = epochTimeSeconds;
        m_timeZone = timeZone;
        m_dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        m_calendar = new GregorianCalendar(m_timeZone);
    }

    /**
     * Creates a new time object.
     *
     * @param epochTimeSeconds as an epoch time in seconds - i.e. UTC since
     * midnight at start of 1 January 1970.
     * @param date representation of date in local time e.g. 2004-11-23
     * @param time representation of time in local time e.g. 17:58:18
     */
    public Time(String epochTimeSeconds, String date, String time, 
        String timezone) {
        m_epochTimeSeconds = epochTimeSeconds;
        if (timezone == null) {
            timezone = "default";
        }
        if (timezone.equalsIgnoreCase("derived")) {
            // Convert date/time to seconds as if it were in UTC
            String dateTime = date + " " + time + " GMT";
            try {
                Date localeDate = LOCAL_DATE_TIME_STAMP_FORMAT.parse(dateTime);
                long localeTimeMillis = localeDate.getTime();
                long utcTimeMillis = Long.valueOf(m_epochTimeSeconds).longValue() * 1000;
                long rawOffset = localeTimeMillis - utcTimeMillis;
                if ((rawOffset % 60000) != 0) {
                    Log.warn(m_log, "Timezone offset is not an integer numbers of minutes " + dateTime);
                    // Round to nearest minute
                    rawOffset = (rawOffset + 30000) / 60000;
                    rawOffset = rawOffset * 60000;
                }
                if ((rawOffset > ONE_DAY_IN_MILLIS) || (rawOffset < -ONE_DAY_IN_MILLIS)) {
                    Log.warn(m_log, "Timezone offset " + rawOffset + " milliseconds is more than a day, " +
                        dateTime);
                }
                m_timeZone = new SimpleTimeZone((int) rawOffset, "DERIVED");
            } catch (java.text.ParseException e) {
                Log.warn(m_log, "Exception parsing date/time " + dateTime + ", using GMT timezone");
                m_timeZone = TimeZone.getTimeZone("GMT");
            }
        } else if (!timezone.equalsIgnoreCase("default")) {
            // This is what happens if no explicit timezone is specified
            // Use current locale
            m_timeZone = TimeZone.getDefault();
        } else {
            // Use time zone specified in parameters
            m_timeZone = TimeZone.getTimeZone(timezone);
        }

        m_dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        m_calendar = new GregorianCalendar(m_timeZone);
    }

    /**
     * Return the epoch time in seconds.
     *
     * @return epoch time in seconds - i.e. since midnight at start of 1
     *         January
     */
    public String getEpoch() {
        return m_epochTimeSeconds;
    }

    /**
     * Return the ISO 8601 format time
     *
     * @return ISO 8601 format time
     */
    public String getIso8601() {
        m_calendar.setTimeZone(m_timeZone);
        m_calendar.setTimeInMillis(new Long(m_epochTimeSeconds).longValue() * 1000);

        return iso8601();
    }

    /**
     * Return the ISO 8601 format time as if the current timezone was UTC.
     *
     * @return ISO 8601 format time as if the locale was UTC
     */
    public String getIso8601Utc() {
        m_calendar.setTimeZone(UTC_TIMEZONE);
        m_calendar.setTimeInMillis(new Long(m_epochTimeSeconds).longValue() * 1000);

        return iso8601();
    }

    /**
     * Return then timezone in use.
     *
     * @return timeZone
     */
    public TimeZone getTimeZone() {
        return m_timeZone;
    }

    /**
     * Obtain ISO 8601 format time from calendar
     *
     * @return ISO 8601 format time
     */
    private String iso8601() {
        int dst = 0;   
        // Check if time zone is within daylight saving time
        if (m_calendar.getTimeZone().inDaylightTime(m_calendar.getTime())) {
            dst = 1;
        }
        m_dateFormat.setTimeZone(m_calendar.getTimeZone());
        StringBuffer isoTime = 
            new StringBuffer(m_dateFormat.format(m_calendar.getTime()));
        // Get offset in minutes
        int offset = (m_calendar.get(java.util.Calendar.ZONE_OFFSET) +
            dst * m_calendar.getTimeZone().getDSTSavings())/ 60000;
        if (offset == 0) {
            isoTime.append("Z");
        } else {
            NumberFormat numberFormat = new DecimalFormat("00");

            // Add in sign
            isoTime.append((offset < 0) ? "-" : "+");

            // Add in hours
            isoTime.append(numberFormat.format(Math.abs(offset / 60)));

            // Add in minutes
            int offsetMinutes = Math.abs(offset % 60);

            if (offsetMinutes > 0) {
                isoTime.append(":").append(numberFormat.format(offsetMinutes));
            }
        }

        return isoTime.toString();
    }
}
