/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Vector;
import org.glite.apel.core.Log;
import org.apache.log4j.Logger;



/**
 * Contains a list of misc methods used throughout the apel application.
 */
public class Utils {

    /**
     * Reference to logging tool.
     */
    private static final Logger LOG = Logger.getLogger(Utils.class.getName());

    /**
     * Constants used to represent the message log file type.
     */
    private static final int MSG_LOGS = 0;
    
    /**
     * Constant used to represent the gk log file type.
     */
    private static final int GK_LOGS = 1;
    
    /**
     * Constant used to represent pbs event logs.
     */
    private static final int PBS_EVENT_LOGS = 2;
    
    /**
     * Constant used to represent lsf event logs.
     */
    private static final int LSF_EVENT_LOGS = 3;

    /**
     * Constant used to represent condor event logs.
     */
    private static final int CONDOR_EVENT_LOGS = 4;

    /**
     * Constant used to represent sge event logs.
     */
    private static final int SGE_EVENT_LOGS = 5;

    /**
     * Constant used to represent blahd logs.
     */
    private static final int BLAHD_LOGS = 6;


    private static Utils UTILS = new Utils();
    
    protected Utils() {}
    
    /**
     * Returns a reference to the global utility object.
     */
    public static Utils getInstance() {
    	return UTILS;
    }
    
    
    /**
     * Returns a list of all message log file names within the 
     * directory and sub directories if the recursive flag is 
     * set to true.
     *
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getMsgLogFiles(String dir, boolean recursive) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, MSG_LOGS);
        }
        return files;
    }
    
    /**
     * Returns a list of all gk log file names within the 
     * directory and sub directories if the recursive flag is 
     * set to true.
     *
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getGkLogFiles(String dir, boolean recursive) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, GK_LOGS);
        }
        return files;
    }

    /**
     * Returns a list of all blahd log file names within the
     * directory and sub directories if the recursive flag is
     * set to true.
     *
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getBlahdLogFiles(String dir, boolean recursive) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, BLAHD_LOGS);
        }
        return files;
    }

   /**
     * Returns a list of all blahd log file names within the
     * directory and sub directories if the recursive flag is
     * set to true.
     * Method is overloaded
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getBlahdLogFiles(String dir, boolean recursive, String logPrefix) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, BLAHD_LOGS, logPrefix);
        }
        return files;
    }



    /**
     * Returns a list of all pbs event log file names within the 
     * directory and sub directories if the recursive flag is 
     * set to true.
     *
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getPbsEventLogFiles(String dir, boolean recursive) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, PBS_EVENT_LOGS);
        }
        return files;
    }

    /**
     * Returns a list of all LSF event log file names within the 
     * directory and sub directories if the recursive flag is 
     * set to true.
     *
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getLsfEventLogFiles(String dir, boolean recursive) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, LSF_EVENT_LOGS);
        }
        return files;
    }


    /**
     * Returns a list of all SGE event log file names within the
     * directory and sub directories if the recursive flag is
     * set to true.
     *
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getSgeEventLogFiles(String dir, boolean recursive) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, SGE_EVENT_LOGS);
        }
        return files;
    }

    /**
     * Returns a list of all condor event log file names within the
     * directory and sub directories if the recursive flag is
     * set to true.
     *
     * @param dir The directory to add files from
     * @param recursive If <code>true</code> searches subdirectories as well.
     */
    public Vector getCondorEventLogFiles(String dir, boolean recursive) {
        Vector files = new Vector();
        if (dir != null) {
            findLogFiles(new File(dir), recursive, files, CONDOR_EVENT_LOGS);
        }
        return files;
    }
   
 
    /**
     * Sifts through a directory (and any sub dirs if the recursive flag 
     * is set) and adds any matching valid log files to the current list.
     * 
     * @param dir The current directory to begin the search.
     * @param recursive If <code>true</code> searches subdirectories as well.
     * @param files A list of current valid files found during the search.
     * @param fileType The type of files to search for (ie MSG_LOGS etc).
     */
    protected void findLogFiles(File dir, boolean recursive, Vector files, int fileType) {
        if (dir != null) {
            File[] directoryList = dir.listFiles();
            if (directoryList == null) {
                return;
            }
            for (int i = 0; i < directoryList.length; ++i) {
                File current = directoryList[i];
                if (current.isFile() && isValid(current, fileType)) {
                    try {
                        files.add(current.getCanonicalPath());
                    } catch (IOException ioe) {
                    	System.err.println(ioe.getMessage());
                    }
                } else if (current.isDirectory() && recursive) {
                    findLogFiles(current, recursive, files, fileType);
                }
            }
        }
    }
    


   /**
     * Sifts through a directory (and any sub dirs if the recursive flag
     * is set) and adds any matching valid log files to the current list.
     * Method is overloaded
     * @param dir The current directory to begin the search.
     * @param recursive If <code>true</code> searches subdirectories as well.
     * @param files A list of current valid files found during the search.
     * @param fileType The type of files to search for (ie MSG_LOGS etc).
     */
    protected void findLogFiles(File dir, boolean recursive, Vector files, int fileType, String logPrefix) {
        if (dir != null) {
            File[] directoryList = dir.listFiles();
            if (directoryList == null) {
                return;
            }
            for (int i = 0; i < directoryList.length; ++i) {
                File current = directoryList[i];
                if (current.isFile() && isValid(current, fileType, logPrefix)) {
                    try {
                        files.add(current.getCanonicalPath());
                    } catch (IOException ioe) {
                        System.err.println(ioe.getMessage());
                    }
                } else if (current.isDirectory() && recursive) {
                    findLogFiles(current, recursive, files, fileType, logPrefix);
                }
            }
        }
    }


    /**
     * Returns true if the log file matches a PBS/Condor/SGE/LSF or globus-gatekeeper or message type file. 
     * Example valid file names are: messages, messages.3, messages.3.gz
     */
    protected boolean isValid(File file, int fileType) {
        String name = file.getName();
        if (fileType == MSG_LOGS) {
            return name.matches("messages") ||
		   name.matches("messages\\.\\d+") || 
                   name.matches("messages\\.\\d+\\.gz") || 
                   name.matches("messages\\.\\d");
        } else if (fileType == GK_LOGS) {
            return name.matches("globus-gatekeeper\\.log$") || 
                   name.matches("^globus-gatekeeper\\.log\\.\\d+\\.\\d+$") || 
                   name.matches("^globus-gatekeeper\\.log\\.\\d+\\.\\d+\\.gz$");
         } else if (fileType == BLAHD_LOGS) {
            return name.matches("acct\\_blahp\\.log$") ||
                   name.matches("^acct\\_blahp\\.log-\\d{8}$") ||
                   name.matches("^acct\\_blahp\\.log-\\d{8}\\.gz$");
        } else if (fileType == PBS_EVENT_LOGS) {
            return name.matches("^\\d\\d\\d\\d\\d\\d\\d\\d$") || 
                   name.matches("^\\d\\d\\d\\d\\d\\d\\d\\d\\.gz$");
        } else if (fileType == LSF_EVENT_LOGS) {
            return name.matches("^lsb\\.acct\\.\\d{8}$") || 
		   name.matches("^lsb\\.acct\\.\\d+") ||
                   name.matches("^lsb\\.acct\\.\\d{8}\\.gz$") ||
                   name.matches("^lsb\\.acct$");
        } else if (fileType == CONDOR_EVENT_LOGS) {
            return name.matches("^history$") || 
                   name.matches("^history\\.gz$") ||
                   name.matches("^history\\.\\d+") ||
                   name.matches("^history\\.\\d+\\.gz") ||
                   name.matches("^history.\\d{8}T\\d{6}.gz") ||
                   name.matches("^history.\\d{8}T\\d{6}");
        } else {
            return name.matches("^accounting$") ||
                   name.matches("^accounting\\.\\d{8}$") ||
                   name.matches("^accounting\\.\\d{8}\\.gz$") || 
                   name.matches("^accounting\\.gz$") ||
		   name.matches("^accounting-\\d{8}$") ||
		   name.matches("^accounting-\\d{8}\\.gz$");
        }
    }

   /**
     * Returns true if the log file matches Accounting Log with prefix.
     */
    protected boolean isValid(File file, int fileType, String logPrefix) {
        // Name of file in Log dir to be tested
        String name = file.getName();
        if (fileType == MSG_LOGS) {
            return name.matches("messages") ||
                   name.matches("messages\\.\\d+") ||
                   name.matches("messages\\.\\d+\\.gz") ||
                   name.matches("messages\\.\\d");
        } else if (fileType == GK_LOGS) {
            return name.matches("globus-gatekeeper\\.log$") ||
                   name.matches("^globus-gatekeeper\\.log\\.\\d+\\.\\d+$") ||
                   name.matches("^globus-gatekeeper\\.log\\.\\d+\\.\\d+\\.gz$");
         } else if (fileType == BLAHD_LOGS) {
            String tmp1="^" + logPrefix + "$";
            String tmp2="^" + logPrefix + "\\d{8}$";
            String tmp3="^" + logPrefix + "\\d{8}\\.gz$";
            //Log.info(LOG, "++++ test expr tmp1 " + name.matches(tmp1) );
            //Log.info(LOG, "++++ test expr tmp2 " + name.matches(tmp2) );
            //Log.info(LOG, "++++ test expr tmp3 " + name.matches(tmp3) );
            return name.matches(tmp1) || name.matches(tmp2) || name.matches(tmp3);
        } else if (fileType == PBS_EVENT_LOGS) {
            return name.matches("^\\d\\d\\d\\d\\d\\d\\d\\d$") ||
                   name.matches("^\\d\\d\\d\\d\\d\\d\\d\\d\\.gz$");
        } else if (fileType == LSF_EVENT_LOGS) {
            return name.matches("^lsb\\.acct\\.\\d{8}$") ||
                   name.matches("^lsb\\.acct\\.\\d+") ||
                   name.matches("^lsb\\.acct\\.\\d{8}\\.gz$") ||
                   name.matches("^lsb\\.acct$");
        } else if (fileType == CONDOR_EVENT_LOGS) {
            return name.matches("^history$") ||
                   name.matches("^history\\.gz$") ||
                   name.matches("^history\\.\\d+") ||
                   name.matches("^history\\.\\d+\\.gz") ||
                   name.matches("^history.\\d{8}T\\d{6}.gz") ||
                   name.matches("^history.\\d{8}T\\d{6}");
        } else {
            return name.matches("^accounting$") ||
                   name.matches("^accounting\\.\\d{8}$") ||
                   name.matches("^accounting\\.\\d{8}\\.gz$") ||
                   name.matches("^accounting\\.gz$") ||
                   name.matches("^accounting-\\d{8}$") ||
	           name.matches("^accounting-\\d{8}\\.gz$");

        }
    }

    /**
     * Return date string with addition or subtraction of specified number of
     * days
     *
     * @param date date string to adjust
     * @param dayChange the number of days to add or substract
     *
     * @return date string adjusted by the specified number of days.
     */
    private String addDays(String date, int dayChange) {
        String[] fields = date.split("-");
        String day = fields[2];
        String month = fields[1];
        String year = fields[0];
        String DATE_FORMAT = "yyyy-MM-dd";
        SimpleDateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
        Calendar cal = Calendar.getInstance();
        // Subtracting one from month because it is zero-based in Calendar.set
        cal.set(new Integer(year).intValue(),
            new Integer(month).intValue() - 1, new Integer(day).intValue());
        cal.add(Calendar.DATE, dayChange);
        return dateFormat.format(cal.getTime());
    }
    
    /**
     * Calculate the earliest date a PBS record may have in order to be matched
     * with a GkRecord. The job number rolls over every few months back to
     * zero again. This potentially means the the GRAM script job ID may be
     * repeated  after a few months.  When this happens there is a risk that
     * the job number will be matched with an old one in the GateKeeper log
     * file which has not been marked as processed. This will result in the
     * local user ID being mapped to the wrong global user DN.  Avoid this by
     * having a start date on the Gate Keeper records which means that they
     * cannot be used to match up Pbs records older than that date.  It only
     * needs to be a day to handle the crossover at midnight. In a perfect
     * world the measurement date would be good enough.  But allow an extra
     * day because times don't always exactly match up.
     *
     * @param measurementDate measurement date of record
     *
     * @return start date for record
     */
    public String calculateValidFrom(String measurementDate) {
        return addDays(measurementDate, -1);
    }

    /**
     * Calculate the most recent date a PBS record may have in order to be
     * matched with a GkRecord. The job number rolls over every few months
     * back to zero again. This potentially means the the GRAM script job ID
     * may be repeated  after a few months.  When this happens there is a risk
     * that the job number will be matched with an old one in the GateKeeper
     * log file which has not been marked as processed. This will result in
     * the local user ID being mapped to the wrong global user DN.  Avoid this
     * by having an expiry date on the Gate Keeper records which means that
     * they cannot be used to match up Pbs records newer than that date.  It
     * only needs to be as long as the time for the longest job to finish, and
     * shorter than the roll-over interval for job numbers.
     *
     * @param measurementDate measurement date of record
     *
     * @return expiry date for record
     */
    public String calculateValidUntil(String measurementDate) {
        return addDays(measurementDate, 28);
    }    
    
}
