/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.config.map;

import org.glite.apel.core.config.parser.ConfigMap;

/**
 * Maps tokens found in the XML config file (used by the apel log processor) to 
 * internal ids that can then be manipulated by the application.
 */
public class FileMap extends ConfigMap {

    /** 
     * XML root element of the XML configruation file. 
     */
    private static final String ROOT_TOK = "/ApelConfiguration";

    /** 
     * XML element representing the Dir element. 
     */
    private static final String DIR_TOK = "/Dir";
    
    /** 
     * XML element representing the Logs element. 
     */
    private static final String LOGS_TOK = "/Logs";
    
    /** 
     * XML element defining database URL. 
     */
    private static final String DBURL_TOK = "/DBURL";

    /** 
     * XML element defining database username. 
     */
    private static final String DBUSER_TOK = "/DBUsername";

    /** 
     * XML element defining database password. 
     */
    private static final String DBPASS_TOK = "/DBPassword";

    /** 
     * XML element defining the site name. 
     */
    private static final String SITENAME_TOK = "/SiteName";
    
    /** 
     * XML element defining whether to process GK logs. 
     */
    private static final String GK_PROC_TOK = "/GKLogProcessor";
    private static final String BLAHD_PROC_TOK = "/BlahdLogProcessor";
    
    /** 
     * XML element defining GK log files. 
     */
    private static final String GKLOGS_TOK = "/GKLogs";

    /** 
     * XML element defining the SubmitHost value. 
     */
    private static final String GK_SUBMIT_TOK    = "/GKLogProcessor/SubmitHost";
    private static final String BLAHD_SUBMIT_TOK = "/BlahdLogProcessor/SubmitHost";
    private static final String EVENT_SUBMIT_TOK = "/EventLogProcessor/SubmitHost";   

    /**
     * XML element defining the Blahpd_accounting_info prefix
     */
    private static final String BLAHD_PREFIX_TOK    = "/BlahdLogProcessor/BlahdLogPrefix";


    /** 
     * XML element defining GK message log files. 
     */
    private static final String MSLOGS_TOK = "/MessageLogs";
    
    /** 
     * XML element defining whether to search subdirs. 
     */
    private static final String GK_SEARCH_SUBDIRS_TOK = "/GKLogProcessor/Logs@searchSubDirs";
    private static final String BLAHD_SEARCH_SUBDIRS_TOK = "/BlahdLogProcessor/Logs@searchSubDirs";

    /** 
     * XML element defining whether to reprocess records. 
     */
    private static final String GK_REPROCESS_TOK = "/GKLogProcessor/Logs@reprocess";
    private static final String BLAHD_REPROCESS_TOK = "/BlahdLogProcessor/Logs@reprocess";    

    /** 
     * XML element defining extra GK log files. 
     */
    private static final String EXTRA_GK_LOGS_TOK = "/GKLogProcessor/Logs/GKLogs/ExtraFile";
    private static final String EXTRA_BLAHD_LOGS_TOK = "/BlahdLogProcessor/Logs/ExtraFile";

    /** 
     * XML element defining extra Message log files. 
     */
    private static final String EXTRA_MSG_FILE_TOK = "/GKLogProcessor/Logs/MessageLogs/ExtraFile";
    
    /** 
     * XML element defining whether to process event logs. 
     */
    private static final String EVENT_PROC_TOK = "/EventLogProcessor";

    /** 
     * XML element defining the extra log file. 
     */
    private static final String EXTRA_EVENT_LOGS_TOK = "/EventLogProcessor/Logs/ExtraFile";

    /** 
     * XML element defining the timezone where the event logs are processed. 
     */
    private static final String TIMEZONE_TOK = "/EventLogProcessor/Timezone";
    
    /** 
     * XML element defining whether to search subdirs. 
     */
    private static final String EVENT_SEARCH_SUBDIRS_TOK = "/EventLogProcessor/Logs@searchSubDirs";
    
    /** 
     * XML element defining whether to reprocess records. 
     */
    private static final String EVENT_REPROCESS_TOK = "/EventLogProcessor/Logs@reprocess";
    
    /**
     * XML element defining whether to use the scaling factor (SpecInt)
     * quoted in the LSF log file.  This only has an effect in the 
     * LSF parser.
     */
    private static final String EVENT_USE_SCALING_TOK = "/EventLogProcessor/Logs@useScalingFactor";
    
    /** 
     * XML element defining whether to process CPU info. 
     */
    private static final String CPU_PROC_TOK = "/CPUProcessor";

    /** 
     * XML element defining the GIIS hostname. 
     */
    private static final String GIIS_TOK = "/CPUProcessor/GIIS";

    /** 
     * XML element defining the default CPU spec. 
     */
    private static final String DEFAULTSPEC_TOK = "/CPUProcessor/DefaultCPUSpec";
    
    /** 
     * XML element defining the delete processor. 
     */
    private static final String DELETE_PROC_TOK = "/DBDeleteProcessor";
    
    /**
     * XML element defining the delete processor.
     */
    private static final String DB_PROC_TOK = "/DBProcessor";

    /** 
     * XML element defining the delete processor. 
     */
    private static final String CLEAN_ALL_TOK = "/DBDeleteProcessor@cleanAll";
    
    /**
     * XML element defining the delete processor.
     */
    private static final String INSPECT_TABLE_TOK = "/DBProcessor@inspectTables";

    /** 
     * XML element defining whether to enable debug logging. 
     */
    private static final String DEBUG_TOK = "@enableDebugLogging";
    
    /** 
     * Internal id - event log directory. 
     */
    public static final String PROP_EVENT_LOG_DIR = "pbslog";
    
    /** 
     * Internal id - gk log directory. 
     */
    public static final String PROP_GKLOG = "gklog";
    public static final String PROP_BLAHDLOG = "blahdlog";
  
   /**
    * Internal id - blahd log prefix
    */
    public static final String PROP_BLAHDLOG_PREFIX = "blahdlogPrefix";
  

    /** 
     * Internal id - message log directory. 
     */
    public static final String PROP_MSLOG = "mslog";
    
    /** 
     * Internal id - GIIS host. 
     */
    public static final String PROP_GIIS = "GIIS";
    
    /** 
     * Internal id - site name. 
     */
    public static final String PROP_SITE = "site";
    
    /** 
     * Internal id - gk submit host name. 
     */
    public static final String PROP_SUBMIT_HOST = "CEhost";
    
    /** 
     * Internal id - time zone. 
     */
    public static final String PROP_TIME_ZONE = "timeZone";
    
    /**
     * Internal id - denotes a default value for the CPU spec should be used.
     */
    public static final String PROP_USE_DEFAULT_SPEC = "noLdap";
    
    /** 
     * Internal id - database URL. 
     */
    public static final String PROP_DB_URL = "DBURL";
    
    /** 
     * Internal id - database user name. 
     */
    public static final String PROP_DB_USERNAME = "DBUsername";

    /** 
     * Internal id - database password. 
     */
    public static final String PROP_DB_PASSWORD = "DBPassword";
    
    /** 
     * Internal id - extra event log files. 
     */
    public static final String PROP_EXTRA_EVENT_LOGS = "extraEventLogFile";

    /** 
     * Internal id - extra gk log files. 
     */
    public static final String PROP_EXTRA_GK_LOGS = "extraGKLogFile";
    public static final String PROP_EXTRA_BLAHD_LOGS = "extraBlahdLogFile";
    
    /** 
     * Internal id - extra message log files. 
     */
    public static final String PROP_EXTRA_MSG_LOG_FILES = "extraMSLogFile";
    
    /**
     * Internal id - option set when using the DeleteLogProcessor.
     */
    public static final String OPT_CLEAN_ALL_FROM_TABLES = "cleanAll";

    /**
     * Internal id - option set when using the DBProcessor.
     */
    public static final String OPT_INSPECT_TABLES = "inspectTables";
    
    /**
     * Internal id - option to reset the process flag on all previously
     * processed gk logs. 
     */
    public static final String OPT_GK_LOG_REPROCESS = "reprocessGKFiles";
    public static final String OPT_BLAHD_LOG_REPROCESS = "reprocessBlahdFiles";

    /**
     * Internal id - option search all sub directories under the
     * gk log directory.
     */
    public static final String OPT_GK_SEARCH_ALL = "recursiveGKFiles";
    public static final String OPT_BLAHD_SEARCH_ALL = "recursiveBlahdFiles";
    
    /**
     * Internal id - option to reset the process flag on all previously
     * prcessed event logs. 
     */
    public static final String OPT_EVENT_LOG_REPROCESS = "reprocessEventFiles";
    
    /**
     * Internal id - whether to use scaling factor from the LSF log file.
     */
    public static final String OPT_EVENT_USE_SCALING_FACTOR = "useScalingFactor";
    

    /**
     * Internal id - option to search all sub directories under the event
     * log directory.
     */
    public static final String OPT_EVENT_SEARCH_ALL = "recursiveEventFiles";
    
    /** 
     * Internal id - option to run the CPU performance of the CE. 
     */
    public static final String OPT_SPEC_PROCESSOR = "c";

    /** 
     * Internal id - option to run gk log processor. 
     */
    public static final String OPT_GK_LOG_PROCESSOR = "g";
    public static final String OPT_BLAHD_LOG_PROCESSOR = "b";

    /** 
     * Internal id - option to run event log processor. 
     */
    public static final String OPT_EVENT_LOG_PROCESSOR = "e";
    
    /** 
     * Internal id - option to run the delete processor. 
     */
    public static final String OPT_DB_DELETE_PROCESSOR = "d";
    
    /**
     * Internal id - option to run the database processor.
     */
    public static final String OPT_DB_PROCESSOR = "p";

    /** 
     * Internal id - option to run in quiet mode for minimal debug logging. 
     */
    public static final String OPT_QUIET = "quiet";    
    
	public FileMap() {
		setupOptions();
		setupParameters();
		setupTokens();
    }
	
	/**
	 * Adds the valid tokens that make up the config syntax. If an XML
	 * element is parsed and does not exist within the list of known tokens,
	 * an error will be flagged.
	 */
	private void setupTokens() {
		addTok(ROOT_TOK);
		addTok(ROOT_TOK + DBURL_TOK);
		addTok(ROOT_TOK + DBUSER_TOK);
		addTok(ROOT_TOK + DBPASS_TOK);
		addTok(ROOT_TOK + SITENAME_TOK);

		addTok(ROOT_TOK + GK_SUBMIT_TOK);
        addTok(ROOT_TOK + BLAHD_SUBMIT_TOK);  
        addTok(ROOT_TOK + EVENT_SUBMIT_TOK);  
        addTok(ROOT_TOK + BLAHD_PREFIX_TOK);  
		addTok(ROOT_TOK + CPU_PROC_TOK);
		addTok(ROOT_TOK + GIIS_TOK);
		addTok(ROOT_TOK + CLEAN_ALL_TOK);
        addTok(ROOT_TOK + INSPECT_TABLE_TOK);
		addTok(ROOT_TOK + GK_PROC_TOK);
        addTok(ROOT_TOK + BLAHD_PROC_TOK);    
		addTok(ROOT_TOK + EVENT_PROC_TOK);
		addTok(ROOT_TOK + DEFAULTSPEC_TOK);
		addTok(ROOT_TOK + DELETE_PROC_TOK);
        addTok(ROOT_TOK + DB_PROC_TOK);
        
		addTok(ROOT_TOK + TIMEZONE_TOK);
		addTok(ROOT_TOK + EXTRA_EVENT_LOGS_TOK);
		addTok(ROOT_TOK + EVENT_PROC_TOK + LOGS_TOK);
		addTok(ROOT_TOK + EVENT_PROC_TOK + LOGS_TOK + DIR_TOK);
		addTok(ROOT_TOK + EXTRA_GK_LOGS_TOK);
        addTok(ROOT_TOK + EXTRA_BLAHD_LOGS_TOK); 
		addTok(ROOT_TOK + EXTRA_MSG_FILE_TOK);

		addTok(ROOT_TOK + GK_PROC_TOK + LOGS_TOK);
		addTok(ROOT_TOK + GK_PROC_TOK + LOGS_TOK + GKLOGS_TOK);
		addTok(ROOT_TOK + GK_PROC_TOK + LOGS_TOK + GKLOGS_TOK + DIR_TOK);
		addTok(ROOT_TOK + GK_PROC_TOK + LOGS_TOK + MSLOGS_TOK); 
		addTok(ROOT_TOK + GK_PROC_TOK + LOGS_TOK + MSLOGS_TOK + DIR_TOK);

                // Element /ApelConfiguration/BlahdProcessor/Logs
                addTok(ROOT_TOK + BLAHD_PROC_TOK + LOGS_TOK);
                // Element /ApelConfiguration/BlahdProcessor/Logs/Dir
                addTok(ROOT_TOK + BLAHD_PROC_TOK + LOGS_TOK + DIR_TOK);
                // Element ApelConfiguration/BlahdProcessor/Logs   
                //addTok(ROOT_TOK + BLAHD_PROC_TOK + BLAHD_SUBMIT_TOK + BLAHD_PREFIX_TOK + LOGS_TOK + DIR_TOK);
	}
	
	/**
	 * Adds all the user options to internal id mappings.
	 */
	private void setupOptions() {
		addOpt(ROOT_TOK + GK_PROC_TOK, OPT_GK_LOG_PROCESSOR);
                addOpt(ROOT_TOK + BLAHD_PROC_TOK, OPT_BLAHD_LOG_PROCESSOR); 
		addOpt(ROOT_TOK + EVENT_PROC_TOK, OPT_EVENT_LOG_PROCESSOR);

		addOpt(ROOT_TOK + CPU_PROC_TOK, OPT_SPEC_PROCESSOR);
		addOpt(ROOT_TOK + DEBUG_TOK, OPT_QUIET);

		addOpt(ROOT_TOK + GK_SEARCH_SUBDIRS_TOK, OPT_GK_SEARCH_ALL);
                addOpt(ROOT_TOK + BLAHD_SEARCH_SUBDIRS_TOK, OPT_BLAHD_SEARCH_ALL);

		addOpt(ROOT_TOK + GK_REPROCESS_TOK, OPT_GK_LOG_REPROCESS);
                addOpt(ROOT_TOK + BLAHD_REPROCESS_TOK, OPT_BLAHD_LOG_REPROCESS); 

		addOpt(ROOT_TOK + EVENT_SEARCH_SUBDIRS_TOK, OPT_EVENT_SEARCH_ALL);
		addOpt(ROOT_TOK + EVENT_REPROCESS_TOK, OPT_EVENT_LOG_REPROCESS);
		
		addOpt(ROOT_TOK + EVENT_USE_SCALING_TOK, OPT_EVENT_USE_SCALING_FACTOR);

                addOpt(ROOT_TOK + DELETE_PROC_TOK, OPT_DB_DELETE_PROCESSOR);
                addOpt(ROOT_TOK + CLEAN_ALL_TOK, OPT_CLEAN_ALL_FROM_TABLES);

                addOpt(ROOT_TOK + DB_PROC_TOK, OPT_DB_PROCESSOR);
                addOpt(ROOT_TOK + INSPECT_TABLE_TOK, OPT_INSPECT_TABLES);
	}
	
	/**
	 * Adds all the user parameters to internal id mappings.
	 */
	private void setupParameters() {
		addParam(ROOT_TOK + DBURL_TOK, PROP_DB_URL);
		addParam(ROOT_TOK + DBUSER_TOK, PROP_DB_USERNAME);
		addParam(ROOT_TOK + DBPASS_TOK, PROP_DB_PASSWORD);
		addParam(ROOT_TOK + SITENAME_TOK, PROP_SITE);

		addParam(ROOT_TOK + EXTRA_EVENT_LOGS_TOK, PROP_EXTRA_EVENT_LOGS);
		addParam(ROOT_TOK + EXTRA_GK_LOGS_TOK, PROP_EXTRA_GK_LOGS);
                addParam(ROOT_TOK + EXTRA_BLAHD_LOGS_TOK, PROP_EXTRA_BLAHD_LOGS); 
		addParam(ROOT_TOK + EXTRA_MSG_FILE_TOK, PROP_EXTRA_MSG_LOG_FILES);

		addParam(ROOT_TOK + GIIS_TOK, PROP_GIIS);
		addParam(ROOT_TOK + DEFAULTSPEC_TOK, PROP_USE_DEFAULT_SPEC);
		addParam(ROOT_TOK + TIMEZONE_TOK, PROP_TIME_ZONE);
                
                // EventLog
                addParam(ROOT_TOK + EVENT_SUBMIT_TOK, PROP_SUBMIT_HOST);
                addParam(ROOT_TOK + EVENT_PROC_TOK + LOGS_TOK + DIR_TOK, PROP_EVENT_LOG_DIR);
                // GK
                addParam(ROOT_TOK + GK_SUBMIT_TOK, PROP_SUBMIT_HOST);
                addParam(ROOT_TOK + GK_PROC_TOK + LOGS_TOK + MSLOGS_TOK + DIR_TOK, PROP_MSLOG);
                addParam(ROOT_TOK + GK_PROC_TOK + LOGS_TOK + GKLOGS_TOK + DIR_TOK, PROP_GKLOG);
                // Blahd
                addParam(ROOT_TOK + BLAHD_PREFIX_TOK, PROP_BLAHDLOG_PREFIX); 
                addParam(ROOT_TOK + BLAHD_SUBMIT_TOK, PROP_SUBMIT_HOST); 
                addParam(ROOT_TOK + BLAHD_PROC_TOK + LOGS_TOK + DIR_TOK, PROP_BLAHDLOG);

	}
	
}
