/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core.config.parser;

import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * Reads-in a command line list and attempts to parse the options/paramters 
 * using a gnu/linux CL style. Properties are key-value pairs represented by:
 * --param [value] or -param [value]. Options are simply switches defined as:
 * --option or -o. 
 * 
 * If a property is found, the following argument is taken to be its value.
 * Multi-valued properties are also supported by the Config Class. A Config
 * object is returned on a valid parse.
 * 
 * Arguments consiting of a single '-' character followed by a string 
 * (e.g. -abcd) are assumed to be single character switches (i.e. equivalent 
 * to -a -b -c -d). These are handled by preprocessing the raw arguments.
 */
public class CLParser implements ConfigParser {
	
    /**
     * Using log4j logging framework.
     */
    private static final Logger LOG = 
    	Logger.getLogger(CLParser.class.getName());
    
    /** 
     * The list of command line arguments. 
     */
    private String[] m_args;

    /**
     * Creates a new CL parser using GNU/linux like conventions for parsing
     * the user input.
     *
     * @param args The command line arguments to parse.
     */
    public CLParser(String[] args) {
        m_args = preprocessArgs(args);
    }
	
	/**
     * Extracts the configuration information from the given command line
     * arguments.
     *
     * @returns a configuration object that wraps all the options and
     * properties parsed on the command line.
     * @throws ConfigurationException If an unknown argument is supplied or a
     * property is specified with no value.
     */
    public Config parse(ConfigMap model) throws ConfigurationException {
    	Config config = new Config();
    	LOG.debug("number of arguments defined: " + m_args.length);
    	for (int i = 0; i < m_args.length; i++) {
        	String arg = m_args[i];
            LOG.debug("examining argument: " + arg);
        	if (arg.equals("")) {
                continue;
            }
            String key = trim(m_args[i]);
            if (key == null) {
                throw new ConfigurationException(ErrorMsg.BAD_CLI_ARG, arg);
            }
            String configId = model.intern(key);
            if (model.isProperty(key)) {
            	LOG.debug("found a property");
                i++;
                if (i == m_args.length) {
                    throw new ConfigurationException(ErrorMsg.MISSING_PROP, arg);
                }
                config.addProperty(configId, m_args[i]);
                LOG.debug("property value: " + m_args[i]);
            } else if (model.isOption(key)) {
            	LOG.debug("found an option");
            	config.addOption(configId);
            } else {
            	throw new ConfigurationException(ErrorMsg.ARG_NOT_FOUND, arg);
            }
        }
        return config;
    }

    /**
     * Trims the argument Striping off all leading '-' characters (up to a 
     * maximum of 2 hyphens). If there are no leading '-' characters or if 
     * more than 2 - are found, null is returned.
     *
     * @param arg the argument to strip.
     * @return the argument without any leading hyphens.
     */
    private String trim(String arg) {
        if (!arg.startsWith("-") || 
        	arg.startsWith("---")) {
        	return null;
        }
        int index = arg.lastIndexOf('-') + 1;
        if (arg.startsWith("--")) {
	        if (arg.length() == (index +1)) {
	        	return null;
	        }
        }
        if (index < arg.length()) {
        	return arg.substring(index);
        } 
        return null;
    }

    /**
     * Test if a key represents a set of single character options. This is true
     * if it begins with a single '-' character followed by one or more other
     * characters.
     *
     * @param key the option string to be checked.
     * @return true if the string starts with no hyphens and the length is
     * greater than one char. 
     */
    private boolean isSingleCharOption(String key) {
        if (!key.startsWith("-")) {
            return false;
        } else if (key.startsWith("--")) {
            return false;
        } else if (key.length() == 1) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * Separate out single-character options into separate arguments. E.g.
     * -abcd turns into -a -b -c -d
     *
     * @param args the command line list to be separated.
     * @return a list containing all separated compound options.
     */
    private String[] preprocessArgs(String[] args) {
        Vector newArgs = new Vector();
        for (int i = 0; i < args.length; i++) {
            if (isSingleCharOption(args[i])) {
                String chars = trim(args[i]);
                for (int j = 0; j < chars.length(); j++) {
                    newArgs.add("-" + chars.charAt(j));
                }
            } else {
                newArgs.add(args[i]);
            }
        }
        return (String[]) newArgs.toArray(new String[0]);
    }
    
}
