/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core.config.parser;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.ArrayList;

import org.apache.log4j.Logger;

// dk
//import org.glite.apel.core.Log;


/**
 * Class used to encapsulate all parsed options and properties read-in from 
 * an input source. An input source in this case could be a configuration 
 * file or a command line string.
 */
public class Config {
	
    private static final Logger LOG = 
    	Logger.getLogger(Config.class.getName());
	
    /**
     * Contains the options currently configured.
     */
	private Vector m_options;
	
    /**
     * Contains all the properties.
     */
	private Hashtable m_properties;
        private Hashtable m_options_hash;
	
	Config() {
		m_options = new Vector();         // good for options that have a "yes" or "no"
                m_options_hash = new Hashtable(); // options where values need to be recorded 
		m_properties = new Hashtable();
	}
	
	/**
	 * Resets the current configuration.
	 */
	public void reset() {
		m_options.clear();
		m_properties.clear();
                m_options_hash.clear();
	}
	
	/**
	 * Adds the option to the list of current user options.
	 */
	public void addOption(Object option) {
		m_options.add(option);
	}
	
	/**
	 * Checks the state of the option. If the option state is defined 
	 * as either 'false' or 'no', the option is removed from the
	 * list of current options (ie its disabled). Otherwise the 
	 * option is added (ie the option is enabled).
	 * 
	 * @param option the internal name of the option to be checked and 
	 * added if necessary. 
	 * @param value the state of the option.
	 */
	public void toggleOption(String option, String value) {
		if (isToggleType(value)) {
			if (booleanValue(value) && !m_options.contains(option)) {
				m_options.add(option);
			} else {
				// Option is disabled, so remove it
				m_options.remove(option);
			}
		} else {
			LOG.debug(option + ": state is invalid: " + value);
		}
	}
	
	/**
	 * Adds the property value to the configuration.
	 * 
	 * @param key the key of the property.
	 * @param value the value associated with the property.
	 */
	public void addProperty(String key, String value) {
        Vector propsList = null;
		if (m_properties.containsKey(key)) {
            propsList = (Vector) m_properties.get(key);
        } else {
            propsList = new Vector();
        }
		propsList.add(value);
		m_properties.put(key, propsList);
	}


        /** 
         * Adds the option value to the configuration.
         *
         * @param key the key of the property.
         * @param value the value associated with the property.
         */
        public void addAttributeOption(String key, String value) {
        Vector propsList = null;
                if (m_options_hash.containsKey(key)) {
            propsList = (Vector) m_options_hash.get(key);
        } else {
            propsList = new Vector();
        }
                propsList.add(value);
                m_options_hash.put(key, propsList);
        }

	
    /**
     * Get the value of a property if it is set. If the property has multiple
     * values, return the last value set.
     *
     * @param property property name
     * @return value of specified property, or null if not present
     */
    public String getOptionalProperty(String property) {
        Vector prop = (Vector) m_properties.get(property);
        if (prop == null) {
            return null;
        }
        return (String) prop.lastElement();
    }

    /**
     * Get all values of a property as a Vector.
     *
     * @param property property name
     * @return Vector of values set for the specified property.
     */
    public Vector getPropertyValues(String property) {
        Vector prop = (Vector) m_properties.get(property);
        if (prop == null) {
            return new Vector();
        }
        return prop;
    }
    
    /**
     * Get the value of a property. If the property has multiple values, return
     * the last value set. This  means that values that are set later in the
     * configuration process can override the earlier set values.
     *
     * @param property property name
     * @param secret if true, do not log value. Only that it has been read
     * @return value of specified property, or null if it does not exist.
     */
    public String getProperty(String property) {
    	Vector prop = (Vector) m_properties.get(property);
        if (prop == null) {
        	return null;
        }
        return (String) prop.lastElement();
    }	

    /** 
     * Get the value of a property. If the property has multiple values, return
     * the last value set. This  means that values that are set later in the
     * configuration process can override the earlier set values.
     *
     * @param property property name
     * @param secret if true, do not log value. Only that it has been read
     * @return value of specified property, or null if it does not exist.
     */
    public String getOptionValue(String property) {
        Vector prop = (Vector) m_options_hash.get(property);
        if (prop == null) {
                return null;
        }
        return (String) prop.lastElement();
    }

    /**
     * Check if the specified option has been marked as set.
     *
     * @param option Option name
     * @return <code>true</code> if it is set, <code>false</code> otherwise.
     */
    public boolean isOptionSet(String option) {
        return m_options.contains(option);
    }

    /**
     * Return <code>true</code> if a property has a value set.
     *
     * @param name property name
     * @return <code>true</code> if a value is set, <code>false</code>
     * otherwise.
     */
    public boolean isPropertySet(String name) {
        return m_properties.containsKey(name);
    }
    
    /**
     * Test if a string represents a Boolean true or false value. Recognizes
     * 'true', 'false', 'yes' and 'no'. Note that we can't use the built-in
     * Boolean.valueOf() method because it doesn't recognize 'yes' and 'no'.
     *
     * @param s String to test
     * @return the bool value depending on the given conditions above.
     */
    private boolean booleanValue(String s) {
        String test = s.trim();
        if (test.equalsIgnoreCase("yes")) {
            return true;
        } else if (test.equalsIgnoreCase("true")) {
            return true;
        } else if (test.equalsIgnoreCase("no")) {
            return false;
        } else if (test.equalsIgnoreCase("false")) {
            return false;
        } else {
            return false;
        }
    }
    
    /**
     * Checks the string is a switchable type ie contains
     * one of the following states: [true, false, yes, no].
     *  
     * @param s the string to check.
     * @return true if the string matches the above constraint, false
     * otherwise.
     */
    private boolean isToggleType(String s) {
        String test = s.trim();
        if (test.equalsIgnoreCase("yes") ||
   			test.equalsIgnoreCase("no") ||
       		test.equalsIgnoreCase("true") ||
			test.equalsIgnoreCase("false")) {
            return true;
        }
        return false;
    }

	/**
	 * Adds an attribute to the list of configs. An attribute is 
	 * a toggle value (ie contains two states: true/false). If the
	 * attribute is found to be false, any existing properties that
	 * belong to this attribute are removed.
	 * 
	 * @param internalId the internal key of the attribute.
	 * @param value the value to be set.
	 */
	public void addAttribute(String internalId, String value) {
    	if (booleanValue(value)) {
    		if (m_properties.containsKey(internalId)) {
    			// Want to reset all the properties to true
    			Vector props = (Vector) m_properties.get(internalId);
    			props.clear();
    		}
    		addProperty(internalId, value);
    	} else {
    		// The property has been set to false, so remove all traces of it
    		m_properties.remove(internalId);
    	}
	}
    
    /**
     * Returns a string view of the internal data structure used. Any
     * properties defined in hideProperties will be concealed.
     */
    public String toString(Vector hideProperties) {
        StringBuffer buf = new StringBuffer(128);
        buf.append(m_options);
        buf.append(" [");
        Enumeration keys = m_properties.keys();
        while (keys.hasMoreElements()) {
            String key = (String) keys.nextElement();
            if (hideProperties.contains(key)) {
                buf.append(key + "=****");
            } else {
                Vector prop = (Vector) m_properties.get(key);
                if (prop.size() > 1) {
                    buf.append(key + "=" + prop);
                } else {
                    buf.append(key + "=" + prop.get(0));
                }
            }
            if (keys.hasMoreElements()) {
            	buf.append(", ");
            }
        }
        buf.append("]");
        return buf.toString();
    }

    
	/**
	 * Gets the number of found options. Usefull for unit testing
	 * purposes.
	 * 
	 * @return the number of options set by the user.
	 */
	public int getNumberOptions() {
		return m_options.size();
	}

	/**
	 * Gets the number of set properties. Usefull for unit testing
	 * purposes.
	 * 
	 * @return the number of properties set by the user.
	 */
	public int getNumberProperties() {
		return m_properties.size();
	}
	
}
