/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.config.parser;

import java.util.Vector;
import java.util.Hashtable;

/**
 * Class is used to encapsulate the config mapping of user options/params
 * to an internal representation. The model is used by the ConfigParser
 * to check a found option/param is valid. 
 */
public class ConfigMap {
	
	/** 
     * Contains a list of all the known user option tokens. 
     */
	Vector m_usrOpts;
	
	/** 
     * Contains a list of all the known user parameter tokens. 
     */
	Vector m_usrProps;
	
	/** 
     * Contains all the valid tokens of the model. 
     */
	Vector m_tokens;
	
    /** 
     * Maps each deprecated option to the suggested replacement. 
     */
    public Hashtable m_deprecated;
    
	/** 
     * Maps the user view of an option/param to an internal view
	 * for easy manipulation. 
     */
	Hashtable m_intern;
	
	public ConfigMap() {
		m_usrOpts = new Vector();
		m_usrProps = new Vector();
		m_tokens = new Vector();
		m_intern = new Hashtable();
        m_deprecated = new Hashtable();
	}
	
	/**
	 * Adds the token to the list of valid and well-known tokens.
	 * When a parse is carried out, if a value does not match the 
	 * token list, the configuration is incorrect. This is used to 
	 * provide simple type checking for XML config formats.
	 *  
	 * @param token the token to be added.
	 */
	public void addTok(String token) {
		m_tokens.add(token);
	}
	
	/**
	 * Adds the option to the ConfigModel.
	 * 
	 * @param usrOpt the option name as manipulated by the user.
	 * @param internalOpt the internal representation of the option.
	 */
	public void addOpt(String usrOpt, String internOpt) {
		if (!m_usrOpts.contains(usrOpt)) {
			m_usrOpts.add(usrOpt);
		}
		m_intern.put(usrOpt, internOpt);
	}
	
	/**
	 * Adds the parameter to the ConfigModel.
	 * 
	 * @param usrParam the parameter name manipulated by the user.
	 * @param internParam the internal representation of the parameter.
	 */
	public void addParam(String usrParam, String internParam) {
		if (!m_usrProps.contains(usrParam)) {
			m_usrProps.add(usrParam);
		}
		m_intern.put(usrParam, internParam);
	}
	
    /**
     * Adds the option/property to the deprecated list.
     * 
     * @param usrConfig the config name manipulated by the user.
     * @param description the suggested option/property to use instead.
     */
    public void addDeprecated(String usrConfig, String description) {
    	m_deprecated.put(usrConfig, description);
    }
    
    /**
     * Returns the description about the deprecated config.
     * 
     * @param config the deprecated property name.
     */
    public String getDeprecatedDescription(String config) {
        return (String) m_deprecated.get(config);
    }
    
    /**
     * Test if a key matches a known option.
     *
     * @param opt the value to be checked.
     * @return true if the key is an option, false otherwise.
     */
    public boolean isOption(String opt) {
        if (opt == null) {
            return false;
        } else if (m_usrOpts.contains(opt)) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Test if a key matches a known property
     *
     * @param prop the property to check.
     * @return true if the key is a property, false otherwise.
     */
    public boolean isProperty(String prop) {
        if (prop == null) {
            return false;
        } else if (m_usrProps.contains(prop)) {
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Returns true if the config is contained in the deprecated
     * map.
     */
    public boolean isDeprecated(String config) {
    	return m_deprecated.containsKey(config);
    }
    
	/**
	 * Checks the token is a valid part of the configuration
	 * syntax.
	 * 
	 * @param name the name of the token to be verified.
	 * @return true if the name is valid, false otherwise.
	 */
	public boolean isValidToken(String name) {
		return m_tokens.contains(name);
	}
    
    /**
     * Gets the internal id of the parameter name. If the token is
     * undefined, returns null.
     * 
     * @param token The name of the parameter.
     * @return The internal id of the parameter.
     */
    public String intern(String token) {
        if (token == null) {
            return null;
        }
        return (String) m_intern.get(token);
    }
    
    /**
     * Returns a representation of the stored options and properties.
     */
    public String toString() {
    	StringBuffer buffer = new StringBuffer();
    	String nl = System.getProperty("line.separator");
    	buffer.append(m_tokens);
    	buffer.append(nl);
    	buffer.append(m_usrOpts);
    	buffer.append(nl);
    	buffer.append(m_usrProps);
    	buffer.append(nl);
    	buffer.append(m_intern);
    	buffer.append(nl);
        return buffer.toString();
    }
    
}
