/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.config.parser;

import org.apache.log4j.Logger;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;

import javax.xml.parsers.SAXParserFactory;

/**
 * The XMLParser is used to decode a generic xml property file. A ConfigModel
 * object containing a list of valid elements/attributes is passed to the
 * parse routine and is used to retrieve the relavant configuration from the
 * XML document.
 */
public class XMLParser extends DefaultHandler implements ConfigParser {
	
    /**
     * Reference to logging tool.
     */
    private static final Logger LOG = 
    	Logger.getLogger(XMLParser.class.getName());
    
    /** 
     * Current column value being parsed. 
     */
    protected String m_propertyName;

    /** 
     * Current column value being parsed. 
     */
    protected StringBuffer m_property;

    /** 
     * Config file reader 
     */
    private Reader m_reader;

    /** 
     * Encapsulates all the options and properties read-in from the 
     * config source. 
     */
    private Config m_config;
    
    /** 
     * The current model the parser uses to validate the XML content. 
     */
    private ConfigMap m_model;
    
    /**
     * Creates a new xml parser object.
     */
    private XMLParser(){
    	m_config = new Config();
    	m_property = new StringBuffer();
    	m_propertyName = "";
    }
    
    /**
     * Creates a new XMLParser which will attempt to parse from the 
     * Reader object.
     *
     * @param reader Reader reading XML configuration information
     */
    public XMLParser(Reader reader) {
    	this();
    	m_reader = reader;
    }

    /**
     * Creates a new XMLParser, the parse routine will be applied to
     * the XMl config file.
     * 
     * @param filePath path of the file to parse.
     * @throws FileNotFoundException if filePath does not exist.
     */
    public XMLParser(String filePath) throws FileNotFoundException {
    	this();
    	m_reader = new FileReader(filePath);
    }

    /**
     * Extracts the configuration information from the XML.
     *
     * @returns a configuration object that wraps all the options and
     * properties parsed from the XML.
     * @throws ConfigurationException If an unknown argument is supplied or a
     * property is specified with no value.
     */
    public Config parse(ConfigMap model) throws ConfigurationException {
        m_model = model;
    	try {
            SAXParserFactory saxFactory = SAXParserFactory.newInstance();
            XMLReader xmlReader = saxFactory.newSAXParser().getXMLReader();
            xmlReader.setContentHandler(this);
            xmlReader.parse(new InputSource(m_reader));
        } catch (Exception e) {
            throw new ConfigurationException(ErrorMsg.ERROR_PARSING_FILE, 
            	e.getMessage());
        }
        return m_config;
    }

    /**
     * Add character data between elements to the current property value.
     */
    public void characters(char[] ch, int start, int length) {
        String text = new String(ch, start, length);
        //LOG.debug("found text: " + text);
        m_property.append(text);
    }

    /**
     * Adds the contents of this element as a property with the element 
     * name as the key.
     */
    public void endElement(String uri, String localName, String elementName)
        throws SAXException {
    	String trimmedProperty = m_property.toString().trim();
        if (!trimmedProperty.equals("")) {
        	String internalId = m_model.intern(m_propertyName);
            if (internalId != null) {
            	//LOG.debug("internal id: " + internalId);
            	m_config.addProperty(internalId, trimmedProperty);
            }
        }
        m_property = new StringBuffer();
        m_propertyName = removeLastElement(m_propertyName);
    }

    /**
     * When an element is encountered the attributes are parsed and
     * added to the current config.
     */
    public void startElement(String uri, String localName, 
    	String elementName, Attributes attributes) throws SAXException {
        m_propertyName = m_propertyName + "/" + elementName;
        if (m_model.isDeprecated(m_propertyName)) {
            //LOG.debug("ignoring deprecated config: " + m_propertyName);
        } else if (m_model.isValidToken(m_propertyName)) {
            if (m_model.isOption(m_propertyName)) {
	            m_config.addOption(m_model.intern(m_propertyName));
        	}
            setAttributes(attributes);
        } else {
        	throw new SAXException(ErrorMsg.get(ErrorMsg.UNKNOWN_ELEMENT) + 
                m_propertyName);
        }
    }

    /**
     * Extracts all attributes from an element and adds them to the hashtable
     * of properties.
     *
     * @param attributes the list of attributes to be extracted.
     */
    protected void setAttributes(Attributes attributes) 
    	throws SAXException {
    	String fullName = m_propertyName + "@";
        for (int i = 0; i < attributes.getLength(); i++) {
            String name = attributes.getLocalName(i);
            if (name.equals("")) {
            	name = attributes.getQName(i);
            }
        	String path = fullName + name;
            //LOG.debug("found attribute: " + path);
            if (m_model.isDeprecated(path)) {
                //LOG.debug("ignorning deprecated attribute: " + path);
            } else {
                String internalId = m_model.intern(path);
                if (internalId == null) {
                	throw new SAXException(
                        ErrorMsg.get(ErrorMsg.UNKNOWN_ATTRIBUTE) +
                		fullName + name);
                }
        		String value = attributes.getValue(i);
            	if (m_model.isOption(path)) {
            		m_config.toggleOption(internalId, value);
                        m_config.addAttributeOption(internalId, value);
            	} else {           	
            		m_config.addAttribute(internalId, value);
                        m_config.addAttributeOption(internalId, value);
            	}
            }
        }
    }

    /**
     * Remove the last element from a simple XPath expression E.g. given
     * /e1/e2/e3 would return /e1/e2
     */
    private static String removeLastElement(String xpath) {
        int lastSlash = xpath.lastIndexOf('/');
        return xpath.substring(0, lastSlash);
    }

    /**
     * Tidies up the Reader object of the XMLParser once gc'ed.
     */
    public void finalize() throws Throwable {
        try {
            if (m_reader != null) {
                m_reader.close();
            }
        } catch (IOException e) {
            LOG.debug("Could not close configuration reader");
        }
        super.finalize();
    }
    
}
