/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.db;

import org.glite.apel.core.ApelException;
import org.glite.apel.core.UTCTime;
import org.glite.apel.core.log.type.AccountRecordIterator;
import org.glite.apel.core.log.type.AccountSummaryIterator;
import org.glite.apel.core.log.type.EventRecordIterator;
import org.glite.apel.core.log.type.GkRecordIterator;
import org.glite.apel.core.log.type.MsgRecordIterator;
import org.glite.apel.core.log.type.BlahdRecordIterator;
import java.util.ArrayList;

/**
 * Contains all the methods that are used when accessing the apel db.
 */
public interface ApelDB {
    /** 
     * Event end event records table. 
     */
    public String EVENT_TABLE_NAME = "EventRecords";

    /** 
     * Gate Keeper log file table. 
     */
    public String GK_TABLE_NAME = "GkRecords";

    /**
     * Blahd log file table.
     */
    public String BLAHD_TABLE_NAME = "BlahdRecords";

    /** 
     * LCG accounts table. 
     */
    public String LCG_RECORDS_TABLE_NAME = "LcgRecords";

    /**
     * RGMA Summary table.
     */
    public String LCG_RECORDS_SYNC_TABLE_NAME = "LcgRecordsSync_v2";

    /** 
     * List of files which have been processed. 
     */
    public String PROCESSED_FILES_TABLE_NAME = "LcgProcessedFiles";
    
    /** 
     * Worker node CPU power table. 
     */
    public String SPEC_TABLE_NAME = "SpecRecords";

   /**
    * A backup of site SpecRecords table for savannah bug fix _xxxxx
    */
    //public String SPEC_TABLE_TMP_NAME = "SpecRecords_25430";
    public String SPEC_TABLE_TMP_NAME = "SpecRecords_28593";
    public String SPEC_TABLE_TMP_NAME2 = "SpecRecords_65723";

    /** 
     * Message records table.
     */
    public String MSG_TABLE_NAME = "MessageRecords";

    /**
     * Reprocess info table.
     */
    public String REPUBLISH_TABLE_NAME = "RepublishInfo";    
    
    /** 
     * Constant used to delimit special table columns that are empty. 
     */
    public static final String NULL = "NULL";
    
    /**
     * Log file type.
     */
    public static final char FILE_TYPE_EVENT_LOG = 'E';
    
    /** 
     * Gate Keeper log file type. 
     */
    public static final char FILE_TYPE_GATE_KEEPER_LOG = 'G';
    
    /**
     * Blahd Accounting log file type.
     */
    public static final char FILE_TYPE_BLAHD_LOG = 'B';

    /** 
     * Message log file type. 
     */
    public static final char FILE_TYPE_MESSAGE_LOG = 'M';
    
    /**
     * Gets Local HostName
     */
    public String getHostName() 
        throws ApelException;

    /**
     * Deletes all records from the table.
     */
    public void deleteAllRecords(String tableName) 
        throws ApelException;

    /**
     * Returns the number of records stored in the table.
     */
    public int count(String tableName) 
        throws ApelException;

    /**
     * Deletes all event records with set processed flag.
     */
    public void deleteProcessedEventRecords() 
        throws ApelException;
    
    /**
     * Deletes all gk records with set processed flag.
     */
    public void deleteProcessedGkRecords() 
        throws ApelException;
   
    /**
     * Deletes all blahd records with set processed flag.
     */
    public void deleteProcessedBlahdRecords()
        throws ApelException;

 
    /**
     * Deletes all message records with set processed flag.
     */
    public void deleteProcessedMsgRecords() 
        throws ApelException;
    
    /**
     * Deletes all records with matching filetype from the processed file
     * table. It is meant specifically as an easy way of allowing complete 
     * data processing from scratch.
     *
     * @param fileType The file type to be deleted ('E' : Pbs log file,
     * 'G' : Gk log file, 'M' : Message log file).
     * @throws ApelException If an error occurs while connecting to the
     * database.
     */
    public void deleteAllFiles(char fileType) 
        throws ApelException;
    
    /**
     * Inserts records into the table from the specified event log file.
     * Returns the number of records inserted.
     */
    public int insertEventRecords(EventRecordIterator eventRecordsIt) 
        throws ApelException;
    
    /**
     * Inserts all records from the specified gk iterator into the db.
     * Returns the number of records inserted.
     */
    public int insertGkRecords(GkRecordIterator gkRecordIt) 
        throws ApelException;
    
   /**
     * Inserts all records from the specified blahd iterator into the db.
     * Returns the number of records inserted.
     */
    public int insertBlahdRecords(BlahdRecordIterator blahdRecordIt)
        throws ApelException;

    /**
     * Adds all records from the message iterator to the message table.
     * Returns the number of records inserted.
     */
    public int insertMsgRecords(MsgRecordIterator msgRecordIt) 
        throws ApelException;
    
    /**
     * If the job manager type is vanila PBS, a short cut exists whereby
     * the GramScriptID is the same as the JobName. Returns the number of
     * message records inserted.
     */
    public int insertMsgRecords(GkRecordIterator gkRecordIt) 
        throws ApelException;
        
    /**
     * Adds a filename to the processed files table. 
     * 
     * @param fileName The base name of the file.
     * @param fileType The file type (eg event, gk or message file).
     * @param fullPath The full path of the file.
     * @throws ApelException If an error occurs while carrying out the batch
     * insert.
     */
    public void insertProcessedFile(String fileName, char fileType, 
        String fullPath) throws ApelException;
   
   /**
    * Processes the SpecRecord - intended to replace insertSpecRecord
    */
    public void processSpecRecord(String clusterId, String subClusterId,
        String specInt, String specFloat, String date, String time)
        throws ApelException;

    /**
     * Attempts to generate new accounting records, inserting them
     * within the accounting record table. Returns an objectified view
     * of the newly inserted records.
     */
    public ArrayList generateAccountRecords(
        boolean ignoreProcessedFlag, int limit, int offset) throws ApelException;
    
    /**
     * Attempts to generate new accounting records for the Glite CE, inserting them
     * within the accounting record table. Returns an objectified view
     * of the newly inserted records.
     */

    public ArrayList generateAccountRecordsGliteCE(boolean ignoreProcessedFlag, int limit, int offset) throws ApelException;


    /**
     * Checks if the file exists in the processed files table.
     *
     * @param fileName The base name of the file.
     * @param fileType The file type to check (eg event, gk or message).
     * @return True if it has an entry in the processed files table, false
     * otherwise.
     * @throws ApelException If an error occurs while connecting to the
     * database.
     */
    public boolean isFileProcessed(String fileName, char fileType) 
        throws ApelException;


    /**
     * Checks if the file exists in the processed files table.
     *
     * @param filePath The Fully qualified fileName
     * @param hostName The local hostName
     * @param fileType The file type to check (eg event, gk or message).
     * @return True if it has an entry in the processed files table, false
     * otherwise.
     * @throws ApelException If an error occurs while connecting to the
     * database.
     */
    public boolean isFileProcessed(String filePath, String hostName, char fileType)
        throws ApelException;

    /**
     * Returns a count of accounting records available after the date 
     * and time period.
     */
    public int getNumAccountingRecords(String date, String time) 
        throws ApelException;

    /**
     * Returns a count of accounting records available after the date
     * and time period for a specified site.
     */
    public int getNumAccountingRecords(String date, String time, String siteName)
        throws ApelException;
 
    /**
     * DaveKant
     * Returns a count of accounting records available in a Gap interval
     */
    public int getNumAccountingRecordsInGap(String date_start, String date_end)
        throws ApelException;

    /**
     * Returns a batch of accounting records that match or exceed the 
     * time stamp.
     */
    public ArrayList getAccountingRecords(String date, String time, int limit, int offset) 
        throws ApelException;
    
   /**
     * Returns a batch of accounting records that match or exceed the
     * time stamp for a specified site.
     */
    public ArrayList getAccountingRecords(String date, String time, String siteName, int limit, int offset)
        throws ApelException;

    /**
     * Returns a batch of accounting records available in a Gap interval
     */
    public ArrayList getAccountingRecordsInGap(String date_start, String date_end, int limit, int offset)
        throws ApelException;
 

    /**
     * Returns a batch of summary accounting records.
     */
    public AccountSummaryIterator getAccountingSummary(String siteName)
        throws ApelException;

    /**
     * Returns all records stored in the accounting table.
     */
    public ArrayList getAllAccountingRecords(int limit, int offset) 
        throws ApelException;

    /**
     * Updates all the MeasurementDate/Time columns of the accounting
     * table to contain the current time. This is used when a complete 
     * reprocess is carried out and ensures when the republish option is set 
     * to 'missing', an accurate audit can be made of how many records were 
     * sent during the last join process.
     */
    public void updateAccountingTimeStamp() 
        throws ApelException;
   
    /**
     * Updates the MeasurementDate/Time columns of the accounting
     * table to contain the current time. This is used to update only the data
     * defined in an AccountingRecord container
     */
    public void updateAccountingTimeStamp(ArrayList arIt)
        throws ApelException;

    /**
     * Updates the republish date/time (represented in UTC) to the current
     * time.
     */
    public void updateRepublishTimeStamp() 
        throws ApelException;
    
    /**
     * Returns true only if a repulish time stamp exists within the 
     * republish table. 
     */
    public boolean containsRepublishTimeStamp() throws ApelException;

    /**
     * Initialises the republish date/time.
     */
    public UTCTime setupTimeStamp() throws ApelException;
    
    /**
     * Returns the date when the apel last published any accounting data.
     */
    public String getRepublishDate() throws ApelException;    

    /**
     * Returns the time when the apel last published any accounting data.
     */
    public String getRepublishTime() throws ApelException;
 
    /**
     * Tidies up any looses db connections etc.
     */
    public void close();
    

    /**
     * Setter method for Gap interval
     */
     //public void gapset();

    /**
     * Attempts to carry out any optimizations on the table
     * to help improve the performance of any queries to it.
     */
    public void optimize(String tableName) throws ApelException;

    /**
     * Inspect the structure of a table.
     */
    public void inspectTable(String tableName) throws ApelException;

    //public String getVOFilter(ArrayList<String> VOList,boolean include);
}


