/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.db;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;
import org.apache.log4j.Logger;
import org.glite.apel.core.ApelException;
import org.glite.apel.core.Log;
import org.glite.apel.core.UTCTime;
import org.glite.apel.core.Utils;
import org.glite.apel.core.log.type.AccountRecord;
import org.glite.apel.core.log.type.EventRecord;
import org.glite.apel.core.log.type.EventRecordIterator;
import org.glite.apel.core.log.type.GkRecord;
import org.glite.apel.core.log.type.GkRecordIterator;
import org.glite.apel.core.log.type.MsgRecord;
import org.glite.apel.core.log.type.MsgRecordIterator;
import org.glite.apel.core.log.type.BlahdRecord;
import org.glite.apel.core.log.type.BlahdRecordIterator;
import org.glite.apel.core.log.type.AccountSummary;
import org.glite.apel.core.log.type.AccountSummaryIterator;
import java.util.HashMap;

/**
 * Provides an implementation of all db accessor methods using MySql. 
 */
public class MySQLImpl implements ApelDB {

    /**
     * Using Log4J.
     */
    private Logger m_log;
    /** 
     * Constant used to flag a new record. 
     */
    private static final String UNPROCESSED = "0";
    /** 
     * Constant used to flag a processed record. 
     */
    private static final String PROCESSED = "1";
    /**
     * The default date used by the catch-all spec record.
     */
    private static final String DEFAULT_SPEC_DATE = "1970-01-01";
    /**
     * The default time used by the catch-all spec record.
     */
    private static final String DEFAULT_SPEC_TIME = "00:00:00";
    /**
     * Mysql error code used when duplicate records are inserted.
     */
    private static final int DUPLICATE_KEY_CODE = 1062;
    /**
     * JDBC driver used (should be in config setup).
     */
    private static final String MYSQL_DRIVER = "org.gjt.mm.mysql.Driver";
    /**
     * DB Connection object.
     */
    private Connection m_dbConnection;
    /**
     * The DN of the Site.
     */
    private String m_siteName;
    /**
     * ArrayList for the mappings between groups and VOs
     */
    private static HashMap<String,String> m_VONames;
    private static String m_VOFilter;

    /**
     * Creates a new Table object.
     */
    public MySQLImpl(String url, String username, String password, String siteName) throws ApelException {
        m_log = Logger.getLogger(getClass().getName());
        m_siteName = siteName;
        try {
            Class.forName(MYSQL_DRIVER);
            m_dbConnection =
                    DriverManager.getConnection(url, username, password);
        } catch (Exception e) {
            throw new ApelException("Unable to setup a database connection: "
                    + e.getMessage());
        }
    }

    /**
     * Reprocesses the distribution of keys within the table and compacts
     * the data into an optimal form. Note, this uses dedicated MySQL 
     * functionality and is therefore not portable SQL.
     */
    public void optimize(String tableName) throws ApelException {
        Statement stmt = null;
        try {
            Log.info(m_log, "Optimising table: " + tableName);
            stmt = m_dbConnection.createStatement();
            stmt.execute("ANALYZE TABLE " + tableName);
            stmt.execute("OPTIMIZE TABLE " + tableName);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Inspect a table.
     * Can be used to create a table, or change an existing schema
     */
    public void inspectTable(String tableName) throws ApelException {
        Statement stmt = null;
        ResultSet rs = null;
        ///////////////////////////////////////////////////
        // BLAHD Records Table
        ///////////////////////////////////////////////////
        if (tableName.equals(ApelDB.BLAHD_TABLE_NAME)) {
            try {
                stmt = null;
                rs = null;
                Log.info(m_log, "Checking the " + tableName + " table");
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("SHOW TABLES LIKE '" + tableName + "'");
                rs.next();
                if (rs.getRow() == 0){
                Log.info(m_log, "Table BlahdRecords doesn't exist");
                    createBlahdTable();
                } else {
                    Log.info(m_log, "The " + tableName + " schema is up-to-date ");
                }

            } catch (SQLException e) {
                throw new ApelException(e.getMessage());
            } finally {
                close(stmt);
            }
        }
        /////////////////////////////////////////////////////////
        // LcgRecords Table
        /////////////////////////////////////////////////////////
        if (tableName.equals(ApelDB.LCG_RECORDS_TABLE_NAME)) {
            Log.info(m_log, "Checking the " + tableName + " table");
            try {
                stmt = null;
                rs = null;
                // Check the LCGUserVO
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " LCGUserVO");
                rs.next();
                String LCGUserVO_type = rs.getString(2);

                // Check the LCGJobID
                stmt = null;
                rs = null;
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " LCGJobID");
                rs.next();
                String LCGJobID_type = rs.getString(2);

                // Check the ExecutingCE
                stmt = null;
                rs = null;
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " ExecutingCE");
                rs.next();
                String ExecutingCE_type = rs.getString(2);

                if (LCGUserVO_type.equals("varchar(50)")
                        || LCGJobID_type.equals("varchar(50)")
                        || ExecutingCE_type.equals("varchar(50)")) {
                    alterLcgRecordsTable();
                } else {
                    Log.info(m_log, "The " + tableName + " schema is up-to-date ");
                }
            } catch (SQLException e) {
                throw new ApelException(e.getMessage());
            }
            close(stmt);
        }


        /////////////////////////////////////////////////////////
        // SpecRecords Table
        // Check schema of SpecRecords
        /////////////////////////////////////////////////////////
        if (tableName.equals(ApelDB.SPEC_TABLE_NAME)) {
            Log.info(m_log, "Checking the " + tableName + " table for patch " +
                    "28593");
            try {
                stmt = null;
                rs = null;
                // Check if ClusterID column is Primary
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " SpecInt2000");
                rs.next();
                String specInt2000_key = rs.getString(4);
                if (specInt2000_key.equals("")) {
                    alterSpecRecordsTable();
                }
            } catch (SQLException e) {
                throw new ApelException(e.getMessage());
            }
            close(stmt);
        }

        /////////////////////////////////////////////////////////
        // SpecRecords Table modification, to allow longer CE
        // names in the ClusterID and SubClusterID fields
        // This is needed for the update described in
        // https://savannah.cern.ch/bugs/index.php?65723
        //
        // Check schema of SpecRecords
        /////////////////////////////////////////////////////////
        if (tableName.equals(ApelDB.SPEC_TABLE_NAME)) {
            Log.info(m_log, "Checking the " + tableName + " table for patch " +
                    "65723");
            try {
                String clusterLength = null;
                String subClusterLength = null;
                stmt = null;
                rs = null;
                // Check if ClusterID column is Primary
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("SHOW COLUMNS FROM " + tableName + " LIKE '%lusterID';");
                rs.next();
                clusterLength = rs.getString("Type");
                rs.next();
                subClusterLength = rs.getString("Type");
                if(clusterLength.equals("varchar(50)"))
                {
                    if(subClusterLength.equals("varchar(50)"))
                    {
                        alterSpecRecordsTable2();
                    }
                }
                Log.info(m_log, "The " + tableName + " schema is up-to-date ");
            } catch (SQLException e) {
                throw new ApelException(e.getMessage());
            }
            close(stmt);
        }



        /////////////////////////////////////////////////////////
        // LcgProcessedFiles Table
        /////////////////////////////////////////////////////////
        if (tableName.equals(ApelDB.PROCESSED_FILES_TABLE_NAME)) {
            Log.info(m_log, "Checking the " + tableName + " table");
            try {
                stmt = null;
                rs = null;
                // Check FilePath Column
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " FilePath");
                rs.next();
                String FilePath_key = rs.getString(4);

                // Check HostName Column
                stmt = null;
                rs = null;
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " HostName");
                rs.next();
                String HostName_key = rs.getString(4);

                if (HostName_key.equals("") || FilePath_key.equals("")) {
                    alterLcgProcessedFilesTable();
                } else {
                    Log.info(m_log, "The " + tableName + " schema is up-to-date ");
                }
            } catch (SQLException e) {
                throw new ApelException(e.getMessage());
            }
            close(stmt);
        }


        /////////////////////////////////////////////////////////
        // GkRecords Table
        /////////////////////////////////////////////////////////
        if (tableName.equals(ApelDB.GK_TABLE_NAME)) {
            Log.info(m_log, "Checking the " + tableName + " table");
            try {
                stmt = null;
                rs = null;
                // Check GramSCriptJobID Column
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " GramScriptJobID");
                rs.next();
                String GramScriptJobID_type = rs.getString(2);

                if (GramScriptJobID_type.equals("varchar(100)")) {
                    alterGkRecordsTable();
                } else {
                    Log.info(m_log, "The " + tableName + " schema is up-to-date ");
                }
            } catch (SQLException e) {
                throw new ApelException(e.getMessage());
            }
            close(stmt);
        }

        /////////////////////////////////////////////////////////
        // MsgRecords Table
        /////////////////////////////////////////////////////////
        if (tableName.equals(ApelDB.MSG_TABLE_NAME)) {
            Log.info(m_log, "Checking the " + tableName + " table");
            try {
                stmt = null;
                rs = null;
                // Check GramSCriptJobID Column
                stmt = m_dbConnection.createStatement();
                rs = stmt.executeQuery("DESCRIBE " + tableName + " GramScriptJobID");
                rs.next();
                String GramScriptJobID_type = rs.getString(2);

                if (GramScriptJobID_type.equals("varchar(100)")) {
                    alterMessageRecordsTable();
                } else {
                    Log.info(m_log, "The " + tableName + " schema is up-to-date ");
                }
            } catch (SQLException e) {
                throw new ApelException(e.getMessage());
            }
            close(stmt);
        }
        

    }

    /**
     * Alter the LcgRecords Table
     */
    private void alterLcgRecordsTable() throws ApelException {
        Log.info(m_log, "A schema update is required, please be patient.");
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            stmt.execute("ALTER TABLE " + ApelDB.LCG_RECORDS_TABLE_NAME
                    + " CHANGE COLUMN LCGUserVO LCGUserVO VARCHAR(255),"
                    + " CHANGE COLUMN LCGJobID LCGJobID VARCHAR(255), "
                    + " CHANGE COLUMN ExecutingCE ExecutingCE VARCHAR(255)");
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Alter the SpecRecords Table
     * -> Copy existing table (for backup), truncate, and change schema
     */
    private void alterSpecRecordsTable() throws ApelException {
        Log.info(m_log, "A schema update is required, please be patient.");
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();

            // Copy table
            stmt.execute("CREATE TABLE " + ApelDB.SPEC_TABLE_TMP_NAME
                    + " SELECT * FROM " + ApelDB.SPEC_TABLE_NAME);

            // Truncate table
            stmt.execute("TRUNCATE TABLE " + ApelDB.SPEC_TABLE_NAME);

            // Modify table
            stmt.execute("ALTER TABLE " + ApelDB.SPEC_TABLE_NAME
                    + " DROP PRIMARY KEY, "
                    + " ADD PRIMARY KEY (SpecID,ClusterID,SpecInt2000)");
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Alter the SpecRecords Table
     * -> Copy existing table (for backup), truncate, and change schema
     */
    private void alterSpecRecordsTable2() throws ApelException {
        Log.info(m_log, "A schema update is required for Savannah patch " +
                "65723, please be patient.");
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();

            // Copy table
            stmt.execute("CREATE TABLE " + ApelDB.SPEC_TABLE_TMP_NAME2
                    + " SELECT * FROM " + ApelDB.SPEC_TABLE_NAME);

            // Modify table
            stmt.execute("ALTER TABLE " + ApelDB.SPEC_TABLE_NAME
                    + " MODIFY COLUMN ClusterID varchar(255);");

            // Modify table
            stmt.execute("ALTER TABLE " + ApelDB.SPEC_TABLE_NAME
                    + " MODIFY COLUMN SubClusterID varchar(255);");

            // Set all columns to "Off" so they don't interfere with
            // the new SpecInt retrieval method
            stmt.execute("UPDATE SpecRecords SET EndDate=CURDATE()," +
                    "EndTime=CURTIME() WHERE EndDate IS NULL " +
                    "AND EndTime IS NULL");
            
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Alter the LcgProcessedFiles
     */
    private void alterLcgProcessedFilesTable() throws ApelException {
        Log.info(m_log, "A schema update is required, please be patient.");
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            stmt.execute("ALTER TABLE " + ApelDB.PROCESSED_FILES_TABLE_NAME
                    + " CHANGE COLUMN FilePath FilePath VARCHAR(200),"
                    + " CHANGE COLUMN HostName HostName VARCHAR(100), "
                    + " DROP PRIMARY KEY, "
                    + " ADD PRIMARY KEY (FileName,FileType,SiteName,FilePath,HostName)");
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Alter the GkRecords Table
     */
    private void alterGkRecordsTable() throws ApelException {
        Log.info(m_log, "A schema update is required, please be patient.");
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            stmt.execute("ALTER TABLE " + ApelDB.GK_TABLE_NAME
                    + " CHANGE COLUMN GramScriptJobID GramScriptJobID VARCHAR(255) NOT NULL");
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Alter the MessagesRecords Table
     */
    private void alterMessageRecordsTable() throws ApelException {
        Log.info(m_log, "A schema update is required, please be patient.");
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            stmt.execute("ALTER TABLE " + ApelDB.MSG_TABLE_NAME
                    + " CHANGE COLUMN GramScriptJobID GramScriptJobID VARCHAR(255) NOT NULL");
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Create the BlahdTable
     */
    private void createBlahdTable() throws ApelException {
        Log.info(m_log, "A schema update is required, please be patient.");
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            stmt.execute("CREATE TABLE IF NOT EXISTS " + ApelDB.BLAHD_TABLE_NAME
                    + " (BlahdID VARCHAR(255) PRIMARY KEY,"
                    + "TimeStamp DATETIME NOT NULL default '0000-00-00 00:00:00',"
                    + "GlobalUserName VARCHAR(255),"
                    + "FullyQualifiedAttributeName VARCHAR(255),"
                    + "ResourceIdentity VARCHAR(255),"
                    + "GlobalJobId VARCHAR(255),"
                    + "LrmsId VARCHAR(255),"
                    + "JobName VARCHAR(255),"
                    + "SiteName VARCHAR(50) NOT NULL,"
                    + "ValidFrom DATE,"
                    + "ValidUntil DATE,"
                    + "Processed INT)");
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Deletes all records from the table.
     */
    public void deleteAllRecords(String tableName) throws ApelException {
        // Operations on tales may be performed such as checking that it exists before 
        // attempting to delete data
        // Here the new BlahdRecords table is checked
        inspectTable(tableName);

        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            int numDels = stmt.executeUpdate("DELETE FROM " + tableName);
            Log.info(m_log, "Rows deleted from " + tableName + ": "
                    + numDels);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Returns the number of records stored in the table.
     */
    public int count(String tableName) throws ApelException {
        Statement stmt = null;
        int count = 0;
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT COUNT(*) FROM " + tableName);
            rs.next();
            count = rs.getInt(1);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
        return count;
    }

    /**
     * Updates the table setting the processed flag where each result row
     * id maps to a row found in the table.
     */
    private void updateProcessedEventRecords(ResultSet joinResult) throws ApelException {
        try {
            joinResult.beforeFirst();
            String update = "UPDATE " + EVENT_TABLE_NAME + " SET Processed = "
                    + PROCESSED + " WHERE EventID = ?";
            PreparedStatement updateStmt =
                    m_dbConnection.prepareStatement(update);
            while (joinResult.next()) {
                updateStmt.clearParameters();
                updateStmt.setObject(1, joinResult.getString("EventID"));
                updateStmt.execute();
            }
            close(updateStmt);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
    }

    /**
     * Deletes all records with set processed flag.
     */
    public void deleteProcessedEventRecords() throws ApelException {
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            int numDels = stmt.executeUpdate("DELETE FROM " + EVENT_TABLE_NAME
                    + " WHERE Processed = " + PROCESSED);
            Log.info(m_log, "Number deleted processed logs from: "
                    + EVENT_TABLE_NAME + ": " + numDels);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Inserts records into the table from the iterator.
     * Returns the number of records inserted.
     */
    public int insertEventRecords(EventRecordIterator eventRecordsIt)
            throws ApelException {
        int insertCount = 0;
        PreparedStatement insertStmt = null;
        try {
            String insert = "REPLACE INTO " + EVENT_TABLE_NAME + " VALUES "
                    + "(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
            insertStmt = m_dbConnection.prepareStatement(insert);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        while (eventRecordsIt.hasNext()) {
            try {
                EventRecord pbsRecord = eventRecordsIt.next();
                // DaveKant : Removes records from the vector once they are written to database.
                // This helps keep memory usage down when processing large log event logs.
                //
                eventRecordsIt.remove();

                String eventDate = pbsRecord.getEventDate();
                String eventTime = pbsRecord.getEventTime();
                String jobName = pbsRecord.getJobName();
                String submitHost = pbsRecord.getSubmitHost();
                String recordId = eventDate + " " + eventTime
                        + " " + jobName + " " + submitHost + " " + m_siteName;
                insertStmt.clearParameters();
                insertStmt.setObject(1, recordId);
                insertStmt.setObject(2, m_siteName);
                insertStmt.setObject(3, jobName);
                insertStmt.setObject(4, pbsRecord.getLocalUserID());
                insertStmt.setObject(5, pbsRecord.getLocalUserGroup());
                insertStmt.setObject(6, pbsRecord.getWallDuration());
                insertStmt.setObject(7, pbsRecord.getCpuDuration());
                insertStmt.setObject(8, pbsRecord.getWallDurationSeconds());
                insertStmt.setObject(9, pbsRecord.getCpuDurationSeconds());
                insertStmt.setObject(10, pbsRecord.getStartTime());
                insertStmt.setObject(11, pbsRecord.getStopTime());
                insertStmt.setObject(12, pbsRecord.getStartTimeUtc());
                insertStmt.setObject(13, pbsRecord.getStopTimeUtc());
                insertStmt.setObject(14, pbsRecord.getStartTimeEpoch());
                insertStmt.setObject(15, pbsRecord.getStopTimeEpoch());
                insertStmt.setObject(16, submitHost);
                insertStmt.setObject(17, pbsRecord.getMemoryReal());
                insertStmt.setObject(18, pbsRecord.getMemoryVirtual());
                insertStmt.setObject(19, UNPROCESSED);
                insertStmt.setObject(20, eventDate);
                insertStmt.setObject(21, eventTime);
                insertStmt.execute();
                insertCount++;

            } catch (SQLException e) {
                if (!isDuplicate(e)) {
                    throw new ApelException(e);
                }
            }
        }
        close(insertStmt);
        return insertCount;
    }

    /**
     * Returns true only if the record already exists within the database.
     */
    private boolean isDuplicate(SQLException e) {
        if (e.getErrorCode() == DUPLICATE_KEY_CODE) {
            return true;
        }
        return false;
    }

    /**
     * Updates the gk table setting the processed flag where each 
     * result row id maps to a row found in the table.
     */
    private void updateProcessedGkRecords(ResultSet joinResult)
            throws ApelException {
        try {
            String update = "UPDATE " + GK_TABLE_NAME + " SET Processed = "
                    + PROCESSED + " WHERE GkID = ?";
            PreparedStatement updateStmt =
                    m_dbConnection.prepareStatement(update);
            joinResult.beforeFirst();
            while (joinResult.next()) {
                updateStmt.clearParameters();
                updateStmt.setObject(1, joinResult.getString("GkID"));
                updateStmt.execute();
            }
            close(updateStmt);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
    }

    /**
     * Deletes all gk records with set processed flag.
     */
    public void deleteProcessedGkRecords() throws ApelException {
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            int numDels = stmt.executeUpdate("DELETE FROM " + GK_TABLE_NAME
                    + " WHERE Processed = " + PROCESSED);
            Log.info(m_log, "Number deleted processed logs from: "
                    + GK_TABLE_NAME + ": " + numDels);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Inserts all records from the specified event log file into the db.
     */
    public int insertGkRecords(GkRecordIterator gkRecordIt) throws ApelException {
        int insertCount = 0;
        String insert = "REPLACE INTO " + GK_TABLE_NAME + " VALUES "
                + "(?,?,?,?,?,?,?,?)";
        PreparedStatement insertStmt = null;
        try {
            insertStmt = m_dbConnection.prepareStatement(insert);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        while (gkRecordIt.hasNext()) {
            GkRecord gkRecord = gkRecordIt.next();
            String measurementDate = gkRecord.getMeasurementDate();
            String measurementTime = gkRecord.getMeasurementTime();
            String gramScriptJobId = gkRecord.getGramScriptJobID();
            //
            // The siteName is not a good variable to add in the recordIdentity.
            // 
            String recordId = measurementDate + " " + measurementTime + " " + gramScriptJobId + " " + m_siteName;
            try {
                insertStmt.clearParameters();
                insertStmt.setObject(1, recordId);
                insertStmt.setObject(2, gramScriptJobId);
                insertStmt.setObject(3, gkRecord.getLocalJobID());
		insertStmt.setObject(4, UNPROCESSED);
                insertStmt.setObject(5, gkRecord.getGlobalUserName());
                insertStmt.setObject(6, m_siteName);
                insertStmt.setObject(7, gkRecord.getValidFrom());
                insertStmt.setObject(8, gkRecord.getValidUntil());
                insertStmt.execute();
                insertCount++;
            } catch (SQLException e) {
                if (!isDuplicate(e)) {
                    throw new ApelException(e);
                }
            }
        }
        close(insertStmt);
        return insertCount;
    }

    /**
     * Updates the blahd table setting the processed flag where each
     * result row id maps to a row found in the table.
     */
    private void updateProcessedBlahdRecords(ResultSet joinResult)
            throws ApelException {
        try {
            String update = "UPDATE " + BLAHD_TABLE_NAME + " SET Processed = " + PROCESSED + " WHERE BlahdID = ?";
            PreparedStatement updateStmt =
                    m_dbConnection.prepareStatement(update);
            joinResult.beforeFirst();
            while (joinResult.next()) {
                updateStmt.clearParameters();
                updateStmt.setObject(1, joinResult.getString("BlahdID"));
                updateStmt.execute();
            }
            close(updateStmt);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
    }

    /**
     * Deletes all blahd records with set processed flag.
     */
    public void deleteProcessedBlahdRecords() throws ApelException {
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            int numDels = stmt.executeUpdate("DELETE FROM " + BLAHD_TABLE_NAME
                    + " WHERE Processed = " + PROCESSED);
            Log.info(m_log, "Number deleted processed logs from: " + BLAHD_TABLE_NAME + ": " + numDels);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Inserts all records from the specified log file into the db.
     */
    public int insertBlahdRecords(BlahdRecordIterator blahdRecordIt) throws ApelException {
        int insertCount = 0;
        String insert = "REPLACE INTO " + BLAHD_TABLE_NAME + " VALUES (?,?,?,?,?,?,?,?,?,?,?,?)";
        PreparedStatement insertStmt = null;
        try {
            insertStmt = m_dbConnection.prepareStatement(insert);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        while (blahdRecordIt.hasNext()) {
            BlahdRecord blahdRecord = blahdRecordIt.next();
            String timeStamp = blahdRecord.getTimeStamp();
            String jobId = blahdRecord.getJobId();
            String ceId = blahdRecord.getCeId();
            String fqan = blahdRecord.getUserFQAN();
            String lrmsId = blahdRecord.getLrmsId();

            ///////////////////////////////////////////////////////////////////////////////////////
            // Unique record identifier : SITENAME NOT USED BECAUSE SITES CAN CHANGE THEIR NAMES!
            ////////////////////////////////////////////////////////////////////////////////////// 
            // Choice of recordId is taken from https://savannah.cern.ch/patch?898
            String recordId = timeStamp + " " + jobId + " " + ceId + " " + lrmsId;
            //String recordId = timeStamp + " " + jobId + " " + ceId ;

            try {
                insertStmt.clearParameters();
                insertStmt.setObject(1, recordId);
                insertStmt.setObject(2, timeStamp);
                insertStmt.setObject(3, blahdRecord.getUserDN());
                // Handle cases when userFQAN is not provided
                // a) "userFQAN=", "userFQAN=NULL", "userFQAN=null"
                // b) no key=value pair for userFQAN in record
                if (fqan == null || fqan.equals("") || fqan.equals("null")) {
                    insertStmt.setObject(4, ApelDB.NULL);
                } else {
                    insertStmt.setObject(4, fqan);
                }
                insertStmt.setObject(5, ceId);
                insertStmt.setObject(6, jobId);
                insertStmt.setObject(7, lrmsId);
                insertStmt.setObject(8, blahdRecord.getJobName());
                insertStmt.setObject(9, m_siteName);
                insertStmt.setObject(10, blahdRecord.getValidFrom());
                insertStmt.setObject(11, blahdRecord.getValidUntil());
                insertStmt.setObject(12, UNPROCESSED);
                insertStmt.execute();
                insertCount++;
            } catch (SQLException e) {
                if (!isDuplicate(e)) {
                    throw new ApelException(e);
                }
            }
        }
        close(insertStmt);
        return insertCount;
    }

    /**
     * Updates the message table setting the processed flag where each 
     * result row id maps to a row found in the table.
     */
    private void updateProcessedMsgRecords(ResultSet joinResult)
            throws ApelException {
        String update = "UPDATE " + MSG_TABLE_NAME + " SET Processed = "
                + PROCESSED + " WHERE MsgID = ?";
        try {
            PreparedStatement updateStmt =
                    m_dbConnection.prepareStatement(update);
            joinResult.beforeFirst();
            while (joinResult.next()) {
                updateStmt.clearParameters();
                updateStmt.setObject(1, joinResult.getString("MsgID"));
                updateStmt.execute();
            }
            close(updateStmt);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
    }

    /**
     * Deletes all message records with set processed flag.
     */
    public void deleteProcessedMsgRecords() throws ApelException {
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            int numDels = stmt.executeUpdate("DELETE FROM " + MSG_TABLE_NAME
                    + " WHERE Processed = " + PROCESSED);
            Log.info(m_log, "Number deleted processed logs from: "
                    + MSG_TABLE_NAME + ": " + numDels);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
    }

    /**
     * Adds all records from the system message log file to the message table.
     * 
     * @param msgLogFile message log file to be inserted into db. The 
     * file must be parsed before calling this method.
     */
    public int insertMsgRecords(MsgRecordIterator msgRecordIt)
            throws ApelException {
        Utils utils = Utils.getInstance();
        int insertCount = 0;
        String insert = "REPLACE INTO " + MSG_TABLE_NAME + " VALUES "
                + "(?,?,?,?,?,?,?)";
        PreparedStatement insertStmt = null;
        try {
            insertStmt = m_dbConnection.prepareStatement(insert);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        while (msgRecordIt.hasNext()) {
            try {
                MsgRecord msgRec = msgRecordIt.next();
                String measurementDate = msgRec.getMeasurementDate();
                String measurementTime = msgRec.getMeasurementTime();
                String recordIdentity = measurementDate + " "
                        + measurementTime + " " + msgRec.getGramScriptJobID()
                        + " " + m_siteName;
                String validUntil = utils.calculateValidUntil(measurementDate);
                String validFrom = utils.calculateValidFrom(measurementDate);
                insertStmt.clearParameters();
                insertStmt.setObject(1, recordIdentity);
                insertStmt.setObject(2, msgRec.getGramScriptJobID());
                insertStmt.setObject(3, msgRec.getJobName());
                insertStmt.setObject(4, UNPROCESSED);
                insertStmt.setObject(5, m_siteName);
                insertStmt.setObject(6, validFrom);
                insertStmt.setObject(7, validUntil);
                insertStmt.execute();
                insertCount++;
            } catch (SQLException e) {
                if (!isDuplicate(e)) {
                    throw new ApelException(e);
                }
            }
        }
        close(insertStmt);
        return insertCount;
    }

    /**
     * If the job manager type is vanila PBS, a short cut exists whereby
     * the GramScriptID is the same as the JobName. Use gkRecords.isLcgPbs()
     * to test this condition before calling this method. 
     */
    public int insertMsgRecords(GkRecordIterator gkRecordIt) throws ApelException {
        Utils utils = Utils.getInstance();
        int insertCount = 0;
        String insert = "REPLACE INTO " + MSG_TABLE_NAME + " VALUES "
                + "(?,?,?,?,?,?,?)";
        PreparedStatement insertStmt = null;
        try {
            insertStmt = m_dbConnection.prepareStatement(insert);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        while (gkRecordIt.hasNext()) {
            try {
                GkRecord gkRecord = gkRecordIt.next();
                if (gkRecord.isLcgPbs()) {
                    continue;
                }
                String gramId = gkRecord.getGramScriptJobID();
                String measurementDate = gkRecord.getMeasurementDate();
                String measurementTime = gkRecord.getMeasurementTime();
                String recordIdentity = measurementDate + " "
                        + measurementTime + " " + gkRecord.getGramScriptJobID()
                        + " " + m_siteName;
                String validUntil = utils.calculateValidUntil(measurementDate);
                String validFrom = utils.calculateValidFrom(measurementDate);
                // Note the JobName = GramId
                insertStmt.clearParameters();
                insertStmt.setObject(1, recordIdentity);
                insertStmt.setObject(2, gramId);
                insertStmt.setObject(3, gramId);
                insertStmt.setObject(4, UNPROCESSED);
                insertStmt.setObject(5, m_siteName);
                insertStmt.setObject(6, validFrom);
                insertStmt.setObject(7, validUntil);
                insertStmt.execute();
                insertCount++;
            } catch (SQLException e) {
                if (!isDuplicate(e)) {
                    throw new ApelException(e);
                }
            }
        }
        close(insertStmt);
        return insertCount;
    }

    /**
     * Deletes all records with matching filetype from the processed file
     * table. It is meant specifically as an easy way of allowing complete 
     * data processing from scratch.
     *
     * @param fileType The file type to be deleted ('E' : Pbs log file,
     * 'G' : Gk log file, 'M' : Message log file).
     * @throws ApelException If an error occurs while connecting to the
     * database.
     */
    public void deleteAllFiles(char fileType) throws ApelException {
        Statement stmt = null;
        Log.info(m_log, "Updating file type: " + fileType);
        try {
            stmt = m_dbConnection.createStatement();
            int count = stmt.executeUpdate("DELETE FROM "
                    + PROCESSED_FILES_TABLE_NAME + " "
                    + "WHERE FileType = '" + fileType + "'");
            Log.info(m_log, "Number of file records deleted: " + count);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        close(stmt);
    }

    /**
     * Adds a filename to the processed files table. This will be added
     * to a batch of records that will be inserted when executeBatch() is
     * called or when a maximum limit of pending inserts is reached.
     *
     * @param fileName The base name of the file.
     * @param fileType The file type (eg event, gk or message file).
     * @param fullPath The full path of the file.
     * @throws ApelException If an error occurs while carrying out the batch
     * insert.
     */
    public void insertProcessedFile(String fileName, char fileType,
            String fullPath) throws ApelException {
        String hostName = getHostName();

        //try {
        //    hostName = InetAddress.getLocalHost().getCanonicalHostName();
        //} catch (UnknownHostException e) {
        //     throw new ApelException("Unable to determine local host name", e);
        //}

        String insert = "INSERT INTO " + PROCESSED_FILES_TABLE_NAME
                + " VALUES (?,?,?,?,?)";
        PreparedStatement insertStmt = null;
        try {
            insertStmt = m_dbConnection.prepareStatement(insert);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        try {
            insertStmt.clearParameters();
            insertStmt.setObject(1, fileName);
            insertStmt.setObject(2, "" + fileType);
            insertStmt.setObject(3, m_siteName);
            insertStmt.setObject(4, fullPath);
            insertStmt.setObject(5, hostName);
            insertStmt.execute();
            
        } catch (SQLException e) {
            if (!isDuplicate(e)) {
                throw new ApelException(e);
            }
        }
    }

    /**
     * Get the Local HostName
     *
     * @return Local HostName
     * @throws ApelException If unable to determine the local HostName
     */
    public String getHostName() throws ApelException {
        String hostName = null;
        try {
            hostName = InetAddress.getLocalHost().getCanonicalHostName();
        } catch (UnknownHostException e) {
            throw new ApelException("Unable to determine local host name", e);
        }
        return hostName;
    }

    /**
     * Checks if the file exists in the processed files table.
     *
     * @param fileName The base name of the file.
     * @param fileType The file type to check (eg event, gk or message).
     * @return True if it has an entry in the processed files table, false
     * otherwise.
     * @throws ApelException If an error occurs while connecting to the
     * database.
     */
    public boolean isFileProcessed(String fileName, char fileType)
            throws ApelException {
        Statement stmt = null;
        boolean isProcessed = false;
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT * FROM "
                    + PROCESSED_FILES_TABLE_NAME + " WHERE SiteName = '"
                    + m_siteName + "' AND FileName = '"
                    + fileName + "' AND FileType = '" + fileType + "'");
            isProcessed = rs.next();
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        close(stmt);
        return isProcessed;
    }

    /**
     * Checks if the file exists in the processed files table.
     * Sites running multiple CEs configurations would have the same fileNames and filePaths.
     * Therfore, the local hostName is used to distinguish them.
     * Examples: Two CEs, two batch systems have identical EventLog fileNames and filePaths
     * Two CEs, one batch system may have identical GK/accounting log fileNames and filePaths.
     * @param filePath The FullyQualified name of file (includes path)
     * @param hostName The local HostName
     * @param fileType The file type to check (eg event, gk or message).
     * @return True if it has an entry in the processed files table, false
     * otherwise.
     * @throws ApelException If an error occurs while connecting to the
     * database.
     */
    public boolean isFileProcessed(String filePath, String hostName, char fileType)
            throws ApelException {
        Statement stmt = null;
        boolean isProcessed = false;
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT * FROM "
                    + PROCESSED_FILES_TABLE_NAME + " WHERE SiteName = '"
                    + m_siteName + "' AND FilePath = '"
                    + filePath + "' AND HostName = '" + hostName + "' AND FileType = '"
                    + fileType + "'");
            isProcessed = rs.next();
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        close(stmt);
        return isProcessed;
    }

    /**
     * Attempts to generate new accounting records, inserting them
     * within the accounting record table. Returns an objectified view
     * of the newly inserted records.
     */
    public ArrayList generateAccountRecords(
            boolean ignoreProcessedFlag, int limit, int offset) throws ApelException {
        ResultSet result = joinAccountingRecords(ignoreProcessedFlag, limit, offset);
        insertAccountingRecords(result);
        updateProcessedEventRecords(result);
        updateProcessedGkRecords(result);
        updateProcessedMsgRecords(result);
        return convertToArrayList(result);
    }

    /**
     * gLite CE
     * Attempts to generate new accounting records, inserting them
     * within the accounting record table. Returns an objectified view
     * of the newly inserted records.
     */
    public ArrayList generateAccountRecordsGliteCE(boolean ignoreProcessedFlag, int limit, int offset) throws ApelException {

        ResultSet result = joinAccountingRecordsGliteCE(ignoreProcessedFlag, limit, offset);
        insertAccountingRecordsGliteCE(result);
        updateProcessedEventRecords(result);
        updateProcessedBlahdRecords(result);
        // Next convertJoinToGliteAccountingIt...
        return convertJoinToArrayListGliteCE(result);

    }

    /**
     * Inserts records into the main accounting record table. Returns the 
     * current running total of inserts carried out by this instance.
     */
    private void insertAccountingRecords(ResultSet rs) throws ApelException {
        try {
            if (rs.next()) {
                Log.info(m_log, "Storing accounting data into local database");
            } else {
                Log.info(m_log, "No accounting data to store");
                return;
            }
            String insert = "REPLACE INTO "
                    + LCG_RECORDS_TABLE_NAME + " VALUES "
                    + "(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
            PreparedStatement insertStmt =
                    m_dbConnection.prepareStatement(insert);
            UTCTime utcTime = new UTCTime();
            do {
                try {
                    String recordId = rs.getString("EventID");
                    insertStmt.clearParameters();
                    insertStmt.setObject(1, recordId);
                    insertStmt.setObject(2, rs.getString("SiteName"));
                    insertStmt.setObject(3, rs.getString("JobName"));
                    insertStmt.setObject(4, NULL);
                    insertStmt.setObject(5, rs.getString("LocalUserID"));
                    insertStmt.setObject(6, rs.getString("GlobalUserName"));
                    insertStmt.setObject(7, rs.getString("LocalUserGroup"));
                    insertStmt.setObject(8, rs.getString("WallDuration"));
                    insertStmt.setObject(9, rs.getString("CpuDuration"));
                    insertStmt.setObject(10, rs.getObject("WallDurationSeconds"));
                    insertStmt.setObject(11, rs.getObject("CpuDurationSeconds"));
                    insertStmt.setObject(12, rs.getString("StartTime"));
                    insertStmt.setObject(13, rs.getString("StopTime"));
                    insertStmt.setObject(14, rs.getString("StartTimeUTC"));
                    insertStmt.setObject(15, rs.getString("StopTimeUTC"));
                    insertStmt.setObject(16, rs.getObject("StartTimeEpoch"));
                    insertStmt.setObject(17, rs.getObject("StopTimeEpoch"));
                    insertStmt.setObject(18, rs.getString("SubmitHost"));
                    insertStmt.setObject(19, rs.getObject("MemoryReal"));
                    insertStmt.setObject(20, rs.getObject("MemoryVirtual"));
                    insertStmt.setObject(21, rs.getObject("SpecInt2000"));
                    insertStmt.setObject(22, rs.getObject("SpecFloat2000"));
                    insertStmt.setObject(23, rs.getString("EventDate"));
                    insertStmt.setObject(24, rs.getString("EventTime"));
                    insertStmt.setObject(25, utcTime.getDate());
                    insertStmt.setObject(26, utcTime.getTime());
                    insertStmt.execute();
                } catch (SQLException e) {
                    if (!isDuplicate(e)) {
                        throw new ApelException(e);
                    }
                }
            } while (rs.next());
            close(insertStmt);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
    }

    /**
     * GliteCE NB: Main differences between accounting records via gLiteCE and LCG CE are:-
     * Column 1 : RecordIdentity does not have siteName
     * Column 3. BLAHD:LocalJobId instead of EVENT:JobName, they should be the same as its a requirments of the join
     * Column 4. LCGJobID is the RB LB Job ID via Blahd Log file
     * Column 6. GlobalUserName is a BASE64 encrypted string 
     * Column 7. LocalUserGroup is the VOMS userFQAN (default is local unix group defined in gridmap file via EVENT Table)
     * Column 18. ExecutingCE taken from BLAHD:RecordIdentity instead of SubmitHost because it contains the CE and Queue, and is unambiguous.
     * Inserts records into the main accounting record table. Returns the
     * current running total of inserts carried out by this instance.
     */
    private void insertAccountingRecordsGliteCE(ResultSet rs) throws ApelException {
        try {
            if (rs.next()) {
                Log.info(m_log, "Storing accounting data into local database");
            } else {
                Log.info(m_log, "No accounting data to store");
                return;
            }
            String insert = "REPLACE INTO "
                    + LCG_RECORDS_TABLE_NAME + " VALUES "
                    + "(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
            PreparedStatement insertStmt =
                    m_dbConnection.prepareStatement(insert);
            UTCTime utcTime = new UTCTime();
            do {
                try {
                    String recordId = rs.getString("BlahdID");
                    insertStmt.clearParameters();
                    insertStmt.setObject(1, recordId);
                    insertStmt.setObject(2, rs.getString("SiteName"));
                    insertStmt.setObject(3, rs.getString("JobName"));
                    insertStmt.setObject(4, rs.getString("GlobalJobId"));
                    insertStmt.setObject(5, rs.getString("LocalUserID"));
                    insertStmt.setObject(6, rs.getString("GlobalUserName"));
                    if (rs.getString("FullyQualifiedAttributeName").equals(ApelDB.NULL)) {
                        insertStmt.setObject(7, rs.getString("LocalUserGroup"));
                    } else {
                        insertStmt.setObject(7, rs.getString("FullyQualifiedAttributeName"));
                    }
                    insertStmt.setObject(8, rs.getString("WallDuration"));
                    insertStmt.setObject(9, rs.getString("CpuDuration"));
                    insertStmt.setObject(10, rs.getObject("WallDurationSeconds"));
                    insertStmt.setObject(11, rs.getObject("CpuDurationSeconds"));
                    insertStmt.setObject(12, rs.getString("StartTime"));
                    insertStmt.setObject(13, rs.getString("StopTime"));
                    insertStmt.setObject(14, rs.getString("StartTimeUTC"));
                    insertStmt.setObject(15, rs.getString("StopTimeUTC"));
                    insertStmt.setObject(16, rs.getObject("StartTimeEpoch"));
                    insertStmt.setObject(17, rs.getObject("StopTimeEpoch"));
                    insertStmt.setObject(18, rs.getString("ResourceIdentity"));
                    insertStmt.setObject(19, rs.getObject("MemoryReal"));
                    insertStmt.setObject(20, rs.getObject("MemoryVirtual"));
                    insertStmt.setObject(21, rs.getObject("SpecInt2000"));
                    insertStmt.setObject(22, rs.getObject("SpecFloat2000"));
                    insertStmt.setObject(23, rs.getString("EventDate"));
                    insertStmt.setObject(24, rs.getString("EventTime"));
                    insertStmt.setObject(25, utcTime.getDate());
                    insertStmt.setObject(26, utcTime.getTime());
                    insertStmt.execute();
                } catch (SQLException e) {
                    if (!isDuplicate(e)) {
                        throw new ApelException(e);
                    }
                }
            } while (rs.next());
            close(insertStmt);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
    }

    /**
     * Returns a count of records available after the date and time period.
     * This code does not require m_siteName to be specified in the WHERE clause.
     * 
     */
    public int getNumAccountingRecords(String date, String time)
            throws ApelException {
        Statement stmt = null;
        int count = 0;
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT COUNT(*) FROM "
                    + LCG_RECORDS_TABLE_NAME + " WHERE ((MeasurementDate = '" + date
                    + "' AND MeasurementTime >= '" + time + "') OR "
                    + "MeasurementDate > '" + date + "')");
            rs.next();
            count = rs.getInt(1);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
        return count;
    }

    /**
     * Returns a count of records available after the date and time period for a given siteName.
     *
     */
    public int getNumAccountingRecords(String date, String time, String siteName)
            throws ApelException {
        Statement stmt = null;
        int count = 0;
        try {
            stmt = m_dbConnection.createStatement();
            String sql_query = "SELECT COUNT(*) FROM "
                    + LCG_RECORDS_TABLE_NAME
                    + " WHERE ((MeasurementDate = '" + date
                    + "' AND MeasurementTime >= '" + time + "') OR "
                    + "MeasurementDate > '" + date + "') AND ExecutingSite like '"
                    + siteName + "' AND " + m_VOFilter;
            ResultSet rs = stmt.executeQuery(sql_query);
            rs.next();
            count = rs.getInt(1);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
        return count;
    }

    /**
     * DaveKant
     * Returns a count of records available in the gap period
     * EventDate is used in the query, not MeasurementDate!
     * Gaps in the accounting record are determined from visual inspections of Gantt charts 
     * via the GOC accounting and reporting web pages
     * EventDate is the actual date of the job Record whereas MeasurementDate is internal to RGMA  
     * which can be modified by functions like updateRepublishTimestamp()
     */
    public int getNumAccountingRecordsInGap(String date_start, String date_end)
            throws ApelException {
        Statement stmt = null;
        int count = 0;
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT COUNT(*) FROM "
                    + LCG_RECORDS_TABLE_NAME + " WHERE EventDate >= '" + date_start + "' AND "
                    + "EventDate <= '" + date_end + "'");
            rs.next();
            count = rs.getInt(1);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
        return count;
    }

    /**
     * Returns a batch of accounting records that match or exceed the time stamp.
     * This code does not require m_siteName to be specified in the WHERE clause.
     * It simply finds all the records!
     * Cristina del Cano Novales 2008-06-04
     * Bug fix #34506 Added limit and offset to query to avoid OutOfMemory error
     */
    public ArrayList getAccountingRecords(String date, String time, int limit, int offset)
            throws ApelException {
        ResultSet rs = null;
        try {
            Statement stmt = m_dbConnection.createStatement();
            rs = stmt.executeQuery("SELECT * FROM "
                    + LCG_RECORDS_TABLE_NAME + " WHERE ((MeasurementDate = '" + date
                    + "' AND MeasurementTime >= '" + time + "') OR "
                    + "MeasurementDate > '" + date + "') LIMIT " + offset + "," + limit);
            
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }

        // Query is against the LcgRecoprd table which conmatins BOTH LCGCE AND GliteCE data
        return convertToAccounting(rs);
    }

    /**
     * Returns a batch of accounting records that match or exceed the time stamp and siteName.
     * Cristina del Cano Novales 2008-06-04
     * Bug fix #34506 Added limit and offset to query to avoid OutOfMemory error
     */
    public ArrayList getAccountingRecords(String date, String time, String siteName, int limit, int offset)
            throws ApelException {
        ResultSet rs = null;
        try {
            Statement stmt = m_dbConnection.createStatement();
            rs = stmt.executeQuery("SELECT * FROM "
                    + ApelDB.LCG_RECORDS_TABLE_NAME
                    + " WHERE ((MeasurementDate = '" + date
                    + "' AND MeasurementTime >= '" + time + "') OR "
                    + "MeasurementDate > '" + date + "') AND ExecutingSite like '"
                    + siteName + "' AND " + m_VOFilter + " LIMIT " + offset + "," + limit);
            
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        // Query is against the LcgRecoprd table which conmatins BOTH LCGCE AND GliteCE data
        return convertToAccounting(rs);
    }
    /**
     * DaveKant
     * Returns a batch of accounting records that match the gap interval
     * EventDate is used in the query, not MeasurementDate!
     * Gaps in the accounting record are determined from visual inspections of Gantt charts
     * via the GOC accounting and reporting web pages
     * EventDate is the actual date of the job Record whereas MeasurementDate is internal to RGMA
     * which can be modified by functions like updateRepublishTimestamp()
     * 
     * Cristina del Cano Novales 2008-06-04
     * Bug fix #34506 Added limit and offset to query to avoid OutOfMemory error
     */
     public ArrayList getAccountingRecordsInGap(String date_start, String date_end, int limit, int offset)
            throws ApelException {
        ResultSet rs = null;
        try {
            Statement stmt = m_dbConnection.createStatement();
            rs = stmt.executeQuery("SELECT * FROM "
                    + LCG_RECORDS_TABLE_NAME + " WHERE EventDate >= '" + date_start + "' AND "
                    + "EventDate <= '" + date_end + "' AND " + m_VOFilter + " LIMIT " + offset + "," + limit);
            
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }

        return convertToAccounting(rs);
    }

    /**
     * Returns a batch of summary accounting records.
     */
    public AccountSummaryIterator getAccountingSummary(String siteName)
            throws ApelException {
        ResultSet rs = null;
        try {
            Statement stmt = m_dbConnection.createStatement();
            rs = stmt.executeQuery("SELECT ExecutingSite, MIN(EventDate) as RecordStart, "
                    + "MAX(EventDate) as RecordEnd, count(*) as Njobs, "
                    + "count(distinct(EventDate)) as Ndays, "
                    + // Some sites have sums which exceed 2^32
                    //"SUM(ElapsedTimeSeconds) as ElapsedTimeSeconds, " +
                    //"SUM(BaseCpuTimeSeconds) as BaseCpuTimeSeconds, " +
                    "YEAR(EventDate) as YEAR, MONTH(EventDate) as MONTH "
                    + "FROM " + LCG_RECORDS_TABLE_NAME + " WHERE ExecutingSite like '" + siteName + "' "
                    + " AND " + m_VOFilter + " GROUP BY YEAR(EventDate), MONTH(EventDate)");
            
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        // Query is against the LcgRecords table which conmatins BOTH LCGCE AND GliteCE data
        return convertToAccountingSummaryIt(rs);
    }


    /**
     * Converts the query set into an accounting summary set with a useful iterator.
     */
    private AccountSummaryIterator convertToAccountingSummaryIt(ResultSet rs) throws ApelException {
        ArrayList records = new ArrayList();
        try {
            while (rs.next()) {
                AccountSummary record = new AccountSummary();
                String RecordIdentity = ApelDB.NULL;
                RecordIdentity = rs.getString("ExecutingSite") + "-" + rs.getString("YEAR") + "-" + rs.getString("MONTH");
                record.setRecordId(RecordIdentity);
                record.setExecutingSite(rs.getString("ExecutingSite"));
                record.setNjobs(rs.getString("Njobs"));
                record.setNdays(rs.getString("Ndays"));
                //record.setElapsedTimeSeconds(rs.getString("ElapsedTimeSeconds"));
                //record.setBaseCpuTimeSeconds(rs.getString("BaseCpuTimeSeconds"));
                record.setRecordStart(rs.getString("RecordStart"));
                record.setRecordEnd(rs.getString("RecordEnd"));
                records.add(record);
            }
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        return new AccountSummaryIterator(records);
    }

    /**
     * Returns all records stored in the accounting table.
     */
    public ArrayList getAllAccountingRecords(int limit, int offset)
            throws ApelException {
        ArrayList it = null;
        try {
            Statement stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT * FROM " + LCG_RECORDS_TABLE_NAME + " WHERE " + m_VOFilter + " LIMIT " + offset + "," + limit);
            it = this.convertToAccounting(rs);
            
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        return it;
    }
    /**
     * Updates all the MeasurementDate/Time columns of the accounting
     * table to contain the current time. This is used when a complete 
     * reprocess is carried out and ensures when the republish option is set 
     * to 'missing', an accurate audit can be made of how many records were 
     * sent during the last join process.
     */
    public void updateAccountingTimeStamp() throws ApelException {
        UTCTime utcTime = new UTCTime();
        try {
            Statement stmt = m_dbConnection.createStatement();
            stmt.execute("UPDATE " + LCG_RECORDS_TABLE_NAME + " SET "
                    + "MeasurementDate = '" + utcTime.getDate() + "', "
                    + "MeasurementTime = '" + utcTime.getTime() + "'");
            close(stmt);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
    }

    /**
     * Updates the MeasurementDate/Time columns of the accounting
     * table to contain the current time for data encapsulated in an AccountRecordIterator.
     * Example: Updating data selected by the GAP publisher.
     */
    public void updateAccountingTimeStamp(ArrayList arIt) throws ApelException {
        //Log.info(m_log, "  updateAccountingTimeStamp(AccountRecordIterator) : ");
        UTCTime utcTime = new UTCTime();
        try {
            int i = 0;
            while (i < arIt.size()) {
                AccountRecord record = (AccountRecord) arIt.get(i);
                Statement stmt = m_dbConnection.createStatement();
                String query = "UPDATE " + ApelDB.LCG_RECORDS_TABLE_NAME + " SET "
                        + "MeasurementDate = '" + utcTime.getDate() + "', "
                        + "MeasurementTime = '" + utcTime.getTime() + "' "
                        + "WHERE RecordIdentity = '" + record.getRecordId() + "'";
                //Log.info(m_log, "query=" + query);
                stmt.execute(query);
                close(stmt);
                i++;
            }
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
    }

    /**
     * Updates the republish date/time (represented in UTC) to the current
     * time.
     */
    public void updateRepublishTimeStamp() throws ApelException {
        UTCTime timestamp = new UTCTime();
        deleteAllRecords(REPUBLISH_TABLE_NAME);
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            stmt.execute("INSERT INTO " + REPUBLISH_TABLE_NAME + " VALUES "
                    + "('" + timestamp.getDate() + "', '" + timestamp.getTime()
                    + "', '')");
            close(stmt);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
    }

    /**
     * Returns true only if a republish time stamp exists within the
     * republish table. 
     */
    public boolean containsRepublishTimeStamp() throws ApelException {
        Statement stmt = null;
        boolean containsTimeStamp = false;
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT * FROM "
                    + REPUBLISH_TABLE_NAME);
            containsTimeStamp = rs.next();
            close(stmt);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        return containsTimeStamp;
    }

    /**
     * Initialises the republish date/time.
     */
    public UTCTime setupTimeStamp() throws ApelException {
        UTCTime timestamp = new UTCTime();
        Statement stmt = null;
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT * FROM "
                    + REPUBLISH_TABLE_NAME);
            if (rs.next()) {
                timestamp.setDate(rs.getString(1));
                timestamp.setTime(rs.getString(2));
            }
            close(stmt);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        return timestamp;
    }

    /**
     * Determine how the SpecRecords table will be updated according to the specInt and ClusterId information
     *  - insert a new record if no ClusterID exists
     *  - terminate existing specInt if ClusterID exists but specInt is a new value
     *  - don't do anything if ClusterId exists and SpecInt is currently active
     */
    public void processSpecRecord(String clusterId, String subClusterId, String specInt, String specFloat, String date, String time) throws ApelException {

        ////////////////////////////////////////////////////
        // Checks if the clusterId is present in the table.
        ////////////////////////////////////////////////////
        String recordIdentity = null;
        if (isClusterInTable(clusterId)) {
            // Log.info(m_log, "Found ClusterId in Table, now checking if SpecInt is latest");
            // ////////////////////////////////////////////////////////
            // Checks if the SpecInt associated with clusterId exists
            // ////////////////////////////////////////////////////////
            if (!isSpecInTableLatest(clusterId, specInt)) {
                Log.info(m_log, "Found New ClusterID:SpecInt" + clusterId + ":" + specInt);
                // get SpecID for the current Active SpecInt associated to ClusterId
                String SpecID = getActiveSpecID(clusterId);
                // terminate
                terminateSpecID(SpecID, date, time);
                // insert new record
                recordIdentity = date + " " + time + " " + m_siteName + " " + clusterId;
                insertSpecRecord(recordIdentity, clusterId, subClusterId, specInt, specFloat, null, null, date, time);
            } else {
                Log.info(m_log, "The ClusterId : SpecInt pair are latest, no update required " + clusterId + ":" + specInt);
            }
        } else {
            // Add new records to database
            Log.info(m_log, "Found New ClusterId, now storing it");
            // The default record for a new ClusterId.
            // NB The default record simulates discovery of a new ClusterId at Epoch=0
            recordIdentity = DEFAULT_SPEC_DATE + " " + DEFAULT_SPEC_TIME + " " + m_siteName + " " + clusterId;
            insertSpecRecord(recordIdentity, clusterId, subClusterId, specInt, specFloat, date, time, DEFAULT_SPEC_DATE, DEFAULT_SPEC_TIME);           // The active record
            recordIdentity = date + " " + time + " " + m_siteName + " " + clusterId;
            insertSpecRecord(recordIdentity, clusterId, subClusterId, specInt, specFloat, null, null, date, time);
        }
    }

    /**
     * Identify the current active record for a given clusterId and return its unique recordIdentity.
     */
    private String getActiveSpecID(String clusterId) throws ApelException {
        String query = "SELECT SpecID FROM " + SPEC_TABLE_NAME
                + " WHERE ClusterID='" + clusterId + "' AND EndDate IS " + ApelDB.NULL + " AND EndTime IS " + ApelDB.NULL;
        Log.info(m_log, "getActiveSpecId query = " + query);
        String specId = null;
        try {
            Statement stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery(query);
            if (rs.next()) {
                specId = rs.getString("SpecID");
            }
            close(stmt);
            rs.close();
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        return specId;
    }

    /**
     * Terminate a SpecID by setting EndDate = MeasurementDate and EndTime = MeasurementTime
     */
    private void terminateSpecID(String specId, String measurementDate, String measurementTime) throws ApelException {
        String query = "UPDATE " + SPEC_TABLE_NAME
                + " SET EndDate='" + measurementDate + "',EndTime='" + measurementTime + "'"
                + " WHERE SpecID='" + specId + "'";
        Log.info(m_log, "terminateSpecID query = " + query);
        try {
            Statement stmt = m_dbConnection.createStatement();
            // NB: Can not issue data manipulation statements with executeQuery() method
            stmt.execute(query);
            close(stmt);
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
    }

    /**
     * Attempts to insert a new spec record.           
     */
    private void insertSpecRecord(String recordId, String clusterId,
            String subClusterId, String specInt, String specFloat, String endDate,
            String endTime, String measurementDate, String measurementTime)
            throws ApelException {
        String insert = "INSERT INTO " + SPEC_TABLE_NAME + " VALUES "
                + "(?,?,?,?,?,?,?,?,?,?)";
        PreparedStatement insertStmt = null;
        try {
            insertStmt = m_dbConnection.prepareStatement(insert);
        } catch (SQLException e) {
            throw new ApelException(e);
        }
        try {
            insertStmt.clearParameters();
            insertStmt.setObject(1, recordId);
            insertStmt.setObject(2, m_siteName);
            insertStmt.setObject(3, clusterId);
            insertStmt.setObject(4, subClusterId);
            insertStmt.setObject(5, specInt);
            insertStmt.setObject(6, specFloat);
            if (endDate == null) {
                insertStmt.setNull(7, Types.DATE);
                insertStmt.setNull(8, Types.TIME);
            } else {
                insertStmt.setObject(7, endDate);
                insertStmt.setObject(8, endTime);
            }
            insertStmt.setObject(9, measurementDate);
            insertStmt.setObject(10, measurementTime);
            insertStmt.execute();
        } catch (SQLException e) {
            if (!isDuplicate(e)) {
                throw new ApelException(e);
            }
        }
        close(insertStmt);
    }

    /**
     * Boolean to check if the clusterId exists in database.
     * return false if ClusterId not in SpecTable
     */
    private boolean isClusterInTable(String clusterId) throws ApelException {
        boolean isEmpty = false;
        String query = "SELECT SpecID FROM " + SPEC_TABLE_NAME
                + " WHERE SiteName = '" + m_siteName + "' AND ClusterID = '" + clusterId + "'";
        //Log.info(m_log, "isClusterInTable query" + query);
        try {
            Statement stmt = m_dbConnection.createStatement();
            ResultSet rs = stmt.executeQuery(query);
            isEmpty = rs.next();
            close(stmt);
            rs.close();
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        //Log.info(m_log, "isClusterInTable ? " + clusterId + "=" + isEmpty);
        return isEmpty;
    }

    /**
     * Boolean to check if the SpecInt value associated with clusterId is the latest
     */
    private boolean isSpecInTableLatest(String clusterId, String specInt) throws ApelException {
        boolean isLatest = false;
        String query = "SELECT SpecID FROM " + SPEC_TABLE_NAME
                + " WHERE SiteName = '" + m_siteName + "' AND ClusterID = '" + clusterId
                + "' AND SpecInt2000 = '" + specInt + "' AND EndDate IS " + ApelDB.NULL;
        //Log.info(m_log, "isSpecInTableLatest query" + query);
        try {
            Statement stmt = m_dbConnection.createStatement();
            ResultSet isSpecrs = stmt.executeQuery(query);
            isLatest = isSpecrs.next();
            close(stmt);
            isSpecrs.close();
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        //Log.info(m_log, "isSpecInTableLatest ? " + specInt + "=" + isLatest);
        return isLatest;
    }

    /**
     * Attempts to join all accounting records using the gk, message
     * and event tables.
     */
    private ResultSet joinAccountingRecords(boolean republishAll, int limit, int offset)
            throws ApelException {
        Statement stmt = null;
        ResultSet rs = null;
        try {
            Log.info(m_log, "Stitching together all accounting records");
            String query = buildJoinQuery(republishAll, limit, offset);
            m_log.debug(query);
            stmt = m_dbConnection.createStatement();
            rs = stmt.executeQuery(query);
            Log.info(m_log, "Stitching completed");
        } catch (SQLException e) {
            // Bug fix #29962
            throw new ApelException(e.getMessage() + "Check that /tmp directory is not full");
        }
        return rs;
    }

    /**
     * Attempts to join all accounting records using the blahd
     * and event tables.
     */
    private ResultSet joinAccountingRecordsGliteCE(boolean republishAll, int limit, int offset)
            throws ApelException {
        Statement stmt = null;
        ResultSet joinGliteCErs = null;
        try {
            Log.info(m_log, "Stitching together all accounting records");
            String query = buildJoinQueryGliteCE(republishAll, limit, offset);
            m_log.info(query);
            stmt = m_dbConnection.createStatement();
            joinGliteCErs = stmt.executeQuery(query);
            Log.info(m_log, "Stitching completed");
        } catch (SQLException e) {
            // Bug fix #29962
            throw new ApelException(e.getMessage() + "Check that /tmp directory is not full");
        }
        return joinGliteCErs;
    }

    /**
     * Builds the SQL join query used to stich together all accounting 
     * records from LCG CE
     */
    private String buildJoinQuery(boolean republishAll, int limit, int offset) {
        String query = "SELECT " + getJoinColumns()
                + "FROM " + EVENT_TABLE_NAME + ", "
                + MSG_TABLE_NAME + ", "
                + GK_TABLE_NAME + ", "
                + SPEC_TABLE_NAME + " "
                + "WHERE " + getJoinPredicate(republishAll) + " LIMIT " + offset + "," + limit;
        m_log.debug("GK Based Join query: " + query);
        return query;
    }

    /**
     * Builds the SQL join query used to stich together all accounting
     * records from gLite CE
     */
    private String buildJoinQueryGliteCE(boolean republishAll, int limit, int offset) {
        String query = "SELECT " + getJoinColumnsGliteCE()
                + "FROM " + EVENT_TABLE_NAME + ", "
                + BLAHD_TABLE_NAME + ", "
                + SPEC_TABLE_NAME + " "
                + "WHERE " + getJoinPredicateGliteCE(republishAll) + " LIMIT " + offset + "," + limit;
        m_log.debug("Grid Accounting Mapping File: Join query: " + query);


        return query;
    }

    /**
     * Returns the join predicate. (for LCG CE)
     */
    private String getJoinPredicate(boolean republishAll) {
        String predicate =
                EVENT_TABLE_NAME + ".JobName = "
                + MSG_TABLE_NAME + ".JobName AND "
                + MSG_TABLE_NAME + ".GramScriptJobID = "
                + GK_TABLE_NAME + ".GramScriptJobID AND "
                + getJoinSpecPredicate() + "AND ";
        String republishBit = EVENT_TABLE_NAME
                + ".Processed = 0 AND " + GK_TABLE_NAME
                + ".Processed = 0 AND " + MSG_TABLE_NAME
                + ".Processed = 0 AND ";
        if (republishAll) {
            republishBit = " ";
        }
        predicate += republishBit;
        predicate += EVENT_TABLE_NAME + ".EventDate >= "
                + GK_TABLE_NAME + ".ValidFrom AND " + EVENT_TABLE_NAME
                + ".EventDate >= " + MSG_TABLE_NAME + ".ValidFrom AND "
                + EVENT_TABLE_NAME + ".EventDate < " + GK_TABLE_NAME
                + ".ValidUntil AND " + EVENT_TABLE_NAME + ".EventDate < "
                + MSG_TABLE_NAME + ".ValidUntil AND " + GK_TABLE_NAME
                + ".SiteName LIKE '" + m_siteName + "' AND " + EVENT_TABLE_NAME
                + ".SiteName LIKE '" + m_siteName + "' AND " + SPEC_TABLE_NAME
                + ".SiteName LIKE '" + m_siteName + "' AND " + MSG_TABLE_NAME
                + ".SiteName LIKE '" + m_siteName + "' AND "
                + EVENT_TABLE_NAME + ".SubmitHost = " + SPEC_TABLE_NAME + ".ClusterID AND "
                + GK_TABLE_NAME + ".SiteName = " + MSG_TABLE_NAME + ".SiteName AND "
                + MSG_TABLE_NAME + ".SiteName = " + EVENT_TABLE_NAME + ".SiteName AND "
                + EVENT_TABLE_NAME + ".SiteName = " + SPEC_TABLE_NAME + ".SiteName"
                + // Remove multiple matches from specRecords - bug # 25430
                " GROUP BY " + EVENT_TABLE_NAME + ".EventID";
        return predicate;
    }

    /**
     * Returns the join predicate. (for gLite CE)
     */
    private String getJoinPredicateGliteCE(boolean republishAll) {
        String predicate =
                EVENT_TABLE_NAME + ".JobName = "
                + BLAHD_TABLE_NAME + ".JobName AND "
                + getJoinSpecPredicate() + "AND ";
        String republishBit = EVENT_TABLE_NAME
                + ".Processed = 0 AND " + BLAHD_TABLE_NAME
                + ".Processed = 0 AND ";
        if (republishAll) {
            republishBit = " ";
        }
        predicate += republishBit;
        predicate += EVENT_TABLE_NAME + ".EventDate >= " + BLAHD_TABLE_NAME + ".ValidFrom AND "
                + EVENT_TABLE_NAME + ".EventDate < " + BLAHD_TABLE_NAME + ".ValidUntil AND "
                + BLAHD_TABLE_NAME + ".SiteName LIKE '" + m_siteName + "' AND " + EVENT_TABLE_NAME
                + ".SiteName LIKE '" + m_siteName + "' AND " + SPEC_TABLE_NAME
                + ".SiteName LIKE '" + m_siteName + "' AND "
                + BLAHD_TABLE_NAME + ".ResourceIdentity RLIKE " + SPEC_TABLE_NAME + ".ClusterID AND "
                + BLAHD_TABLE_NAME + ".SiteName = " + EVENT_TABLE_NAME + ".SiteName AND "
                + EVENT_TABLE_NAME + ".SiteName = " + SPEC_TABLE_NAME + ".SiteName";
        return predicate;
    }

    /**
     * Returns a set of column names used in the join query. (for LCG CE)
     */
    private String getJoinColumns() {
        return EVENT_TABLE_NAME + ".EventID, "
                + EVENT_TABLE_NAME + ".SiteName, "
                + EVENT_TABLE_NAME + ".JobName, "
                + EVENT_TABLE_NAME + ".LocalUserID, "
                + EVENT_TABLE_NAME + ".LocalUserGroup, "
                + EVENT_TABLE_NAME + ".WallDuration, "
                + EVENT_TABLE_NAME + ".CpuDuration, "
                + EVENT_TABLE_NAME + ".WallDurationSeconds, "
                + EVENT_TABLE_NAME + ".CpuDurationSeconds, "
                + EVENT_TABLE_NAME + ".StartTime, "
                + EVENT_TABLE_NAME + ".StopTime, "
                + EVENT_TABLE_NAME + ".StartTimeUTC, "
                + EVENT_TABLE_NAME + ".StopTimeUTC, "
                + EVENT_TABLE_NAME + ".StartTimeEpoch, "
                + EVENT_TABLE_NAME + ".StopTimeEpoch, "
                + EVENT_TABLE_NAME + ".SubmitHost, "
                + EVENT_TABLE_NAME + ".MemoryReal, "
                + EVENT_TABLE_NAME + ".MemoryVirtual, "
                + EVENT_TABLE_NAME + ".EventDate, "
                + EVENT_TABLE_NAME + ".EventTime,"
                + GK_TABLE_NAME + ".GlobalUserName, "
                + GK_TABLE_NAME + ".GkID, "
                + SPEC_TABLE_NAME + ".SpecInt2000, "
                + SPEC_TABLE_NAME + ".SpecFloat2000, "
                + MSG_TABLE_NAME + ".MsgID ";
    }

    /**
     * Returns a set of column names used in the join query. (for gLite CE)
     */
    private String getJoinColumnsGliteCE() {
        return EVENT_TABLE_NAME + ".EventID, "
                + EVENT_TABLE_NAME + ".SiteName, "
                + EVENT_TABLE_NAME + ".JobName, "
                + EVENT_TABLE_NAME + ".LocalUserID, "
                + EVENT_TABLE_NAME + ".LocalUserGroup, "
                + EVENT_TABLE_NAME + ".WallDuration, "
                + EVENT_TABLE_NAME + ".CpuDuration, "
                + EVENT_TABLE_NAME + ".WallDurationSeconds, "
                + EVENT_TABLE_NAME + ".CpuDurationSeconds, "
                + EVENT_TABLE_NAME + ".StartTime, "
                + EVENT_TABLE_NAME + ".StopTime, "
                + EVENT_TABLE_NAME + ".StartTimeUTC, "
                + EVENT_TABLE_NAME + ".StopTimeUTC, "
                + EVENT_TABLE_NAME + ".StartTimeEpoch, "
                + EVENT_TABLE_NAME + ".StopTimeEpoch, "
                + EVENT_TABLE_NAME + ".SubmitHost, "
                + EVENT_TABLE_NAME + ".MemoryReal, "
                + EVENT_TABLE_NAME + ".MemoryVirtual, "
                + EVENT_TABLE_NAME + ".EventDate, "
                + EVENT_TABLE_NAME + ".EventTime,"
                + BLAHD_TABLE_NAME + ".BlahdID, "
                + BLAHD_TABLE_NAME + ".JobName, "
                + BLAHD_TABLE_NAME + ".LrmsId, "
                + BLAHD_TABLE_NAME + ".GlobalUserName, "
                + BLAHD_TABLE_NAME + ".FullyQualifiedAttributeName, "
                + BLAHD_TABLE_NAME + ".ResourceIdentity, "
                + BLAHD_TABLE_NAME + ".GlobalJobId, "
                + SPEC_TABLE_NAME + ".SpecInt2000, "
                + SPEC_TABLE_NAME + ".SpecFloat2000 ";
    }

    /**
     * Returns the predicate for matching each pbs record with
     * an appropriate cpu spec value.
     */
    private String getJoinSpecPredicate() {
        String predicate = "((("
                + EVENT_TABLE_NAME + ".EventDate = "
                + SPEC_TABLE_NAME + ".MeasurementDate AND "
                + EVENT_TABLE_NAME + ".EventTime >= "
                + SPEC_TABLE_NAME + ".MeasurementTime) OR "
                + EVENT_TABLE_NAME + ".EventDate > "
                + SPEC_TABLE_NAME + ".MeasurementDate) AND ("
                + SPEC_TABLE_NAME + ".EndDate IS NULL OR ("
                + EVENT_TABLE_NAME + ".EventDate = "
                + SPEC_TABLE_NAME + ".EndDate AND "
                + EVENT_TABLE_NAME + ".EventTime < "
                + SPEC_TABLE_NAME + ".EndTime) OR "
                + EVENT_TABLE_NAME + ".EventDate < "
                + SPEC_TABLE_NAME + ".EndDate))";
        return predicate;
    }

    /**
     * Tidies up any open resources.
     */
    public void close() {
        try {
            m_dbConnection.close();
        } catch (SQLException e) {
        }
    }

    /**
     * Attempts to close the Statement object. Any thrown exceptions are
     * quietly ignored.
     */
    private void close(Statement stmt) {
        if (stmt != null) {
            try {
                stmt.close();
            } catch (SQLException e) {
            }
        }
    }

    /**
     * Attempts to close the PreparedStatement object. Any thrown exceptions 
     * are quietly ignored.
     */
    private void close(PreparedStatement stmt) {
        if (stmt != null) {
            try {
                stmt.close();
            } catch (SQLException e) {
            }
        }
    }

    /**
     * Returns the date when the apel last published any accounting data.
     */
    public String getRepublishDate() throws ApelException {
        Statement stmt = null;
        String date = "";
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs =
                    stmt.executeQuery("SELECT * FROM " + REPUBLISH_TABLE_NAME);
            if (rs.next()) {
                date = rs.getString(1);
            }
            rs.close();
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
        return date;
    }

    /**
     * Returns the time when the apel last published any accounting data.
     */
    public String getRepublishTime() throws ApelException {
        Statement stmt = null;
        String time = "";
        try {
            stmt = m_dbConnection.createStatement();
            ResultSet rs =
                    stmt.executeQuery("SELECT * FROM " + REPUBLISH_TABLE_NAME);
            if (rs.next()) {
                time = rs.getString(2);
            }
            rs.close();
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        close(stmt);
        return time;
    }

    private ArrayList convertToArrayList(ResultSet rs) throws ApelException {
        ArrayList records = new ArrayList();
        try {
            rs.beforeFirst();
            while (rs.next()) {
                AccountRecord record = new AccountRecord();
                record.setRecordId(rs.getString("EventID"));
                record.setExecutingSite(rs.getString("SiteName"));
                record.setLocalJobId(rs.getString("JobName"));
                record.setLocalUserId(rs.getString("LocalUserID"));
                record.setLcgUserId(rs.getString("GlobalUserName"));
                record.setLocalUserVO(rs.getString("LocalUserGroup"));
                record.setElapsedTime(rs.getString("WallDuration"));
                record.setBaseCPUTime(rs.getString("CpuDuration"));
                record.setElapsedTimeSeconds(rs.getString("WallDurationSeconds"));
                record.setBaseCPUTimeSeconds(rs.getString("CpuDurationSeconds"));
                record.setStartTime(rs.getString("StartTime"));
                record.setStopTime(rs.getString("StopTime"));
                record.setStartTimeUtc(rs.getString("StartTimeUTC"));
                record.setStopTimeUtc(rs.getString("StopTimeUTC"));
                record.setStartTimeEpoch(rs.getString("StartTimeEpoch"));
                record.setStopTimeEpoch(rs.getString("StopTimeEpoch"));
                record.setExecutingCE(rs.getString("SubmitHost"));
                record.setMemoryReal(rs.getString("MemoryReal"));
                record.setMemoryVirtual(rs.getString("MemoryVirtual"));
                record.setSpecInt2000(rs.getString("SpecInt2000"));
                record.setSpecFloat2000(rs.getString("SpecFloat2000"));
                record.setEventDate(rs.getString("EventDate"));
                record.setEventTime(rs.getString("EventTime"));
                records.add(record);
            }
        } catch (SQLException e) {
            throw new ApelException(e);
        }

        return records;
    }

    private ArrayList convertJoinToArrayListGliteCE(ResultSet rs) throws ApelException {
        ArrayList records = new ArrayList();
        try {
            rs.beforeFirst();
            while (rs.next()) {
                AccountRecord record = new AccountRecord();
                record.setRecordId(rs.getString("BlahdID"));
                record.setExecutingSite(rs.getString("SiteName"));
                record.setLocalJobId(rs.getString("JobName"));
                record.setLocalUserId(rs.getString("LocalUserID"));
                record.setLcgUserId(rs.getString("GlobalUserName"));
                if (rs.getString("FullyQualifiedAttributeName").equals(ApelDB.NULL)) {
                    record.setLocalUserVO(rs.getString("LocalUserGroup"));
                } else {
                    record.setLocalUserVO(rs.getString("FullyQualifiedAttributeName"));
                }
                record.setElapsedTime(rs.getString("WallDuration"));
                record.setBaseCPUTime(rs.getString("CpuDuration"));
                record.setElapsedTimeSeconds(rs.getString("WallDurationSeconds"));
                record.setBaseCPUTimeSeconds(rs.getString("CpuDurationSeconds"));
                record.setStartTime(rs.getString("StartTime"));
                record.setStopTime(rs.getString("StopTime"));
                record.setStartTimeUtc(rs.getString("StartTimeUTC"));
                record.setStopTimeUtc(rs.getString("StopTimeUTC"));
                record.setStartTimeEpoch(rs.getString("StartTimeEpoch"));
                record.setStopTimeEpoch(rs.getString("StopTimeEpoch"));
                record.setExecutingCE(rs.getString("ResourceIdentity"));
                record.setMemoryReal(rs.getString("MemoryReal"));
                record.setMemoryVirtual(rs.getString("MemoryVirtual"));
                record.setSpecInt2000(rs.getString("SpecInt2000"));
                record.setSpecFloat2000(rs.getString("SpecFloat2000"));
                record.setEventDate(rs.getString("EventDate"));
                record.setEventTime(rs.getString("EventTime"));
                records.add(record);
            }
        } catch (SQLException e) {
            throw new ApelException(e);
        }

        return records;
    }

    private ArrayList convertToAccounting(ResultSet rs) throws ApelException {
        ArrayList records = new ArrayList();
        try {
            while (rs.next()) {
                AccountRecord record = new AccountRecord();
                record.setRecordId(rs.getString("RecordIdentity"));
                record.setExecutingSite(rs.getString("ExecutingSite"));
                record.setLocalJobId(rs.getString("LocalJobID"));
                record.setLocalUserId(rs.getString("LocalUserID"));
                record.setLcgJobId(rs.getString("LcgJobID"));
                record.setLcgUserId(rs.getString("LCGUserID"));
                String LCGUserVO = rs.getString("LCGUserVO");
                if (m_VONames.containsKey(LCGUserVO)){
                    record.setLocalUserVO(m_VONames.get(LCGUserVO));
                }
                else{
                    record.setLocalUserVO(LCGUserVO);
                }
                
                record.setElapsedTime(rs.getString("ElapsedTime"));
                record.setBaseCPUTime(rs.getString("BaseCpuTime"));
                record.setElapsedTimeSeconds(rs.getString("ElapsedTimeSeconds"));
                record.setBaseCPUTimeSeconds(rs.getString("BaseCpuTimeSeconds"));
                record.setStartTime(rs.getString("StartTime"));
                record.setStopTime(rs.getString("StopTime"));
                record.setStartTimeUtc(rs.getString("StartTimeUTC"));
                record.setStopTimeUtc(rs.getString("StopTimeUTC"));
                record.setStartTimeEpoch(rs.getString("StartTimeEpoch"));
                record.setStopTimeEpoch(rs.getString("StopTimeEpoch"));
                record.setExecutingCE(rs.getString("ExecutingCE"));
                record.setMemoryReal(rs.getString("MemoryReal"));
                record.setMemoryVirtual(rs.getString("MemoryVirtual"));
                record.setSpecInt2000(rs.getString("SpecInt2000"));
                record.setSpecFloat2000(rs.getString("SpecFloat2000"));
                record.setEventDate(rs.getString("EventDate"));
                record.setEventTime(rs.getString("EventTime"));
                records.add(record);
            }
        } catch (SQLException e) {
            throw new ApelException(e.getMessage());
        }
        return records;
    }

    public static void setVOFilter(ArrayList<String> VOList, boolean include) {

        String filter = "'";

        for (String vo : VOList) {
            filter = filter + "^/" + vo + "|" + "^" + vo + "$|";
        }
        filter = filter.substring(0, filter.length() - 1);

        if (filter.length() == 0) {
            filter = "'.*'";
        } else {
            filter = filter + "'";
        }

        if (include){
            filter = "LCGUserVO RLIKE " + filter;
        } else {
            filter = "LCGUserVO NOT RLIKE " + filter;
        }

        m_VOFilter = filter;

    }

    public static void setVOMapping(ArrayList<String> VOGroups){

        m_VONames = new HashMap<String,String>();
        for (String pair:VOGroups){
            String[] vogroup = pair.split(":");
            m_VONames.put(vogroup[0], vogroup[1]);
        }
    }
}
