/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.log.parser;

import org.apache.log4j.Logger;
import org.glite.apel.core.Log;
import org.glite.apel.core.Time;
import org.glite.apel.core.TimeInterval;
import org.glite.apel.core.log.type.BlahdRecord;
import org.glite.apel.core.log.type.BlahdRecordIterator;
import org.glite.apel.core.Utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.Vector;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

/**
 * A file containing a BLAHD Log record
 */
public class BlahdLogParser extends File {
    
    /**
     * Using log4j logging tool.
     */
    protected static final Logger LOG = 
        Logger.getLogger(BlahdLogParser.class.getName());
    
    /**
     * Buffer which reads-in data from the log file.
     */
    protected BufferedReader m_reader;
    
    /**
     * Holds all blahd records read-in from the log file.
     */
    protected Vector m_blahdRecords;
    
    /** 
     * The format used to encode the current date. 
     */
    private static final String EVENT_DATE_FORMAT = "yyyyMMdd";
    
    /** 
     * UTC constant used when converting the current date. 
     */
    
    /**
     * Formatter converts the current date into the same
     * format used by the event log file.
     */
    private SimpleDateFormat m_dateFormatter;
    
    /**
     * The timezone to be used when interpreting time stamps 
     * parsed from the log files.
     */
     private static final String UTC = "UTC";
     private static final TimeZone tz_UTC = TimeZone.getTimeZone(UTC);
       
    /**
     *  
     */
    private String m_submitHost = null;
 
    /**
     * Pattern used to chop up a line of data that matches an Record into a convenient list.
     */
    private Pattern m_delimiter;
    private Pattern m_delimiter_tmp;
    
    /**
     * Pattern used to match a blahd event record in the log file.
     */
    private Pattern m_endEventPattern;
    
    /**
     * Creates a new event log file.
     */
      //public BlahdLogParser(File file, String timezone) {
      public BlahdLogParser(File file, String submitHost) {
        super(file.getPath());
        if (submitHost != null) {
         m_submitHost = submitHost;
         //Log.info(LOG, "Warning: <SubmitHost> tag found in APEL configuration = " + submitHost);
         //Log.info(LOG, "         APEL will use this value instead of the CeId in the accounting.log!");
        }

        m_blahdRecords = new Vector();
        try {
            FileInputStream fileInputStream = new FileInputStream(file);
            if (getName().matches("^.*\\.gz$")) {
                m_reader = new BufferedReader(new InputStreamReader(
                    new GZIPInputStream(fileInputStream)));
            } else {
                m_reader = new BufferedReader(new InputStreamReader(fileInputStream));
            }
        } catch (Exception e) {
            Log.warn(LOG, "Exception opening file " + file.getPath(), e);
        }

        m_delimiter = Pattern.compile("\" \"");
        // Used to split timeStamp String into "YYYY-MM-DD" and "HH:MM:SS"
        m_delimiter_tmp = Pattern.compile(" ");
    }
    
    /**
     * Returns true if the file name matches today's date (ie the log
     * is currently being used).
     * For blah logs, rotation is based in UTC. This may not be the case for PBS where local is used.
     * This is confirmed at CESGA whose locale is CEST
     */
    public boolean isActiveLog(String blahdLogPrefix) {
        SimpleDateFormat m_dateFormatter = new SimpleDateFormat(EVENT_DATE_FORMAT);
        //m_dateFormatter.setTimeZone(TimeZone.getDefault());
        m_dateFormatter.setTimeZone(tz_UTC);
        Date currentDate = Calendar.getInstance().getTime();
        String formattedDate = m_dateFormatter.format(currentDate);
        ///////////////////////////////////////////////////////////////////////////////////////////////
        // This prefex should be specified in APEL config file as the siteAdmins are free to
        // override the default specified by BLAHPD_ACCOUNTING_INFO_LOG in /opt/glite/etc/blah.config
        // Current Log assumed to be blahp.log-YYYYMMDD or grid-jobmap_YYYYMMDD with no "gz" postfix.
        // ////////////////////////////////////////////////////////////////////////////////////////////
        String defaultLogPrefix="grid-jobmap_";
        String currentLog = null;
        if (blahdLogPrefix == null) {
          currentLog = defaultLogPrefix + formattedDate;
        } else {
          currentLog = blahdLogPrefix + formattedDate;
        }
        //Log.info(LOG,"The Current log is        : " + currentLog);
        return currentLog.equals(getName());
    }
    
    /**
     * Parses the log file and returns an iterator over all the 
     * event records found within the file.
     */
    public void parse() {
        Log.info(LOG, "Processing " + getFullPath());
        
        String line = null;
        try {
            // process each line
            while ((line = m_reader.readLine()) != null) {
                    //Log.info(LOG, "Line = " + line);
                    BlahdRecord record = createBlahdRecord(line);
                    // Add record to Vector
                    m_blahdRecords.add(record); 
            }
        } catch (IOException e) {
            Log.warn(LOG, "Exception reading file " + getName(), e);
        }
    }
    
    /**
     * Returns a new EventRecord derived from the line of data 
     * (parsed from the event log). 
     */
    BlahdRecord createBlahdRecord(String line) {
        BlahdRecord record = new BlahdRecord();
        String[] fields = m_delimiter.split(line);
        String fqan_chain="";
        Utils utils = Utils.getInstance();
        for (int index = 0; index < fields.length; index++) {
            String field = fields[index].replaceAll("\"","");
            if (field.startsWith("timestamp=")) {
                record.setTimeStamp(strip(field, '='));
                String tmp = strip(field, '=');
                String [] dateTime = m_delimiter_tmp.split(tmp);
                record.setDate(dateTime[0]);
                record.setTime(dateTime[1]);
                record.setValidFrom(utils.calculateValidFrom(dateTime[0]));
                record.setValidUntil(utils.calculateValidUntil(dateTime[0]));
            } else if (field.startsWith("userDN=")) {
                record.setUserDN(strip(field, '='));
            } else if (field.startsWith("userFQAN=")) {
                // MUST Support multiple UserFQAN in the accounting Record 
                fqan_chain+=strip(field, '=') + ";";
                // check chain length does not exceed 255
                if (fqan_chain.length() < 255 ) {
                 record.setUserFQAN(fqan_chain); 
                }
            } else if (field.startsWith("ceID=")) {
                record.setCeId(strip(field, '='));
                String [] CE = record.getCeId().split(":");
                if (m_submitHost == null) {
                   record.setSubmitHost(CE[0]);
                }else{
                   record.setSubmitHost(m_submitHost);
                }
            } else if (field.startsWith("jobID=")) {
                record.setJobId(strip(field, '='));
            } else if (field.startsWith("lrmsID=")) {
               record.setLrmsId(strip(field, '='));
               ///////////////////////////////////////////////////////////////////////
               // This code appears to be redundant!
               // Blah Data is joined to Event Data by means of the JobName variable.
               // JobName in EventRecords is ALWAYS the batch LRMS_ID
               // Therefore JobName in BlahdRecords should be set to LRMSID
               // /////////////////////////////////////////////////////////////////////
               record.setJobName(record.getLrmsId());
               ////////////////////////////////////////////////////////////////////////
               // String [] lrmsId = record.getLrmsId().split("\\.");
               //if (m_submitHost == null) {
               //    record.setJobName(record.getLrmsId()); 
               //}else{
               //    record.setJobName(lrmsId[0] + "." + m_submitHost);
               //}
               /////////////////////////////////////////////////////////////////////////
            }
        }
        return record;
    }

    /**
     * Remove the everything from the start of the string to the 
     * specified character.
     *
     * @param string string to strip.
     * @param delimiter character to look for.
     * @return string with leading text stripped off.
     */
    private String strip(String string, char delimiter) {
        int index = string.indexOf(delimiter);
        if (index != -1) {
            return string.substring(index + 1);
        }
        return string;
    }
    
    /**
     * Closes down the log file object. This is called once all meaningful 
     * data has been read from the file.
     */
    public void close() {
        if (m_reader != null) {
            try {
                m_reader.close();
            } catch (IOException e) {
                Log.warn(LOG, "Exception closing file " + getName(), e);
            }
        }
    }
    
    /**
     * Return canonical path, with built-in exception handling.
     * If an exception is thrown only the relative path is returned.
     */
    public String getFullPath() {
        String path;
        try {
            path = getCanonicalPath();
        } catch (Exception e) {
            Log.warn(LOG, "Exception trying to get filename", e);
            path = getName();
        }
        return path;
    }

 
    /**
     * Returns an iterator of all the parsed event records.
     */
    public BlahdRecordIterator getRecords() {
        return new BlahdRecordIterator(m_blahdRecords.iterator());
    }
    
}
