/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core.log.parser;

import org.apache.log4j.Logger;
import org.glite.apel.core.Log;
import org.glite.apel.core.Utils;
import org.glite.apel.core.log.type.GkRecord;
import org.glite.apel.core.log.type.GkRecordIterator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

/**
 * Provides a 'default' implementation of GkLogFile that handles globus
 * style gk logs. The log parser interprets either SGE, Condor, LSF or PBS entries.
 * 
 * Calling parse() will read-in all the log data and cause a list of gk 
 * records to be built.
 * 
 * The ingredients required to build a GkRecord are split between two 
 * lines within the log file. These lines are prefixed with 'JMA'.
 */
public class GkLogParser extends File {
    
    /**
     * Using log4j logging tool.
     */
    protected static final Logger LOG = 
        Logger.getLogger(GkLogParser.class.getName());
    
    /**
     * Constant which defines a pbs format which should be used when 
     * parsing the gk log file.
     */
    public static final int PBS_FORMAT = 0;
    
    /**
     * Constants which defines an lsf format which should be used when
     * parsing the gk log file.
     */
    public static final int LSF_FORMAT = 1;
    
    /**
     * Constants which defines an condor format which should be used when
     * parsing the gk log file.
     */
    public static final int CONDOR_FORMAT = 2;

    /**
     * Constants which defines an sge format which should be used when
     * parsing the gk log file.
     */
    public static final int SGE_FORMAT = 3;


    /**
     * Buffer which reads-in data from the log file.
     */
    protected BufferedReader m_reader;
    
    /**
     * Holds all gk records read-in from the gk log file.
     */
    protected Vector m_gkRecords;
    
    /**
     * Pattern used to match the various attributes within the JMA record,
     * where 'd' represents a decimal digit and '.' represents a dot. This
     * string is used to calculate field lengths so no extra backslashes 
     * must be added.
     */
    private String m_jobIdPrefix = 
         "JMA dddd/dd/dd dd:dd:dd GATEKEEPER_JM_ID ";
    
    /**
     * See above. 
     */
    private String m_prefixString = "^" + m_jobIdPrefix +
        "dddd-dd-dd.dd:dd:dd.dddddddddd.dddddddddd";
    
    /**
     * The compiled pattern holding the regular expression used to pick out 
     * the first line of a JMA record stored within the log file.
     */
    private Pattern m_jmaHeader;
     
    /**
     * The compiled pattern holding the regular expression used to pick out 
     * the second line of a complete JMA record. This one will check for lines
     * that can be identified with the 'pbs' or 'lsf' or 'condor' or 'sge' job manager 
     * type.
     */
    private Pattern m_jmaEventType;
    
   /**
     * The compiled pattern holding the regular expression used to pick out
     * the second line of a complete JMA record for the Imperial College SGE implemetation.
     */
    private Pattern m_jmaEventType_SGE_IC;
    private Pattern m_GramScriptJobId_SGE;
    private Pattern m_GramScriptJobId_SGE_IC;
 
    /**
     * The compiled pattern holding the regular expression used to pick out 
     * the second line of a complete JMA record. This one will check for lines
     * that include the 'fork' identifier (ie which means the pair is dud
     * and should be ignored).
     */
    private Pattern m_jmaForkType;
     
    /**
     * Compiled pattern is used to check if the gk record has been
     * generated by an lcg varient of pbs/lsf.
     */
    private Pattern m_vanilaType;
     
    /**
     * Contains a map of JMA records which match the 'header' regular
     * expression pattern.
     */
    private HashMap line1Map = new HashMap();
     
    /**
     * Contains a map of JMA records which matches the 'footer' 
     * regular expression pattern.
     */
    private HashMap line2Map = new HashMap();
    
    /**
     * Creates a new GateKeeperLogFile object.
     */
    public GkLogParser(File file, int eventType) {
        super(file.getPath());
        m_gkRecords = new Vector();
        try {
            FileInputStream fileInputStream = new FileInputStream(file);
            if (getName().matches("^.*\\.gz$")) {
                m_reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(fileInputStream)));
            } else {
                m_reader = new BufferedReader(new InputStreamReader(fileInputStream));
            }
        } catch (Exception e) {
            Log.warn(LOG, "Exception opening file " + file.getPath(), e);
        }
        String eventTypeStr = null;
        if (eventType == PBS_FORMAT) {
        	eventTypeStr = "pbs";
        } else if (eventType == LSF_FORMAT) {
        	eventTypeStr = "lsf";
        } else if (eventType == CONDOR_FORMAT) {
                eventTypeStr = "condor";
       } else if (eventType == SGE_FORMAT) {
                eventTypeStr = "sge";
        } else {
            Log.warn(LOG, "Unknown log format defined: " + eventType);
        }
        String actualPrefixString = m_prefixString.replaceAll("d", "\\\\d").replaceAll("\\.", "\\\\.");
        m_jmaHeader = Pattern.compile(actualPrefixString + " for /.*");
        m_jmaEventType = Pattern.compile(actualPrefixString + " has GRAM_SCRIPT_JOB_ID .* .*" + eventTypeStr + ".*");
        m_vanilaType = Pattern.compile("\\d*:.*" + eventTypeStr + ".*:internal_\\d*:\\d*\\.\\d*");
        m_jmaForkType = Pattern.compile(actualPrefixString + " has GRAM_SCRIPT_JOB_ID .* fork");
        // Pattern to match a JMA record
        m_jmaEventType_SGE_IC = Pattern.compile(actualPrefixString + " has GRAM_SCRIPT_JOB_ID \\d+\\|[^\\|]+\\|[^\\|]+ " + eventTypeStr + ".*");
        // Pattern to split up the GramScriptJobID into smaller chunks to derive a new GramScriptJobID
        m_GramScriptJobId_SGE = Pattern.compile("\\|");
        // Patter used to determine if GramSCriptJobId is derived from SGE_IC 
        m_GramScriptJobId_SGE_IC = Pattern.compile("\\d+\\|[^\\|]+");
    }
    
    /**
     * Parses the log file and returns an iterator of all the 
     * generated records.
     */
    public void parse() {
        try {
            String line = null;
            while ((line = m_reader.readLine()) != null) {
                if (m_jmaHeader.matcher(line).matches()) {
                    line1Map.put(getLocalJobID(line), line);
                } else if (m_jmaEventType.matcher(line).matches() ||
                    m_jmaForkType.matcher(line).matches()) {
                    line2Map.put(getLocalJobID(line), line);
                }
            }
        } catch (IOException e) {
            Log.warn(LOG, "Error reading gk log file " + getPath(), e);
        }
        buildGkRecords();
    }  
    
    /**
     * Scan the two Maps containing lines matching line1regex and line2regex
     * for the first pair of lines corresponding to single valid gate keeper
     * JMA record.
     */
    private void buildGkRecords() {
        Iterator it = line1Map.keySet().iterator();
        while (it.hasNext()) {
            String localJobID = (String) it.next();
            if (line2Map.containsKey(localJobID)) {
                String line1 = (String) line1Map.get(localJobID);
                String line2 = (String) line2Map.get(localJobID);
                it.remove();
                line2Map.remove(localJobID);
               if (!m_jmaForkType.matcher(line2).matches()) {
                    // DK: Determine GramScriptJobID for SGE_IC line format
                    String gramScriptJobID = null;
                    if (m_jmaEventType_SGE_IC.matcher(line2).matches()) {
                     gramScriptJobID = getGramScriptJobID_SGE_IC(line2);
                    }else{
                     gramScriptJobID = getGramScriptJobID(line2);
                    }
                    String globalUserName = getGlobalUserName(line1);
                    m_gkRecords.add(createGkRecord(localJobID, gramScriptJobID, globalUserName));
                }
            }
        }
        if (line2Map.size() != 0) {
            Log.warn(LOG, "Could not find first part of " +
                "JMA record in " + line2Map.size() + " case/s");
        }
        if (line1Map.size() != 0) {
            Log.warn(LOG, "Could not find second part of JMA " +
                "record in " + line1Map.size() + " case/s");
        }
    }
    
    /**
     * Attempts to create a new gk record.
	 */
	private GkRecord createGkRecord(String localJobID, String gramScriptJobID, String globalUserName) {
                Utils utils = Utils.getInstance();        
		String date = localJobID.replaceFirst("\\..*", "");
		String time = localJobID.replaceFirst("^.*?\\d\\d\\.", "").replaceFirst("\\..*", "");
		GkRecord gkRecord = new GkRecord();
		gkRecord.setGramScriptJobID(gramScriptJobID);
		gkRecord.setLocalJobID(localJobID);
		gkRecord.setGlobalUserName(globalUserName);
		gkRecord.setMeasurementDate(date);
		gkRecord.setMeasurementTime(time);
		gkRecord.setValidFrom(utils.calculateValidFrom(date));
		gkRecord.setValidUntil(utils.calculateValidUntil(date));

                // Dealing with Vanilla Cases
                // SGE_IC has no vanilla case
                if (m_GramScriptJobId_SGE_IC.matcher(gramScriptJobID).matches()) {
                 gkRecord.setLcgPbs(true);
                }else{
                 gkRecord.setLcgPbs(m_vanilaType.matcher(gramScriptJobID).matches());
                }

                return gkRecord;
	}

	/**
     * Returns the user DN string contained the specified JMA line.
     */
    private String getGlobalUserName(String line) {
        String name = line.substring(m_prefixString.length());
        name = name.replaceFirst("for ", "");
        return name.replaceFirst(" on [\\.\\d]*$", "");
    }
    
    /**
     * Returns the Gram job ID contained in a JMA line.
     */
    private String getGramScriptJobID(String line) {
        String name = line.substring(m_prefixString.length());
        name = name.replaceFirst("has GRAM_SCRIPT_JOB_ID ", "");
        int startOfManager = name.indexOf(" manager type");
        name = name.substring(0, startOfManager);
        LOG.debug("job name is " + name);
        return name;
    }

    /**
     * Returns the Gram job ID contained in a JMA line.
     */
    private String getGramScriptJobID_SGE_IC(String line) {
        String name = line.substring(m_prefixString.length());
        name = name.replaceFirst("has GRAM_SCRIPT_JOB_ID ", "");
        int startOfManager = name.indexOf(" manager type");
        name = name.substring(0, startOfManager);
        // The name is bigger than a VARCHAR(255), so we now split it on the "|"
        // and rebuild a smaller GramScriptJobID
        String [] values = m_GramScriptJobId_SGE.split(name);
        String lrms_id = values[0];
        String gram_id_part1 = values[1];
        String gram_id_part2 = values[2];
        LOG.debug("job name is " + name);
        return lrms_id + "|" + gram_id_part1;
    }

    /**
     * Returns the local job ID contained in a JMA line.
     */
    private String getLocalJobID(String line) {
        return line.substring(m_jobIdPrefix.length()).replaceAll(" .*$", "");
    }
    
    /**
     * Closes down the log file object. This is called once all meaningful 
     * data has been read from the file.
     */
    public void close() {
        try {
            m_reader.close();
        } catch (IOException e) {
            Log.warn(LOG, "Exception closing file " + getName(), e);
        }
    }

    /**
     * Return canonical path, with built-in exception handling.
     * If an exception is thrown only the relative path is returned.
     */
    public String getFullPath() {
        String path;
        try {
            path = getCanonicalPath();
        } catch (Exception e) {
            Log.warn(LOG, "Exception trying to get filename", e);
            path = getName();
        }
        return path;
    }
    
    /**
     * Returns an iterator of all the parsed gk records.
     */
    public GkRecordIterator getRecords() {
        return new GkRecordIterator(m_gkRecords.iterator());
    }
    
}
