/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.log.parser;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Vector;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

import org.apache.log4j.Logger;
import org.glite.apel.core.ApelException;
import org.glite.apel.core.Log;
import org.glite.apel.core.Utils;
import org.glite.apel.core.log.type.MsgRecord;
import org.glite.apel.core.log.type.MsgRecordIterator;

/**
 * Provides parsing capability to read-in system message logs generated
 * by the job manager. The parser can be configured to read-in PBS or
 * LSF entries stored in the log. 
 */
public class MsgLogParser extends File {
    
    /**
     * Using log4j logging tool.
     */
    protected static final Logger LOG = 
        Logger.getLogger(MsgLogParser.class.getName());
    
    /**
     * Contains a list of all message records parsed from the log file.
     */
    protected Vector m_msgRecords;
    
    /**
     * Buffer which reads-in data from the log file.
     */
    protected BufferedReader m_reader;
    
    /**
     * Regular expression used to identify if a parsed line
     * of data matches a valid message record.
     */
    private Pattern m_validMessage;
    
    /**
     * Formats date/time derived from log file into a predifined format.
     */
    private SimpleDateFormat m_inputDateFormat;
    
    /**
     * Formats the date/time into a predefined format 
     * (compatible will MeasurementDate/Time).
     */
    private SimpleDateFormat m_outputDateFormat;
    
    /**
     * Constant which defines a pbs format which should be used when 
     * parsing the gk log file.
     */
    public static final int PBS_FORMAT = 0;
    
    /**
     * Constants which defines an lsf format which should be used when
     * parsing the gk log file.
     */
    public static final int LSF_FORMAT = 1;
    
    /**
     * Constants which defines an condor format which should be used when
     * parsing the gk log file.
     */
    public static final int CONDOR_FORMAT = 2;


    /**
     * Constants which defines an sge format which should be used when
     * parsing the gk log file.
     */
    public static final int SGE_FORMAT = 3;

    ///////////////////////////////////////////////////
    // DK: Added from GkLofParser
    ///////////////////////////////////////////////////
    private String m_jobIdPrefix;
    private String m_prefixString;
    // Pattern to match a JMA record
    private Pattern m_jmaEventType_SGE_IC;
    // Pattern to split up the GramScriptJobID into smaller chunks
    private Pattern m_GramScriptJobId_SGE;
    ////////////////////////////////////////////////



    /**
     * Creats a new pbs message log file.
     */
    public MsgLogParser(File file, int eventType) {
        super(file.getPath());
        m_msgRecords = new Vector();
        try {
            FileInputStream fileInputStream = new FileInputStream(file);
            if (getName().matches("^.*\\.gz$")) {
                m_reader = new BufferedReader(new InputStreamReader(
                    new GZIPInputStream(fileInputStream)));
            } else {
                m_reader = new BufferedReader(new InputStreamReader(
                    fileInputStream));
            }
        } catch (Exception e) {
            Log.warn(LOG, "Exception opening file " + file.getPath(), e);
        }
        m_inputDateFormat = new SimpleDateFormat("MMM d yyyy");
        m_outputDateFormat = new SimpleDateFormat("yyyy-MM-dd");
        String eventTypeStr = null;
        if (eventType == PBS_FORMAT) {
            eventTypeStr = "pbs";
        } else if (eventType == LSF_FORMAT) {
            eventTypeStr = "lsf";
        } else if (eventType == CONDOR_FORMAT) {
            eventTypeStr = "condor";
        } else if (eventType == SGE_FORMAT) {
            eventTypeStr = "sge";
        } else {
            Log.warn(LOG, "Unknown log format defined: " + eventType);
        }

        //
        // Default way to determine the BatchID
        //
        m_validMessage = Pattern.compile("^.*? gridinfo: .*? Submitted job " +
            "\\d*:.*" + eventTypeStr + ".*:internal_\\d*:\\d*\\.\\d* " +
            "to batch system .*" + eventTypeStr + ".* with ID .*");

        /////////////////////////////////////////////////////////////////////////////////////
        // In SGE_IC messages, the line above doesn't exist. Therfore we try to find the JMA
        // lines that contain the GramScriptJobID. From this line we can get LRMSJob_id.
        // Note: that we must be very specific to match the jobmanager type as JMA receords
        // exist in standard implementations of lcgpbs, lcglsf  etc
        //
        // Code taken from GkLogParser, m_jobIdPrefix modified to contain gridinfo
        //////////////////////////////////////////////////////////////////////////////////////
        String gridinfo_prefix    = "^.*? gridinfo\\[\\d+\\]: ";
        m_jobIdPrefix             = "JMA dddd/dd/dd dd:dd:dd GATEKEEPER_JM_ID ";
        m_prefixString            = m_jobIdPrefix + "dddd-dd-dd.dd:dd:dd.dddddddddd.dddddddddd";
        String actualPrefixString = gridinfo_prefix + m_prefixString.replaceAll("d", "\\\\d").replaceAll("\\.", "\\\\.");
        // Pattern to match a JMA record
        m_jmaEventType_SGE_IC = Pattern.compile(actualPrefixString + " has GRAM_SCRIPT_JOB_ID \\d+\\|[^\\|]+\\|[^\\|]+ " + eventTypeStr + ".*");

        // Pattern to split up the GramScriptJobID into smaller chunks
        m_GramScriptJobId_SGE = Pattern.compile("\\|");
    }
    
    /**
     * Parses the line of data, returning a new MessageLogRecord.
     */
    MsgRecord createMessageRecord(String line) throws ApelException {
        if (m_jmaEventType_SGE_IC.matcher(line).matches()) {
             return getMessageRecord_SGE_IC(line);
        }else{
             return getMessageRecord(line);
        }
    }

    private MsgRecord getMessageRecord(String line) throws ApelException {        
        MsgRecord record = new MsgRecord();
        String gramId = line.replaceFirst("^.*? Submitted job", "");
        gramId = gramId.replaceFirst(" to batch system .*? with ID .*", "");
        String jobName = line.replaceFirst("^.* to batch system .*? with ID ", "");
        record.setGramId(gramId.trim());
        record.setJobName(jobName.trim());
        long epochTimeSeconds = 0;
        String epochTimeString = gramId.replaceFirst(":.*", "").trim();
        try {
            epochTimeSeconds = Long.parseLong(epochTimeString);
        } catch (NumberFormatException e) {
            throw new ApelException("Cannot translate " + epochTimeString + " to long");
        }
        Calendar cal = Calendar.getInstance();
        cal.setTimeInMillis(epochTimeSeconds * 1000);
        int year = cal.get(Calendar.YEAR);
        int month = cal.get(Calendar.MONTH);
        String dateString = line.substring(0, 6);
        // Correct year in case the epoch time is in different locale timezone
        if ((month == Calendar.JANUARY) &&
            (dateString.substring(0, 3).equalsIgnoreCase("Dec"))) {
            year = year + 1;
        } else if ((month == Calendar.DECEMBER) &&
            (dateString.substring(0, 3).equalsIgnoreCase("Jan"))) {
            year = year - 1;
        }
        dateString = dateString + " " + String.valueOf(year);
        String measurementDate = null;
        try {
            Date date = m_inputDateFormat.parse(dateString);
            measurementDate = m_outputDateFormat.format(date);
        } catch (ParseException e) {
            throw new ApelException("Unable to translate date " + dateString, e);
        }
        record.setMeasurementDate(measurementDate);
        record.setMeasurementTime(line.substring(7, 15));
        Utils utils = Utils.getInstance();
        record.setValidFrom(utils.calculateValidFrom(measurementDate));
        record.setValidUntil(utils.calculateValidUntil(measurementDate));
        return record;
    }
    
    private MsgRecord getMessageRecord_SGE_IC(String line) throws ApelException {
        Utils utils = Utils.getInstance();
        MsgRecord record = new MsgRecord();

        String name = line;

        int startOfGramScript = name.indexOf(" has GRAM_SCRIPT_JOB_ID");
        int startOfGkJm       = name.indexOf("GATEKEEPER_JM_ID");
        int startOfManager    = name.indexOf(" manager type");

        String gram = name.substring(startOfGramScript, startOfManager).replaceFirst(" has GRAM_SCRIPT_JOB_ID ", "");
        gram = gram.trim();
        //Log.info(LOG,"GRAM_SCRIPT_JOB_ID: =\"" + gram + "\"");

        String localJobID = name.substring(startOfGkJm, startOfGramScript).replaceFirst("GATEKEEPER_JM_ID ", "");
        localJobID = localJobID.trim();
        //Log.info(LOG,"GATEKEEPER_JM_ID: = \"" + localJobID + "\"");

        // The name is bigger than a VARCHAR(255), so we now split it on the "|"
        // and rebuild a smaller GramScriptJobID
        String [] values = m_GramScriptJobId_SGE.split(gram);
        
        String lrms_id = values[0];
        String gram_id_part1 = values[1];
        String gram_id_part2 = values[2];

        record.setGramId(lrms_id + "|" + gram_id_part1);
        record.setJobName(lrms_id);

        String date = localJobID.replaceFirst("\\..*", "");
        String time = localJobID.replaceFirst("^.*?\\d\\d\\.", "").replaceFirst("\\..*", "");
        record.setMeasurementDate(date);
        record.setMeasurementTime(time);
        record.setValidFrom(utils.calculateValidFrom(date));
        record.setValidUntil(utils.calculateValidUntil(date));
        return record;
    }

    /**
     * Parses the log file and returns an iterator over all the message records found within the file.
     */
	public void parse() {
        String line = null;
        try {
            while ((line = m_reader.readLine()) != null) {
                //
                // Looking for the standard validMessage or the SGE_IC JMA Header in the gridinfo records
                //
                if ( m_validMessage.matcher(line).matches() || 
                     m_jmaEventType_SGE_IC.matcher(line).matches() ) {
                    try {
                    	MsgRecord record = createMessageRecord(line);
                    	m_msgRecords.add(record);
                    } catch (ApelException e) {
                        Log.warn(LOG, "Error parsing: " + getFullPath(), e);
                    }
                }
            }
        } catch (IOException e) {
            Log.warn(LOG, "Exception reading file " + getName(), e);
        }
	}

    /**
     * Closes down the log file object. This is called once all meaningful 
     * data has been read from the file.
     */
    public void close() {
        if (m_reader != null) {
            try {
                m_reader.close();
            } catch (IOException e) {
                Log.warn(LOG, "Exception closing file " + getName(), e);
            }
        }
    }

    /**
     * Return canonical path, with built-in exception handling.
     * If an exception is thrown only the relative path is returned.
     */
    public String getFullPath() {
        String path;
        try {
            path = getCanonicalPath();
        } catch (Exception e) {
            Log.warn(LOG, "Exception trying to get filename", e);
            path = getName();
        }
        return path;
    }
    
    /**
     * Returns an iterator of all the parsed message records.
     */
    public MsgRecordIterator getRecords() {
        return new MsgRecordIterator(m_msgRecords.iterator());
    }
    
}
