/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.processor;

import org.apache.log4j.Logger;
import org.glite.apel.core.ApelException;
import org.glite.apel.core.Log;
import org.glite.apel.core.Utils;
import org.glite.apel.core.db.ApelDB;
import org.glite.apel.core.db.MySQLImpl;
import org.glite.apel.core.log.parser.BlahdLogParser;
import java.io.File;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.ArrayList;

/**
 * Parser is used to parse all the available PBS event logs.
 */
public class BlahdLogProcessor {
    /**
     * Reference to logging tool.
     */
    private static final Logger LOG = Logger.getLogger(BlahdLogProcessor.class.getName());
    /**
     * Mysql db interface.
     */
    private ApelDB m_db;
    /**
     * Contains a list of all valid event files (each element
     * is a pbs file type).
     */
    private List m_logFiles;
   /**
    * The Blahd Log Prefix
    */ 
    private String m_logPrefix;
   /**
    * The LRMS Host name
    */
    private String m_submitHost;
    /**
     * Creates a new Blahd log processor.
     */
    public BlahdLogProcessor(String url, String username, String password, 
     String siteName, String eventLogDir, boolean searchSubDirs,
        boolean reprocessLogs, List extraLogs, String logPrefix, String submitHost) throws ApelException {
        //Log.info(LOG, "==== BlahdLogProcessor ====");
        m_logFiles  = new ArrayList();
        m_logPrefix = logPrefix;
        //Log.info(LOG, "> logPrefix is " + m_logPrefix );
        m_submitHost = submitHost;
        m_db = new MySQLImpl(url, username, password, siteName);
        //
        // Check that the BlahdTable exists: 
        // If not, create it....but is this necessary, we already steer Apel via the config file.
        //
        m_db.inspectTable(ApelDB.BLAHD_TABLE_NAME);

        if (reprocessLogs) {
            Log.info(LOG, "Reprocess enabled, resetting all event logs");
            m_db.deleteAllFiles(ApelDB.FILE_TYPE_BLAHD_LOG);
         } else {
            Log.info(LOG, "Reprocess disabled, checking new event logs only");
        }
        setupLogFiles(eventLogDir, searchSubDirs, extraLogs);
    }
    /**
     * Reads-in all valid event log files and stores them as file
     * objects within the list m_logFiles.
     */
    private void setupLogFiles(String eventLogDir, boolean recursive,List extraFiles) {
        //Log.info(LOG, "==== org.glite.apel.core.BlahdLogProcessor:setupLogFiles ====");
        //Log.info(LOG, "> logPrefix is " + m_logPrefix );
        //Log.info(LOG, "> LogDirectory is " + eventLogDir );
        Vector logFilePaths = Utils.getInstance().getBlahdLogFiles(eventLogDir, recursive, m_logPrefix);
        Iterator my = logFilePaths.iterator();
        if (extraFiles != null) {
            Iterator it = extraFiles.iterator();
            while (it.hasNext()) {
                String extraFile = (String) it.next();
                logFilePaths.add(extraFile);
            }
        }
        Iterator it = logFilePaths.iterator();
        while (it.hasNext()) {
            String filePath = (String) it.next();
             m_logFiles.add(new BlahdLogParser(new File(filePath), m_submitHost) );
        }
    }
    /**
     * Processes the blahd log files which have not so far been processed.
     */
    public void processLogs() throws ApelException {
        String hostName=m_db.getHostName();
	int size = m_logFiles.size();
        int eventCount = 0;
	for (int i=0;i<size;i++){
            BlahdLogParser file = (BlahdLogParser)m_logFiles.get(0);
            //DK - Handle multiple CE scenario
            //if (m_db.isFileProcessed(file.getName(), ApelDB.FILE_TYPE_BLAHD_LOG)) {
            if (m_db.isFileProcessed(file.getFullPath(), hostName, ApelDB.FILE_TYPE_BLAHD_LOG)) {
                Log.info(LOG, "Ignoring already processed accounting log file: " + 
                hostName + " " + file.getFullPath());
            } else {
                if (!file.isActiveLog(m_logPrefix)) {
                  m_db.insertProcessedFile(file.getName(), ApelDB.FILE_TYPE_BLAHD_LOG, file.getFullPath());
                }
                file.parse();
                eventCount += m_db.insertBlahdRecords(file.getRecords());
            }
            // 
            // Savannah Bug # 23796 
            // Close the file after it has been processed
            file.close();
	    m_logFiles.remove(0);
        }
        Log.info(LOG, "Blahd records inserted: " + eventCount);
    }
    /**
     * Tidies-up any loose resources.
     */
    public void close() {
        m_db.close();
        Iterator eventLogFiles = m_logFiles.iterator();
        while (eventLogFiles.hasNext()) {
            ((BlahdLogParser) eventLogFiles.next()).close();
        }
    }
}
