/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core.processor;

import org.apache.log4j.Logger;
import org.glite.apel.core.ApelException;
import org.glite.apel.core.Log;
import org.glite.apel.core.Utils;
import org.glite.apel.core.db.ApelDB;
import org.glite.apel.core.db.MySQLImpl;
import org.glite.apel.core.log.parser.GkLogParser;
import org.glite.apel.core.log.parser.MsgLogParser;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Vector;

/**
 * Parser is used to parse the Globus Gatekeeper log and Job Manager message
 * log files.
 */
public class GkLogProcessor {
    
    /**
     * Using log4j logging tool.
     */
    private static final Logger LOG = 
        Logger.getLogger(GkLogProcessor.class.getName());
    
    /**
     * The GK log file that is currently being used by the Globus GK.
     */
    private final String CURRENT_GK_LOG = "globus-gatekeeper.log";
    
    /**
     * Mysql db interface.
     */
    private ApelDB m_db;
    
    /**
     * Contains a list of all valid gk files (each element
     * is a gk file type).
     */
    private List m_gkFiles;
    
    /**
     * Contains a list of all valid message files (each element
     * is a message file type).
     */
    private List m_msgFiles;
    
    /**
     * The parsing mode (this defines if PBS or LSF log parsing
     * is used).
     */
    private int m_parserType;
    
    /**
     * Creates a new gk processor.
     */
    public GkLogProcessor(String url, String username, String password, 
        String siteName, String gkLogDir, String msgLogDir, 
        boolean searchSubDirs, boolean reprocessLogs, List extraGkLogs, 
        List extraMsgLogs, int parserType) throws ApelException {
        m_parserType = parserType;
        m_gkFiles = new ArrayList();
        m_msgFiles = new ArrayList();
        m_db = new MySQLImpl(url, username, password, siteName);
        if (reprocessLogs) {
            Log.info(LOG, "Reprocess option enabled, resetting all " +
                "Gk and Message logs stored in db");
            m_db.deleteAllFiles(ApelDB.FILE_TYPE_MESSAGE_LOG);
            m_db.deleteAllFiles(ApelDB.FILE_TYPE_GATE_KEEPER_LOG);
        }
        setupGkLogFiles(gkLogDir, searchSubDirs, extraGkLogs);
        setupMsgLogFiles(msgLogDir, searchSubDirs, extraMsgLogs);
    }

    /**
     * Reads-in all valid globus gk log files and stores them as file
     * objects within the list m_gkFiles.
     */
    private void setupGkLogFiles(String gkLogDir, boolean recursive, 
        List extraFiles) {
        Vector logFilePaths = 
            Utils.getInstance().getGkLogFiles(gkLogDir, recursive);
        if (extraFiles != null) {
            Iterator it = extraFiles.iterator();
            while (it.hasNext()) {
                String extraFile = (String) it.next();
                logFilePaths.add(extraFile);
            }
        }        
        Iterator it = logFilePaths.iterator();
        while (it.hasNext()) {
            String filePath = (String) it.next();
            m_gkFiles.add(new GkLogParser(new File(filePath), m_parserType));
        }
    }    
    
    /**
     * Reads-in all valid system message log files and stores them as file
     * objects within the list m_msgFiles.
     */
    private void setupMsgLogFiles(String msgLogDir, boolean recursive, List extraFiles) {
        Vector logFilePaths = Utils.getInstance().getMsgLogFiles(msgLogDir, recursive);
        if (extraFiles != null) {
            Iterator it = extraFiles.iterator();
            while (it.hasNext()) {
                String extraFile = (String) it.next();
                logFilePaths.add(extraFile);
            }
        }        
        Iterator it = logFilePaths.iterator();
        while (it.hasNext()) {
            String filePath = (String) it.next();
            m_msgFiles.add(new MsgLogParser(new File(filePath), m_parserType));
        }
    }    
    
    public void processLogs() throws ApelException {
    	processGkLogs();
        processMsgLogs();
    }
    
    /**
     * Carry out the processing of the gate keeper log file.
     */
    public void processGkLogs() throws ApelException {
        Log.info(LOG, "**** Updating Gate Keeper log table (" + ApelDB.GK_TABLE_NAME + ") ****");

        // DK - get local hostName
        String hostName=m_db.getHostName();

        Iterator gkLogFiles = m_gkFiles.iterator();
        int gkCount = 0;
        int msgCount = 0;
        while (gkLogFiles.hasNext()) {
          GkLogParser file = (GkLogParser) gkLogFiles.next();
          //DK - Changed to support multiple CE configurations
          //if (m_db.isFileProcessed(file.getName(), ApelDB.FILE_TYPE_GATE_KEEPER_LOG)) {
          if (m_db.isFileProcessed(file.getFullPath(), hostName, ApelDB.FILE_TYPE_GATE_KEEPER_LOG)) {
              Log.info(LOG, "Already processed GK log file: " + hostName + " " + file.getFullPath());
            } else {
                Log.info(LOG, "Processing gate keeper log file: " + hostName + " " + file.getFullPath());
                file.parse();
                gkCount += m_db.insertGkRecords(file.getRecords());
                msgCount += m_db.insertMsgRecords(file.getRecords());
                if (!file.getName().equals(CURRENT_GK_LOG)) {
                    m_db.insertProcessedFile(file.getName(), ApelDB.FILE_TYPE_GATE_KEEPER_LOG, file.getFullPath());
                }
            }
            file.close();
        }
        Log.info(LOG, "Gk records inserted: " + gkCount);
        Log.info(LOG, "Vanilla Job Mananager Records inserted into MessageRecord: " + msgCount);
    }
   
    /**
     * Carries out the processing of the message log file.
     */
    private void processMsgLogs() throws ApelException {
        Log.info(LOG, "**** Updating Job ID table (" + ApelDB.MSG_TABLE_NAME + ") from system message log ****");
        int msgCount = 0;
	int size=m_msgFiles.size();
        // Get a list of all of the log files
        for (int i=0; i<size; i++) {
	    System.gc();
            MsgLogParser file = (MsgLogParser) m_msgFiles.get(0);
            if (!m_db.isFileProcessed(file.getName(), 
                ApelDB.FILE_TYPE_MESSAGE_LOG)) {
                Log.info(LOG, "Processing message log file: " + file.getFullPath());
                file.parse();
                msgCount += m_db.insertMsgRecords(file.getRecords());
                checkMsgFileType(file);
            } else {
                Log.info(LOG, "Already processed message log file: " + file.getName());
            }
            file.close();
	    m_msgFiles.remove(0);
        }
        Log.info(LOG, "Msg records inserted: " + msgCount);
    }

    /**
     * Carries out the processing of the message log file.
     */
/*    private void processMsgLogs() throws ApelException {
        Log.info(LOG, "**** Updating Job ID table (" + ApelDB.MSG_TABLE_NAME + ") from system message log ****");
        int msgCount = 0;
        // Get a list of all of the log files
        Iterator msgLogFiles = m_msgFiles.iterator();
        while (msgLogFiles.hasNext()) {
            MsgLogParser file = (MsgLogParser) msgLogFiles.next();
            if (!m_db.isFileProcessed(file.getName(),
                ApelDB.FILE_TYPE_MESSAGE_LOG)) {
                Log.info(LOG, "Processing message log file: " + file.getFullPath());
                file.parse();
                msgCount += m_db.insertMsgRecords(file.getRecords());
                checkMsgFileType(file);
            } else {
                Log.info(LOG, "Already processed message log file: " + file.getName());
            }
            file.close();
        }
        Log.info(LOG, "Msg records inserted: " + msgCount);
    }*/

    
    /**
     * Reprocess the current log file each time, plus the rotated old ones 
     * as they change. All other files go into the processed files table as
     * they change.
     */
    private void checkMsgFileType(MsgLogParser file) 
        throws ApelException {
        String fileName = file.getName();
        if (!fileName.equals("messages") && 
            !fileName.matches("messages\\.\\d\\.gz") &&
            !fileName.matches("messages\\.\\d")) {
            try {
                String path = file.getCanonicalPath();
                m_db.insertProcessedFile(fileName, ApelDB.FILE_TYPE_MESSAGE_LOG, path);
            } catch (IOException e) {
                Log.warn(LOG, "Error trying to obtain filename: " + e);
            }
        }
    }
    
    /**
     * Tidies-up any loose resources.
     */
    public void close() {
        m_db.close();
        Iterator gkLogFiles = m_gkFiles.iterator();
        while (gkLogFiles.hasNext()) {
            ((GkLogParser) gkLogFiles.next()).close();
        }
        Iterator msgLogFiles = m_msgFiles.iterator();
        while (msgLogFiles.hasNext()) {
           ((MsgLogParser) msgLogFiles.next()).close();
        }
    }



    
}
