/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.core.processor;

import java.util.logging.Level;
import org.apache.log4j.Logger;
import org.glite.apel.core.ApelException;
import org.glite.apel.core.Log;
import org.glite.apel.core.db.ApelDB;
import org.glite.apel.core.db.MySQLImpl;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.TimeZone;
import java.util.regex.Pattern;
import javax.naming.CommunicationException;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;

/**
 * Obtains the current spec values for the site. This is normally retireved
 * from an LDAP query on the GIIS, but can be set to a default value.
 */
public class SpecProcessor {

    /**
     * Using log4j.
     */
    private Logger m_log;
    /**
     * Mysql db interface.
     */
    private ApelDB m_db;
    /**
     * Date formatter.
     */
    private SimpleDateFormat m_dateFormat;
    /**
     * Time formatter.
     */
    private SimpleDateFormat m_timeFormat;
    /**
     * JNDI context factory used when querying the GIIS via LDAP.
     */
    private static final String LDAP_CTX_FACTORY = "com.sun.jndi.ldap.LdapCtxFactory";
    /**
     * Timeout before giving up on the LDAP query.
     */
    private static final int LDAP_QUERY_TIMEOUT_SECONDS = 60;
    /**
     * Constant used during LDAP query.
     */
    private static final String GLUE_CHUNK = "GlueChunkKey";
    private static final String GLUE_FOREIGN = "GlueForeignKey";
    /**
     * Constant used during LDAP query.
     */
    private static final String GLUE_SUB_CLUS = "GlueSubClusterUniqueID";
    private static final String GLUE_CE = "GlueCEUniqueID";
    /**
     * Constant used during LDAP query.
     */
    private static final String SPEC_INT = "GlueHostBenchmarkSI00";
    private static final String CE_CAP = "GlueCECapability";
    /**
     * Constant used during LDAP query.
     */
    private static final String SPEC_FLOAT = "GlueHostBenchmarkSF00";
    /**
     * Pattern to match the giisHost
     */
    private static final Pattern pattern = Pattern.compile(".*:[0-9]*");
    /**
     * Pattern to match the CPUScalingReferenceSI00 value
     */
    private static final Pattern scaling_pattern = Pattern.compile("CPUScalingReferenceSI00=[0-9]*");
    /**
     * The giisHost and port
     */
    private String m_host = null;
    private String m_port = null;
    /**
     * The siteName
     */
    private String m_siteName;

    /**
     * Creates a new CpuPerformanceProcessor object.
     */
    public SpecProcessor(String url, String username, String password,
            String siteName) throws ApelException {
        m_log = Logger.getLogger(getClass().getName());
        m_db = new MySQLImpl(url, username, password, siteName);
        m_dateFormat = new SimpleDateFormat("yyyy-MM-dd");
        m_dateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
        m_timeFormat = new SimpleDateFormat("HH:mm:ss");
        m_timeFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
        m_siteName = siteName;
    }

    /**
     * Gets the CPU spec values from an LDAP query (to the GIIS). The
     * value is then stored in the SpecRecords table.
     */
    public void setLDAPSpec(String giisHost) throws ApelException {
        Collection<String> CEs = new ArrayList();
        Hashtable env = new Hashtable();
        Hashtable found = new Hashtable();
        NamingEnumeration enumer;
        DirContext ctx;
        SearchControls sc = new SearchControls();
        String specInt = null;
        String capability = null;
        NamingEnumeration capabilities = null;
        SearchResult searchResults;
        Attributes attr;
        String clusId = null;
        Log.info(m_log, "Finding CPU spec values from GIIS server: " + giisHost);
        //////////////////////////////////////////////////////////////////////////
        //Construct Host and Port parameters for LDAP query
        //          <GIIS>lcgce01.gridpp.rl.ac.uk:1234</GIIS> will query port 1234
        // Default: <GIIS>lcgce01.gridpp.rl.ac.uk</GIIS> will query port 2170
        //
        //////////////////////////////////////////////////////////////////////////
        if (pattern.matcher(giisHost).matches()) {
            String[] input = giisHost.split(":");
            m_host = input[0];
            m_port = input[1];
        } else {
            m_host = giisHost;
            m_port = "2170";
        }

        String url = "ldap://" + m_host + ":" + m_port;
        Log.info(m_log, "LDAP URL: " + url);
        env.put(Context.PROVIDER_URL, url);
        env.put(Context.INITIAL_CONTEXT_FACTORY, LDAP_CTX_FACTORY);

        // Get the SpecInt from CPUScalingReferenceSI00 in GlueCECapability
        // If this value isn't found, find the Cluster for a CE, go to the
        // SubCluster for the Cluster and get the GlueHostBenchmarkSI00.
        try {

            sc.setSearchScope(SearchControls.SUBTREE_SCOPE);
            sc.setTimeLimit(LDAP_QUERY_TIMEOUT_SECONDS * 1000);
            ctx = (DirContext) new InitialDirContext(env);

            // First try to get the value from the GlueCECapability attribute
            // APEL reads the GlueCE objects for the CECapability value
            if (m_port.equals("2135")) {
                enumer = ctx.search("mds-vo-name=local,o=grid",
                        "(objectclass=GlueCE)", sc);
            } else {
                enumer = ctx.search("o=grid", "(objectclass=GlueCE)", sc);
            }

            while (enumer.hasMore()) {

                // Get the cluster ID
                searchResults = (SearchResult) enumer.next();
                attr = searchResults.getAttributes();
                clusId = (String) attr.get(GLUE_CE).get();
                CEs.add(clusId);

                // Find the SpecInt value from CPUScalingReference
                try {
                    capabilities = attr.get(CE_CAP).getAll();
                    while (capabilities.hasMore()) {
                        capability = (String) capabilities.next();
                        if (scaling_pattern.matcher(capability).lookingAt()) {
                            String[] input = capability.split("=");
                            specInt = input[1];
                            found.put(clusId, specInt);
                        }
                    }
                } catch (NullPointerException npe) {
                    // For the cases when CECapability is not present
                }
            }


        } catch (CommunicationException e) {
            throw new ApelException("Communication Exception: Error making LDAP query to: " + url, e);
        } catch (NamingException e) {
            throw new ApelException("Naming Exception: Error making LDAP query to: " + url, e);
        }

        // If the GlueCECapability value is not set, get SpecInt2000 from the
        // CE's Cluster's SubCluster's GlueHostBenchmarkSI00.
        for (String CE : CEs) {
            // Only if the CE doesn't already have a SpecInt associated with it,
            // get it from the GlueHostBenchmarkSI00
            if (!found.containsKey(CE)) {
                try {
                    // Get the Cluster name from the CE
                    if (m_port.equals("2135")) {
                        enumer = ctx.search("mds-vo-name=local,o=grid",
                                "(GlueCEUniqueID=" + CE + ")", sc);
                    } else {
                        enumer = ctx.search("o=grid", "(GlueCEUniqueID="
                                + CE + ")", sc);
                    }

                    while (enumer.hasMore()) {
                        // Get the ClusterID from the CE
                        searchResults = (SearchResult) enumer.next();
                        attr = searchResults.getAttributes();
                        clusId = (String) attr.get("GlueCEHostingCluster").get();

                        // Assuming that the ClusterID is the SubClusterID,
                        // find the appropriate SubCluster
                        NamingEnumeration subClusters;
                        if (m_port.equals("2135")) {
                            subClusters = ctx.search("mds-vo-name=local,o=grid",
                                    "(GlueSubClusterUniqueID=" + clusId + ")", sc);
                        } else {
                            subClusters = ctx.search("o=grid", "(GlueSubClusterUniqueID="
                                    + clusId + ")", sc);
                        }

                        while (subClusters.hasMore()) {
                            searchResults = (SearchResult) subClusters.next();
                            attr = searchResults.getAttributes();
                            specInt = (String) attr.get("GlueHostBenchmarkSI00").get();
                            if (!found.containsKey(CE)) {
                                found.put(CE, specInt);
                            }
                        }
                    }
                } catch (NamingException ex) {
                    java.util.logging.Logger.getLogger(SpecProcessor.class.getName()).log(Level.SEVERE, null, ex);
                } catch (NullPointerException npe) {
                    // For the cases when CECapability is not present
                }
            }
        }
        Enumeration j = found.keys();
        while (j.hasMoreElements()) {
            String ce = (String) j.nextElement();
            specInt = (String) found.get(ce);
            String specFloat = null;
            Date currentTime = Calendar.getInstance().getTime();
            String measurementDate = m_dateFormat.format(currentTime);
            String measurementTime = m_timeFormat.format(currentTime);
            m_db.processSpecRecord(ce, ce, specInt, specFloat,
             measurementDate, measurementTime);

        }


    }

    /**
     * Checks the ldap values that were read-in. Prints out any suspicious
     * looking null values. Returns true if an invalid value was found.
     */
    private boolean skipSpec(String subClusId, String clusId,
            String specInt, String specFloat) {
        Log.info(m_log, "Cluster: " + clusId + ", Subcluster: "
                + subClusId + ", SpecInt2000: " + specInt
                + ", SpecFloat2000: " + specFloat);
        if (subClusId == null) {
            Log.warn(m_log, "GlueSubClusterID is undefined");
            return true;
        } else if (clusId == null) {
            Log.warn(m_log, "clusterID is undefined");
            return true;
        } else if (specInt == null) {
            Log.warn(m_log, "specInt2000 is undefined");
            return true;
        } else if (specFloat == null) {
            Log.warn(m_log, "specFloat2000 is undefined");
            return true;
        } else {
            return false;
        }
    }

    /**
     * Parses the default spec value defined by the user and then
     * publishes the value to the SpecRecords table.
     */
    public void setDefaultSpec(String defaultSpec) throws ApelException {
        throw new ApelException("APEL no longer supports default Spec values!");
    }

    /**
     * Tidies-up any loose resources.
     */
    public void close() {
        m_db.close();
    }
}
