/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
package org.glite.apel.core.db;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Vector;

import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.glite.apel.core.ApelException;
import org.glite.apel.core.UTCTime;
import org.glite.apel.core.log.type.AccountRecordIterator;
import org.glite.apel.core.log.type.EventRecord;
import org.glite.apel.core.log.type.EventRecordIterator;
import org.glite.apel.core.log.type.GkRecord;
import org.glite.apel.core.log.type.GkRecordIterator;
import org.glite.apel.core.log.type.MsgRecord;
import org.glite.apel.core.log.type.MsgRecordIterator;

/**
 * Unit tests that check the functionality of the mysql db wrapper.
 */
public class MySQLImplTest extends TestCase {
    
    private Connection m_connection;
    
    private MySQLImpl m_mysqlImpl;
    
    public static junit.framework.Test suite() {
        return new TestSuite(MySQLImplTest.class);
    }
    
    /**
     * Scrubs all the tables used by the Apel program.
     */
    private void deleteRecords() {
        Connection con = null;
        Statement stmt = null;
        try {
            stmt = m_connection.createStatement();
            stmt.executeUpdate("DELETE FROM " + ApelDB.REPUBLISH_TABLE_NAME);
            stmt.executeUpdate("DELETE FROM " + ApelDB.MSG_TABLE_NAME);
            stmt.executeUpdate("DELETE FROM " + ApelDB.GK_TABLE_NAME);
            stmt.executeUpdate("DELETE FROM " + ApelDB.SPEC_TABLE_NAME);
            stmt.executeUpdate("DELETE FROM " + ApelDB.EVENT_TABLE_NAME);
            stmt.executeUpdate("DELETE FROM " + ApelDB.LCG_RECORDS_TABLE_NAME);
            stmt.executeUpdate("DELETE FROM " + ApelDB.PROCESSED_FILES_TABLE_NAME);
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            close(stmt);
            close(con);
        }
    }
    
    public void setUp() {
        String dbName = System.getProperty("DB_NAME");
        String userName = System.getProperty("DB_USERNAME");
        String passWord = System.getProperty("DB_PASSWORD");
        if (dbName == null) {
            fail("DB_NAME, DB_USERNAME and DB_PASSWORD must all be " +
                "defined as system properties");
        }
        String dburl = "jdbc:mysql://localhost/" + dbName;
        String driver = "org.gjt.mm.mysql.Driver";
        try {
            Class.forName(driver);
			m_connection = DriverManager.getConnection(dburl, userName, passWord);
            m_mysqlImpl = new MySQLImpl(dburl, userName, passWord, "EXAMPLE-SITE");
            deleteRecords();
        } catch (Exception e) {
			e.printStackTrace();
			fail();
        }
    }

    public void tearDown() {
        deleteRecords();
    	close(m_connection);
    }
    
    /**
     * Closes an sql Connection, catching any exceptions that are thrown.
     */
    void close(Connection conn) {
        try {
            conn.close();
        } catch (Exception e) {}
    }

    /**
     * Closes an sql Statement, catching any exceptions that are thrown.
     */
    void close(Statement stmt) {
        try {
            stmt.close();
        } catch (Exception e) {}
    }
    
    /**
     * Tests all records are removed from a table using the delete all
     * records method.
     */
	public void testDeleteAllRecords() throws ApelException {
		Statement stmt = null;
        try {
			stmt = m_connection.createStatement();
            stmt.execute("INSERT INTO " + ApelDB.REPUBLISH_TABLE_NAME + 
                " VALUES ('0000-00-00', '00:00:00', 'junk')");
            assertEquals(1, m_mysqlImpl.count(ApelDB.REPUBLISH_TABLE_NAME));
            m_mysqlImpl.deleteAllRecords(ApelDB.REPUBLISH_TABLE_NAME);
            assertEquals(0, m_mysqlImpl.count(ApelDB.REPUBLISH_TABLE_NAME));
		} catch (SQLException e) {
			e.printStackTrace();
		}
        close(stmt);
	}

    /**
     * Tests if an event record has its processed flag set, the delete
     * operation will remove it.
     */
	public void testDeleteProcessedEventRecords() {
        EventRecord record = new EventRecord();
        record.setJobName("");
        record.setSubmitHost("");
        record.setEventDate("");
        record.setEventTime("");
        Vector list = new Vector();
        list.add(record);
        EventRecordIterator eventIt = 
            new EventRecordIterator(list.iterator());
        try {
            m_mysqlImpl.insertEventRecords(eventIt);
            assertEquals(1, m_mysqlImpl.count(ApelDB.EVENT_TABLE_NAME));
            setProcessedFlag(ApelDB.EVENT_TABLE_NAME);
            m_mysqlImpl.deleteProcessedEventRecords();
            assertEquals(0, m_mysqlImpl.count(ApelDB.EVENT_TABLE_NAME));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Sets the processed flag for all records (applies only for 
     * gk, event and msg tables).
     */
    private void setProcessedFlag(String tableName) {
        Statement stmt = null;
        try {
            stmt = m_connection.createStatement();
            stmt.execute("UPDATE " + tableName + " SET Processed = '1'");
        } catch (SQLException e) {
            e.printStackTrace();
            fail();
        }
        close(stmt);
    }
    
    /**
     * Tests if a gk record has its processed flag set, the delete
     * operation will remove it.
     */
	public void testDeleteProcessedGkRecords() {
        GkRecord record = new GkRecord();
        record.setGramScriptJobID("");
        Vector list = new Vector();
        list.add(record);
        GkRecordIterator eventIt = new GkRecordIterator(list.iterator());
        try {
            m_mysqlImpl.insertGkRecords(eventIt);
            assertEquals(1, m_mysqlImpl.count(ApelDB.GK_TABLE_NAME));
            setProcessedFlag(ApelDB.GK_TABLE_NAME);
            m_mysqlImpl.deleteProcessedGkRecords();
            assertEquals(0, m_mysqlImpl.count(ApelDB.GK_TABLE_NAME));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Tests message records which have their processed flag set are
     * deleted ok.
     */
	public void testDeleteProcessedMsgRecords() {
		MsgRecord record = new MsgRecord();
		record.setGramId("");
        record.setJobName("");
        record.setMeasurementDate("2005-04-18");
        Vector list = new Vector();
        list.add(record);
        MsgRecordIterator msgIt = new MsgRecordIterator(list.iterator());
        try {
            m_mysqlImpl.insertMsgRecords(msgIt);
            assertEquals(1, m_mysqlImpl.count(ApelDB.MSG_TABLE_NAME));
            setProcessedFlag(ApelDB.MSG_TABLE_NAME);
            m_mysqlImpl.deleteProcessedMsgRecords();
            assertEquals(0, m_mysqlImpl.count(ApelDB.MSG_TABLE_NAME));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Tests all files are removed from the gk, event and msg tables 
     * when the delete all files method is called.
     */
	public void testDeleteAllFiles() {
        try {
            m_mysqlImpl.insertProcessedFile("event-file", 
                ApelDB.FILE_TYPE_EVENT_LOG, "");
            m_mysqlImpl.insertProcessedFile("gk-file", 
                ApelDB.FILE_TYPE_GATE_KEEPER_LOG, "");
            m_mysqlImpl.insertProcessedFile("msg-file", 
                    ApelDB.FILE_TYPE_MESSAGE_LOG, "");
            assertEquals(3, m_mysqlImpl.count(ApelDB.PROCESSED_FILES_TABLE_NAME));
            m_mysqlImpl.deleteAllFiles(ApelDB.FILE_TYPE_EVENT_LOG);
            assertEquals(2, m_mysqlImpl.count(ApelDB.PROCESSED_FILES_TABLE_NAME));
            m_mysqlImpl.deleteAllFiles(ApelDB.FILE_TYPE_GATE_KEEPER_LOG);
            assertEquals(1, m_mysqlImpl.count(ApelDB.PROCESSED_FILES_TABLE_NAME));
            m_mysqlImpl.deleteAllFiles(ApelDB.FILE_TYPE_MESSAGE_LOG);
            assertEquals(0, m_mysqlImpl.count(ApelDB.PROCESSED_FILES_TABLE_NAME));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Test event records can be inserted ok.
     */
	public void testInsertEventRecords() {
        EventRecord record = new EventRecord();
        record.setJobName("");
        record.setSubmitHost("");
        record.setEventDate("");
        record.setEventTime("");
        Vector eventList = new Vector();
        eventList.add(record);
        EventRecordIterator eventIt = 
            new EventRecordIterator(eventList.iterator());
        try {
            m_mysqlImpl.insertEventRecords(eventIt);
            assertEquals(1, m_mysqlImpl.count(ApelDB.EVENT_TABLE_NAME));
            eventIt = new EventRecordIterator(eventList.iterator());
            m_mysqlImpl.insertEventRecords(eventIt);
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}
    
    /**
     * Tests gk records can be inserted ok.
     */
	public void testInsertGkRecords() {
        GkRecord record = new GkRecord();
        record.setGramScriptJobID("");
        Vector list = new Vector();
        list.add(record);
        GkRecordIterator eventIt = new GkRecordIterator(list.iterator());
        try {
            m_mysqlImpl.insertGkRecords(eventIt);
            assertEquals(1, m_mysqlImpl.count(ApelDB.GK_TABLE_NAME));
            // Test duplicates are ignored
            m_mysqlImpl.insertGkRecords(new GkRecordIterator(list.iterator()));
            assertEquals(1, m_mysqlImpl.count(ApelDB.GK_TABLE_NAME));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Tests a message record can be successfully inserted.
     */
	public void testInsertMsgRecords() {
        MsgRecord record = new MsgRecord();
        record.setGramId("");
        record.setJobName("");
        record.setMeasurementDate("2005-04-18");
        Vector list = new Vector();
        list.add(record);
        MsgRecordIterator msgIt = new MsgRecordIterator(list.iterator());
        try {
            m_mysqlImpl.insertMsgRecords(msgIt);
            assertEquals(1, m_mysqlImpl.count(ApelDB.MSG_TABLE_NAME));
            // Test duplicates are ignored
            m_mysqlImpl.insertMsgRecords(new MsgRecordIterator(list.iterator()));
            assertEquals(1, m_mysqlImpl.count(ApelDB.MSG_TABLE_NAME));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Tests records can be inserted into the processed files table.
     */
	public void testInsertProcessedFile() {
        try {
            m_mysqlImpl.insertProcessedFile("event-file", 
                ApelDB.FILE_TYPE_EVENT_LOG, "");
            // Test duplicates are ignored
            m_mysqlImpl.insertProcessedFile("event-file", 
                ApelDB.FILE_TYPE_EVENT_LOG, "");
            assertEquals(1, m_mysqlImpl.count(ApelDB.PROCESSED_FILES_TABLE_NAME));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Tests there exists at most one catch all record.
     */
    public void testCheckCatchAll() throws Exception {
        String clusterId = "lcgce02.gridpp.rl.ac.uk";
        String specValue = "200";
        String date = "2004-07-04";
        m_mysqlImpl.insertSpecRecord(clusterId, clusterId, specValue, 
            specValue, date, "08:25:00");
        m_mysqlImpl.insertSpecRecord(clusterId, clusterId, specValue, 
            specValue, date, "08:30:00");
        int numCatchAlls = getNumberCatchAlls();
        assertTrue(numCatchAlls == 1);
    }
    
    /**
     * Returns the number of 'catch all' SpecRecords.
     */
    private int getNumberCatchAlls() {
        Statement stmt = null;
        int numCatchAlls = 0;
        try {
            stmt = m_connection.createStatement();
            String query = "SELECT MeasurementDate, MeasurementTime FROM " +
                "SpecRecords WHERE siteName = 'EXAMPLE-SITE' " +
                " ORDER BY MeasurementDate, MeasurementTime";
            ResultSet result = stmt.executeQuery(query);
            while (result.next()) {
                String date = result.getString(1);
                if (date.equals("1970-01-01")) {
                    numCatchAlls++;
                }
            }
        } catch (SQLException e) {
        	e.printStackTrace();
            fail();
        }
        close(stmt);
        return numCatchAlls;
    }
    
    /**
     * Tests spec records can be inserted ok.
     */
	public void testInsertSpecRecord() {
		try {
            m_mysqlImpl.insertSpecRecord("cluster", "sub-cluster", "200", 
                "220.1", "2005-01-01", "00:00:00");
            // A default spec is inserted - making the count at least 2
            assertEquals(2, m_mysqlImpl.count(ApelDB.SPEC_TABLE_NAME));
            m_mysqlImpl.insertSpecRecord("cluster", "sub-cluster", "200", 
                "220.1", "2005-01-01", "00:00:00");
            // Test duplicates are ignored
            assertEquals(2, m_mysqlImpl.count(ApelDB.SPEC_TABLE_NAME));
        } catch (Exception e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Checks the functionality to prevent duplicate spec int/float values 
     * from being inserted.
     */
    public void testIgnoreDuplicateSpec() throws Exception {
        String specInt = "211";
        String specFloat = "220";
        m_mysqlImpl.insertSpecRecord("lcgce02", "lcgce02", specInt, 
            specFloat, "2004-07-03", "08:25:00");
        assertEquals(2, m_mysqlImpl.count(ApelDB.SPEC_TABLE_NAME));
        // Add another spec with the same int/float value
        m_mysqlImpl.insertSpecRecord("lcgce02", "lcgce02", specInt, 
            specFloat, "2004-07-03", "08:25:30");
        assertEquals(2, m_mysqlImpl.count(ApelDB.SPEC_TABLE_NAME));
    }
    
    /**
     * Tests the catch-all spec value is updated to the latest GIIS/default
     * value when a new spec is added. 
     */
    public void testCatchAllIsUpdated() throws Exception {
        String clusterId = "lcgce02.gridpp.rl.ac.uk";
        String specInt = "200";
        String specFloat = "400";
        m_mysqlImpl.insertSpecRecord(clusterId, clusterId, specInt, 
            specFloat, "2005-01-01", "00:00:00");
        checkCatchAllSpec(specInt, specFloat);
        specInt = "100";
        specFloat = "250";
        m_mysqlImpl.insertSpecRecord(clusterId, clusterId, specInt, 
            specFloat, "2005-01-10", "00:00:00");
        checkCatchAllSpec(specInt, specFloat);
    }
    
    /**
     * Checks the specInt and specFloat values matchup.
     */
    private void checkCatchAllSpec(String specInt, String specFloat) {
        Statement stmt = null;
        try {
            stmt = m_connection.createStatement();
            String query = "SELECT SpecInt2000, SpecFloat2000 FROM " +
                "SpecRecords ORDER BY MeasurementDate, MeasurementTime";
            ResultSet result = stmt.executeQuery(query);
            if (result.next()) {
                assertEquals(specInt, result.getString(1));
                assertEquals(specFloat, result.getString(2));
            } else {
                fail();
            }
        } catch (SQLException e) {
            e.printStackTrace();
            fail();
        }
    }    
    
    /**
     * Checks if a file has already been processed.
     */
	public void testIsFileProcessed() {
        try {
            m_mysqlImpl.insertProcessedFile("event-file", 
                ApelDB.FILE_TYPE_EVENT_LOG, "");
            m_mysqlImpl.insertProcessedFile("gk-file", 
                ApelDB.FILE_TYPE_GATE_KEEPER_LOG, "");
            m_mysqlImpl.insertProcessedFile("msg-file", 
                ApelDB.FILE_TYPE_MESSAGE_LOG, "");
            assertTrue(m_mysqlImpl.isFileProcessed("event-file", 
                ApelDB.FILE_TYPE_EVENT_LOG));
            assertTrue(m_mysqlImpl.isFileProcessed("gk-file", 
                ApelDB.FILE_TYPE_GATE_KEEPER_LOG));
            assertTrue(m_mysqlImpl.isFileProcessed("msg-file", 
                ApelDB.FILE_TYPE_MESSAGE_LOG));
            assertFalse(m_mysqlImpl.isFileProcessed("missing-file", 
                ApelDB.FILE_TYPE_MESSAGE_LOG));
        } catch (ApelException e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Tests the date/time are correctly updated in the RepublishInfo table.
     */
    public void testUpdateTimeStamp() throws SQLException, ApelException {
        assertFalse(m_mysqlImpl.containsRepublishTimeStamp());
        m_mysqlImpl.updateRepulishTimeStamp();
        assertTrue(m_mysqlImpl.containsRepublishTimeStamp());
        compareDateTime(m_mysqlImpl.getRepublishDate(), 
            m_mysqlImpl.getRepublishTime());
    }

    /**
     * Checks the date/time correspond to the equivalent values found in
     * the first record of the RepublishInfo table.
     */
    private void compareDateTime(String date, String time) 
        throws SQLException{
        ResultSet rs = getRecords(ApelDB.REPUBLISH_TABLE_NAME);
        rs.next();
        assertEquals(rs.getString(1), date);
        assertEquals(rs.getString(2), time);
        rs.close();
    }
    
    /**
     * Selects all the records from a particular table.
     */
    private ResultSet getRecords(String tableName) {
        Statement stmt = null;
        ResultSet result = null;
        try {
            stmt = m_connection.createStatement();
            result = stmt.executeQuery("SELECT * FROM " + tableName);
        } catch (SQLException e) {
            fail();
        }
        return result;
    }
    
    /**
     * Tests the correct number of accounting records are returned after
     * a particular date/time.
     */
	public void testGetNumAccountingRecords() {
        UTCTime utcTime = new UTCTime();        
        insertAccountingRecord(utcTime, 1);
        try {
            int num = m_mysqlImpl.getNumAccountingRecords(utcTime.getDate(), 
                utcTime.getTime());
            assertEquals(num, 1);
            Thread.sleep(2000);
            utcTime.updateDateTime();
            num = m_mysqlImpl.getNumAccountingRecords(utcTime.getDate(), 
                utcTime.getTime());
            assertEquals(num, 0);
        } catch (Exception e) {
        	e.printStackTrace();
            fail();
        }
	}

    private void insertAccountingRecord(UTCTime time, int id) {
        Statement stmt = null;
        try {
            stmt = m_connection.createStatement();
            stmt.executeUpdate("INSERT INTO " + 
                ApelDB.LCG_RECORDS_TABLE_NAME + " (RecordIdentity, " +
                "MeasurementDate, MeasurementTime) VALUES ('" +
                id + "', '" + time.getDate() + "', '" + time.getTime() + "')"); 
        } catch (SQLException e) {
            e.printStackTrace();
            fail();
        }
        close(stmt);
    }
    
    /**
     * Test all accounting records are returned after a particular
     * date/time.
     */
	public void testGetAccountingRecords() {
        UTCTime utcTime = new UTCTime();        
        insertAccountingRecord(utcTime, 1);
        try {
            AccountRecordIterator it = 
                m_mysqlImpl.getAccountingRecords(utcTime.getDate(), 
                utcTime.getTime());
            assertTrue(it.hasNext());
            assertEquals(it.next().getRecordId(), "1");
            assertFalse(it.hasNext());
        } catch (Exception e) {
            e.printStackTrace();
            fail();
        }
	}

    /**
     * Tests the time stamp stored on each accounting record is updated
     * correctly.
     */
	public void testUpdateAccountingTimeStamp() {
        UTCTime utcTime = new UTCTime();        
        insertAccountingRecord(utcTime, 1);
        try {
            int num = m_mysqlImpl.getNumAccountingRecords(utcTime.getDate(), 
                utcTime.getTime());
            assertEquals(num, 1);
            Thread.sleep(2000);
            utcTime.updateDateTime();
            num = m_mysqlImpl.getNumAccountingRecords(utcTime.getDate(), 
                utcTime.getTime());
            assertEquals(num, 0);
            m_mysqlImpl.updateAccountingTimeStamp();
            num = m_mysqlImpl.getNumAccountingRecords(utcTime.getDate(), 
                utcTime.getTime());
            assertEquals(num, 1);
        } catch (Exception e) {
            e.printStackTrace();
            fail();
        }
        
        
        
	}

    /**
     * Test the utc time stamp is correctly setup.
     */
	public void testSetupTimeStamp() {
        Statement stmt = null;
        try {
            stmt = m_connection.createStatement();
            stmt.execute("INSERT INTO " + ApelDB.REPUBLISH_TABLE_NAME + 
                " VALUES ('2005-04-18', '00:00:00', '')"); 
        } catch (SQLException e) {
            e.printStackTrace();
            fail();
        }
        close(stmt);
        try {
            UTCTime time = m_mysqlImpl.setupTimeStamp();
            assertEquals(time.getDate(), "2005-04-18");
            assertEquals(time.getTime(), "00:00:00");
        } catch (ApelException e) {
        	e.printStackTrace();
            fail();
        }
    }

}
