/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core.config.parser;

import org.glite.apel.core.config.map.CLMap;
import org.glite.apel.core.config.parser.CLParser;
import org.glite.apel.core.config.parser.Config;
import org.glite.apel.core.config.parser.ConfigurationException;

import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Units tests for the org.edg.info.apel.CLParserTest class.
 */
public class CLParserTest extends TestCase {

    public static void main(String[] args) {
        CLParserTest test = new CLParserTest();
        test.setUp();
        test.testHelp();
    }

    public static junit.framework.Test suite() {
        return new TestSuite(CLParserTest.class);
    }

    public void setUp() {}

    /**
     * Tests the command line parser will throw a ConfigurationException
     * when a syntax error is found.
     */
    public void testBadOption() {
    	String[] args = new String[] {"--gobble-de-gook"};
        try {
        	new CLParser(args).parse(new CLMap());
        	fail("bad command line did not throw an exception");
        } catch (ConfigurationException e) {}    	
    }

    /**
     * Tests the command line parser will throw a ConfigurationException
     * when an option/property is missing a leading '-'.
     */
    public void testBadSyntax() {
    	String[] args = new String[] {"gobble-de-gook"};
        try {
        	new CLParser(args).parse(new CLMap());
        	fail("bad command line did not throw an exception");
        } catch (ConfigurationException e) {}    	
    }
    
    /**
	 * Tests the command line parser will throw a ConfigurationException
	 * when a user property is missing a value pair.
	 */
	public void testPropertyWithoutValue() {
	    String[] args = new String[] {"-f"};
	    try {
	    	new CLParser(args).parse(new CLMap());
	    	fail("property missing a value but no exception raised");
	    } catch (ConfigurationException e) {}    	
	}
    
    /**
     * Tests the help option is recognized by the command line parser.
     */
    public void testHelp() {
    	String[] args = new String[] {"--help"};
        CLParser parser = new CLParser(args);
        try {
        	Config config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 1);
            assertEquals(config.getNumberProperties(), 0);
            assertTrue(config.isOptionSet(CLMap.OPT_HELP));
            args = new String[] {"-h"};
            parser = new CLParser(args);
        	config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 1);
            assertEquals(config.getNumberProperties(), 0);
            assertTrue(config.isOptionSet(CLMap.OPT_HELP));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }

    /**
     * Tests the version option is recognized by the command line parser.
     */
    public void testVersion() {
    	String[] args = new String[] {"--version"};
        CLParser parser = new CLParser(args);
        try {
        	Config config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 1);
            assertEquals(config.getNumberProperties(), 0);
            assertTrue(config.isOptionSet(CLMap.OPT_VERSION));
            args = new String[] {"-v"};
            parser = new CLParser(args);
        	config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 1);
            assertEquals(config.getNumberProperties(), 0);
            assertTrue(config.isOptionSet(CLMap.OPT_VERSION));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    
    /**
     * Tests grouped options such as -hv are parsed correctly.  
     */
    public void testGroupedOptions() {
    	String[] args = new String[] {"-hv"};
        CLParser parser = new CLParser(args);
        try {
        	Config config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 2);
            assertEquals(config.getNumberProperties(), 0);
            assertTrue(config.isOptionSet(CLMap.OPT_HELP));
            assertTrue(config.isOptionSet(CLMap.OPT_VERSION));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }  
    
    /**
     * Tests the config file property is correctly parsed from the command
     * line.
     */
    public void testConfigFile() {
    	String fileName = "/beans/milk";
    	String[] args = new String[] {"--file", fileName};
        CLParser parser = new CLParser(args);
        try {
        	Config config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 0);
            assertEquals(config.getNumberProperties(), 1);
            assertEquals(config.getProperty(CLMap.PROP_CONFIG_FILE), 
            	fileName);
            fileName = "beans";
        	args = new String[] {"-f", fileName};
            parser = new CLParser(args);
        	config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 0);
            assertEquals(config.getNumberProperties(), 1);
            assertEquals(config.getProperty(CLMap.PROP_CONFIG_FILE),
            	fileName);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    
    /**
     * Tests grouped options such as -hv and parameters are parsed correctly.  
     */
    public void testOptionsAndParameters() {
    	String fileName = "/beans/milk";
    	String[] args = new String[] {"-hvf", fileName};
        CLParser parser = new CLParser(args);
        try {
        	Config config = parser.parse(new CLMap());
            assertEquals(config.getNumberOptions(), 2);
            assertEquals(config.getNumberProperties(), 1);
            assertTrue(config.isOptionSet(CLMap.OPT_HELP));
            assertTrue(config.isOptionSet(CLMap.OPT_VERSION));
            assertEquals(config.getProperty(CLMap.PROP_CONFIG_FILE),
            	fileName);
        	args = new String[] {"-hv", "--file", fileName};
            parser = new CLParser(args);
            assertEquals(config.getNumberOptions(), 2);
            assertEquals(config.getNumberProperties(), 1);
            assertTrue(config.isOptionSet(CLMap.OPT_HELP));
            assertTrue(config.isOptionSet(CLMap.OPT_VERSION));
            assertEquals(config.getProperty(CLMap.PROP_CONFIG_FILE),
            	fileName);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    } 

}
