/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package org.glite.apel.core.config.parser;

import java.io.FileNotFoundException;
import java.io.StringReader;
import java.util.Vector;
import org.glite.apel.core.config.map.FileMap;
import org.glite.apel.core.config.parser.Config;
import org.glite.apel.core.config.parser.ConfigurationException;
import org.glite.apel.core.config.parser.XMLParser;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Unit tests to verify the XML parser using the Apel config file 
 * format.
 */
public class XMLParserTest extends TestCase {
    private static String m_configFileStart = "<ApelConfiguration>";
    private static String m_configFileEnd = "</ApelConfiguration>";
    private static String m_yesValue = "yes";
    private static String m_noValue = "no";
    public static junit.framework.Test suite() {
        return new TestSuite(XMLParserTest.class);
    }
    public void setUp() {}
    /**
     * Tests if an invalid file is given, a FileNotFoundException 
     * is thrown.
     */
    public void testFileNotFound() {
        try {
            new XMLParser("/a/file/that/doesnt/exist");
            fail("invalid file does not through exception");
        } catch (FileNotFoundException e) {
        	// Test passes ok
        }
    }
    /**
     * Tests a ConfigurationException is thrown when parsing
     * badly formed XML.
     */
    public void testFileNotWellFormed() {
        try {
            XMLParser source = new XMLParser(new StringReader(m_configFileStart));
            source.parse(new FileMap());
            fail("parser did not throw exception for badly formed xml");
        } catch (ConfigurationException e) {
        	// Test passes ok
        }
    }    
    /**
     * Tests an exception is thrown for XML which is not part of the
     * standard Apel config syntax.
     */
    public void testInvalidConfig() {
        try {
            String xml = m_configFileStart + "<funky-monkey>" + "</funky-monkey>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            parser.parse(new FileMap());
            fail("invalid xml element should thrown an exception");
        } catch (ConfigurationException e) {
        	// Test passes ok
        }
    }
    /**
     * Tests when a CPU Processor config is defined, the related option
     * for performance processing is set.
     */
    public void testCPUProcessing() {
        try {
            String xml = m_configFileStart + "<CPUProcessor>" + "</CPUProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_SPEC_PROCESSOR));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the db password can be parsed from the XML template.
     */
    public void testDBPassword() {
        String dbPassword = "password";
        try {
            String xml = m_configFileStart + "<DBPassword>" + dbPassword + "</DBPassword>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_DB_PASSWORD, dbPassword);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the db url can be parsed from the XML template.
     */
    public void testDBURL() {
        String dbURL = "jdbc:mysql://localhost:3306/lcg";
        try {
            String xml = m_configFileStart + "<DBURL>" + dbURL + "</DBURL>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_DB_URL, dbURL);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the db username can be parsed from the XML template.
     */
    public void testDBUsername() {
        String dbUsername = "lcg";
        try {
            String xml = m_configFileStart + "<DBUsername>" + dbUsername + "</DBUsername>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_DB_USERNAME, dbUsername);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the default spec value is set when the related XML element is
     * parsed.
     */
    public void testDefaultSpec() {
        String defaultSpec = "300:400";
        try {
            String xml = m_configFileStart + "<CPUProcessor>" + "<DefaultCPUSpec>" + defaultSpec + 
            "</DefaultCPUSpec></CPUProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_USE_DEFAULT_SPEC, defaultSpec);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the enable debug logging flag is set when the related XML 
     * attribute is parsed.
     */
    public void testEnableDebugLogging() {
        try {
            String xml = "<ApelConfiguration" + " enableDebugLogging=\"" + m_noValue + "\"/>";
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertFalse(config.isOptionSet(FileMap.OPT_QUIET));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }    	
    }   
    /**
     * Tests the extra event log is set when the related XML 
     * element is parsed.
     */
    public void testExtraEventLogFile() {
        String extraFile1 = "extra/file1";
        try {
            String xml = m_configFileStart + "<EventLogProcessor>" + "<Logs><ExtraFile>" + extraFile1 + 
            "</ExtraFile></Logs></EventLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_EXTRA_EVENT_LOGS, extraFile1);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the extra GK log file is set when the related XML 
     * element is parsed.
     */
    public void testExtraGKLogFile() {
        String extraFile1 = "extra/file1";
        try {
            String xml = m_configFileStart + "<GKLogProcessor>" + "<Logs><GKLogs><ExtraFile>" + extraFile1 + 
            "</ExtraFile></GKLogs></Logs></GKLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_EXTRA_GK_LOGS, extraFile1);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the extra GK message log file is set when the related XML 
     * element is parsed.
     */
    public void testExtraMessageLogFile() {
        String extraFile1 = "extra/file1";
        try {
            String xml = m_configFileStart + "<GKLogProcessor>" + "<Logs><MessageLogs><ExtraFile>" + extraFile1 + 
            "</ExtraFile></MessageLogs></Logs></GKLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_EXTRA_MSG_LOG_FILES, extraFile1);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests multiple extra event logs are set when the related XML 
     * element is parsed.
     */
    public void testMultipleExtraEventLogFile() {
        String extraFile1 = "extra/file1";
        String extraFile2 = "extra/file2";
        try {
            String xml = m_configFileStart + "<EventLogProcessor>" + "<Logs><ExtraFile>" + extraFile1 + 
            "</ExtraFile><ExtraFile>" + extraFile2 + "</ExtraFile></Logs></EventLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            Vector files = config.getPropertyValues(FileMap.PROP_EXTRA_EVENT_LOGS);
            assertTrue(files.contains(extraFile1));
            assertTrue(files.contains(extraFile2));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the GIIS value is set when the related XML 
     * element is parsed.
     */
    public void testGIIS() {
        String giis = "test.ldap.server";
        try {
            String xml = m_configFileStart + "<CPUProcessor><GIIS>" + giis + 
            "</GIIS></CPUProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_GIIS, giis);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the GKLogs directory value is set when the related XML 
     * element is parsed.
     */
    public void testGKLogs() {
        String gklogs = "/var/log/gklogs";
        try {
            String xml = m_configFileStart + "<GKLogProcessor><Logs>" + "<GKLogs><Dir>" + gklogs + 
            "</Dir></GKLogs></Logs>" + "</GKLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_GKLOG, gklogs);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the DB delete processor option is set when the related XML 
     * element is parsed.
     */
    public void testDBDelete() {
        try {
            String xml = m_configFileStart + "<DBDeleteProcessor " + "cleanAll=\"" + m_yesValue + 
            "\"/>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_DB_DELETE_PROCESSOR));
            assertTrue(config.isOptionSet(FileMap.OPT_CLEAN_ALL_FROM_TABLES));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests when a GK element is defined, the option for GK log
     * processing is set.
     */
    public void testGKProcessing() {
        try {
            String xml = m_configFileStart + "<GKLogProcessor/>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_GK_LOG_PROCESSOR));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the message log directory is set when the related XML
     * element is parsed.
     */
    public void testMessageLogs() {
        String mslogs = "/var/log/messagelogs";
    	try {
            String xml = m_configFileStart + "<GKLogProcessor><Logs>" + "<MessageLogs><Dir>" + mslogs + 
            "</Dir></MessageLogs>" + "</Logs></GKLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_MSLOG, mslogs);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the Event log directory is set when the related XML 
     * element is parsed.
     */
    public void testEventLogs() {
        String eventLogs = "/var/log/pbslogs";
        try {
            String xml = m_configFileStart + "<EventLogProcessor><Logs>" +
            "<Dir>" + eventLogs + "</Dir></Logs>" + "</EventLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_EVENT_LOG_DIR, eventLogs);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the Event message processor is set when the related
     * XML element is parsed.
     */
    public void testEventProcessing() {
        try {
            String xml = m_configFileStart + "<EventLogProcessor/>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_EVENT_LOG_PROCESSOR));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the recursive flag is set when the GK processor element
     * is parsed.
     */
    public void testGKSubDirs() {
        try {
            String xml = m_configFileStart + "<GKLogProcessor>" + "<Logs searchSubDirs=\"" 
            + m_yesValue + "\"/></GKLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_GK_SEARCH_ALL));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the recursive flag is set when the Event processor element
     * is parsed.
     */
    public void testEventSubDirs() {
        try {
            String xml = m_configFileStart + "<EventLogProcessor>" + "<Logs searchSubDirs=\"" 
            + m_yesValue + "\"/></EventLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_EVENT_SEARCH_ALL));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the reprocess flag is set when the related GKProcessor 
     * element is parsed.
     */
    public void testGKLogReprocess() {
        try {
            String xml = m_configFileStart + "<GKLogProcessor><Logs " + "reprocess=\"" 
            + m_yesValue + "\"/></GKLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_GK_LOG_REPROCESS));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the reprocess flag is set when the related EventProcessor 
     * element is parsed.
     */
    public void testEventLogReprocess() {
       try {
            String xml = m_configFileStart + "<EventLogProcessor><Logs " + "reprocess=\"" 
            + m_yesValue + "\"/></EventLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            assertTrue(config.isOptionSet(FileMap.OPT_EVENT_LOG_REPROCESS));
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the site name is set when the related XML 
     * element is parsed.
     */
    public void testSiteName() {
        String siteName = "testsite";
        try {
            String xml = m_configFileStart + "<SiteName>" + siteName + "</SiteName>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_SITE, siteName);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the submit host name is set when the related XML 
     * element is parsed.
     */
    public void testSubmitHost() {
        String submitHost = "testHost";
        try {
            String xml = m_configFileStart + "<GKLogProcessor><SubmitHost>" 
            + submitHost + "</SubmitHost></GKLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_SUBMIT_HOST, submitHost);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    /**
     * Tests the timezone value is set when the related XML
     * element is parsed.
     */
    public void testTimezone() {
        String timeZone = "UTC";
        try {
            String xml = m_configFileStart + "<EventLogProcessor><Timezone>" 
            + timeZone + "</Timezone></EventLogProcessor>" + m_configFileEnd;
            XMLParser parser = new XMLParser(new StringReader(xml));
            Config config = parser.parse(new FileMap());
            checkSingleProp(config, FileMap.PROP_TIME_ZONE, timeZone);
        } catch (ConfigurationException e) {
            fail(e.toString());
        }
    }
    private void checkSingleProp(Config config, String prop, String value) {
        String propVal = config.getProperty(prop);
    	assertTrue(propVal != null);
        assertEquals(config.getNumberProperties(), 1);
        assertEquals(value, propVal);
    } 
}
